@@ -15,7 +15,7 @@ const bench = common.createBenchmark(main, {
   len: [64 * 1024, 128 * 1024, 256 * 1024, 1024 * 1024],
   c: [100],
   method: ['write', 'end'],
-  duration: 5
+  duration: 5,
 });
 
 function main({ len, type, method, c, duration }) {
@@ -48,10 +48,11 @@ function main({ len, type, method, c, duration }) {
     fn(res);
   });
 
-  server.listen(common.PORT, () => {
+  server.listen(0, () => {
     bench.http({
       connections: c,
-      duration
+      duration,
+      port: server.address().port,
     }, () => {
       server.close();
     });
