@@ -1,17 +1,16 @@
 'use strict';
 
 const common = require('../common.js');
-const PORT = common.PORT;
 
 const bench = common.createBenchmark(main, {
   n: [1e3],
-  nheaders: [0, 10, 100, 1000]
+  nheaders: [0, 10, 100, 1000],
 }, { flags: ['--no-warnings'] });
 
 function main({ n, nheaders }) {
   const http2 = require('http2');
   const server = http2.createServer({
-    maxHeaderListPairs: 20000
+    maxHeaderListPairs: 20000,
   });
 
   const headersObject = {
@@ -21,7 +20,7 @@ function main({ n, nheaders }) {
     'accept-language': 'en',
     'content-type': 'text/plain',
     'referer': 'https://example.org/',
-    'user-agent': 'SuperBenchmarker 3000'
+    'user-agent': 'SuperBenchmarker 3000',
   };
 
   for (let i = 0; i < nheaders; i++) {
@@ -32,9 +31,9 @@ function main({ n, nheaders }) {
     stream.respond();
     stream.end('Hi!');
   });
-  server.listen(PORT, () => {
-    const client = http2.connect(`http://localhost:${PORT}/`, {
-      maxHeaderListPairs: 20000
+  server.listen(0, () => {
+    const client = http2.connect(`http://localhost:${server.address().port}/`, {
+      maxHeaderListPairs: 20000,
     });
 
     function doRequest(remaining) {
