@@ -13,14 +13,14 @@ const bench = common.createBenchmark(main, {
     'foo\\bar\\..baz.quux',
     'foo\\bar\\...baz.quux',
     'D:\\foo\\bar\\baz\\asdf\\quux',
-    '\\foo\\bar\\baz\\asdf\\quux.foobarbazasdfquux'
+    '\\foo\\bar\\baz\\asdf\\quux.foobarbazasdfquux',
   ],
-  n: [1e5]
+  n: [1e5],
 });
 
 function main({ n, path }) {
   bench.start();
-  for (var i = 0; i < n; i++) {
+  for (let i = 0; i < n; i++) {
     win32.extname(i % 3 === 0 ? `${path}${i}` : path);
   }
   bench.end(n);
