@@ -20,13 +20,14 @@ function main({ n }) {
     if (count === n)
       bench.end(n);
   }
+
   function cb2() {
     count++;
     if (count === n)
       bench.end(n);
   }
 
-  for (var i = 0; i < n; i++) {
+  for (let i = 0; i < n; i++) {
     setTimeout(i % 2 ? cb : cb2, 1);
   }
 
