@@ -33,11 +33,14 @@
 #include "nghttp3_err.h"
 #include "nghttp3_conv.h"
 #include "nghttp3_http.h"
+#include "nghttp3_unreachable.h"
 
 /* NGHTTP3_QPACK_ENCODER_MAX_DTABLE_CAPACITY is the upper bound of the
    dynamic table capacity that QPACK encoder is willing to use. */
 #define NGHTTP3_QPACK_ENCODER_MAX_DTABLE_CAPACITY 4096
 
+nghttp3_objalloc_def(chunk, nghttp3_chunk, oplent);
+
 /*
  * conn_remote_stream_uni returns nonzero if |stream_id| is remote
  * unidirectional stream ID.
@@ -56,7 +59,7 @@ static int conn_call_begin_headers(nghttp3_conn *conn, nghttp3_stream *stream) {
     return 0;
   }
 
-  rv = conn->callbacks.begin_headers(conn, stream->node.nid.id, conn->user_data,
+  rv = conn->callbacks.begin_headers(conn, stream->node.id, conn->user_data,
                                      stream->user_data);
   if (rv != 0) {
     /* TODO Allow ignore headers */
@@ -74,8 +77,8 @@ static int conn_call_end_headers(nghttp3_conn *conn, nghttp3_stream *stream,
     return 0;
   }
 
-  rv = conn->callbacks.end_headers(conn, stream->node.nid.id, fin,
-                                   conn->user_data, stream->user_data);
+  rv = conn->callbacks.end_headers(conn, stream->node.id, fin, conn->user_data,
+                                   stream->user_data);
   if (rv != 0) {
     /* TODO Allow ignore headers */
     return NGHTTP3_ERR_CALLBACK_FAILURE;
@@ -92,8 +95,8 @@ static int conn_call_begin_trailers(nghttp3_conn *conn,
     return 0;
   }
 
-  rv = conn->callbacks.begin_trailers(conn, stream->node.nid.id,
-                                      conn->user_data, stream->user_data);
+  rv = conn->callbacks.begin_trailers(conn, stream->node.id, conn->user_data,
+                                      stream->user_data);
   if (rv != 0) {
     /* TODO Allow ignore headers */
     return NGHTTP3_ERR_CALLBACK_FAILURE;
@@ -110,8 +113,8 @@ static int conn_call_end_trailers(nghttp3_conn *conn, nghttp3_stream *stream,
     return 0;
   }
 
-  rv = conn->callbacks.end_trailers(conn, stream->node.nid.id, fin,
-                                    conn->user_data, stream->user_data);
+  rv = conn->callbacks.end_trailers(conn, stream->node.id, fin, conn->user_data,
+                                    stream->user_data);
   if (rv != 0) {
     /* TODO Allow ignore headers */
     return NGHTTP3_ERR_CALLBACK_FAILURE;
@@ -127,7 +130,7 @@ static int conn_call_end_stream(nghttp3_conn *conn, nghttp3_stream *stream) {
     return 0;
   }
 
-  rv = conn->callbacks.end_stream(conn, stream->node.nid.id, conn->user_data,
+  rv = conn->callbacks.end_stream(conn, stream->node.id, conn->user_data,
                                   stream->user_data);
   if (rv != 0) {
     return NGHTTP3_ERR_CALLBACK_FAILURE;
@@ -144,7 +147,7 @@ static int conn_call_stop_sending(nghttp3_conn *conn, nghttp3_stream *stream,
     return 0;
   }
 
-  rv = conn->callbacks.stop_sending(conn, stream->node.nid.id, app_error_code,
+  rv = conn->callbacks.stop_sending(conn, stream->node.id, app_error_code,
                                     conn->user_data, stream->user_data);
   if (rv != 0) {
     return NGHTTP3_ERR_CALLBACK_FAILURE;
@@ -161,7 +164,7 @@ static int conn_call_reset_stream(nghttp3_conn *conn, nghttp3_stream *stream,
     return 0;
   }
 
-  rv = conn->callbacks.reset_stream(conn, stream->node.nid.id, app_error_code,
+  rv = conn->callbacks.reset_stream(conn, stream->node.id, app_error_code,
                                     conn->user_data, stream->user_data);
   if (rv != 0) {
     return NGHTTP3_ERR_CALLBACK_FAILURE;
@@ -179,7 +182,7 @@ static int conn_call_deferred_consume(nghttp3_conn *conn,
     return 0;
   }
 
-  rv = conn->callbacks.deferred_consume(conn, stream->node.nid.id, nconsumed,
+  rv = conn->callbacks.deferred_consume(conn, stream->node.id, nconsumed,
                                         conn->user_data, stream->user_data);
   if (rv != 0) {
     return NGHTTP3_ERR_CALLBACK_FAILURE;
@@ -188,6 +191,22 @@ static int conn_call_deferred_consume(nghttp3_conn *conn,
   return 0;
 }
 
+static int conn_call_recv_settings(nghttp3_conn *conn) {
+  int rv;
+
+  if (!conn->callbacks.recv_settings) {
+    return 0;
+  }
+
+  rv = conn->callbacks.recv_settings(conn, &conn->remote.settings,
+                                     conn->user_data);
+  if (rv != 0) {
+    return NGHTTP3_ERR_CALLBACK_FAILURE;
+  }
+
+  return 0;
+}
+
 static int ricnt_less(const nghttp3_pq_entry *lhsx,
                       const nghttp3_pq_entry *rhsx) {
   nghttp3_stream *lhs =
@@ -204,7 +223,7 @@ static int cycle_less(const nghttp3_pq_entry *lhsx,
   const nghttp3_tnode *rhs = nghttp3_struct_of(rhsx, nghttp3_tnode, pe);
 
   if (lhs->cycle == rhs->cycle) {
-    return lhs->seq < rhs->seq;
+    return lhs->id < rhs->id;
   }
 
   return rhs->cycle - lhs->cycle <= NGHTTP3_TNODE_MAX_CYCLE_GAP;
@@ -264,7 +283,6 @@ static int conn_new(nghttp3_conn **pconn, int server, int callbacks_version,
   nghttp3_settings_default(&conn->remote.settings);
   conn->mem = mem;
   conn->user_data = user_data;
-  conn->next_seq = 0;
   conn->server = server;
   conn->rx.goaway_id = NGHTTP3_VARINT_MAX + 1;
   conn->tx.goaway_id = NGHTTP3_VARINT_MAX + 1;
@@ -454,12 +472,6 @@ nghttp3_ssize nghttp3_conn_read_stream(nghttp3_conn *conn, int64_t stream_id,
         stream->tx.hstate = NGHTTP3_HTTP_STATE_REQ_INITIAL;
       }
     }
-  } else if (nghttp3_stream_uni(stream_id) &&
-             stream->type == NGHTTP3_STREAM_TYPE_PUSH) {
-    if (stream->rx.hstate == NGHTTP3_HTTP_STATE_NONE) {
-      stream->rx.hstate = NGHTTP3_HTTP_STATE_RESP_INITIAL;
-      stream->tx.hstate = NGHTTP3_HTTP_STATE_RESP_INITIAL;
-    }
   }
 
   if (srclen == 0 && !fin) {
@@ -603,8 +615,7 @@ nghttp3_ssize nghttp3_conn_read_uni(nghttp3_conn *conn, nghttp3_stream *stream,
     }
     break;
   default:
-    /* unreachable */
-    assert(0);
+    nghttp3_unreachable();
   }
 
   if (nconsumed < 0) {
@@ -686,6 +697,11 @@ nghttp3_ssize nghttp3_conn_read_control(nghttp3_conn *conn,
       case NGHTTP3_FRAME_SETTINGS:
         /* SETTINGS frame might be empty. */
         if (rstate->left == 0) {
+          rv = conn_call_recv_settings(conn);
+          if (rv != 0) {
+            return rv;
+          }
+
           nghttp3_stream_read_state_reset(rstate);
           break;
         }
@@ -735,11 +751,21 @@ nghttp3_ssize nghttp3_conn_read_control(nghttp3_conn *conn,
       }
       break;
     case NGHTTP3_CTRL_STREAM_STATE_SETTINGS:
-      for (; p != end;) {
+      for (;;) {
         if (rstate->left == 0) {
+          rv = conn_call_recv_settings(conn);
+          if (rv != 0) {
+            return rv;
+          }
+
           nghttp3_stream_read_state_reset(rstate);
           break;
         }
+
+        if (p == end) {
+          return (nghttp3_ssize)nconsumed;
+        }
+
         /* Read Identifier */
         len = (size_t)nghttp3_min(rstate->left, (int64_t)(end - p));
         assert(len > 0);
@@ -845,6 +871,11 @@ nghttp3_ssize nghttp3_conn_read_control(nghttp3_conn *conn,
         break;
       }
 
+      rv = conn_call_recv_settings(conn);
+      if (rv != 0) {
+        return rv;
+      }
+
       nghttp3_stream_read_state_reset(rstate);
       break;
     case NGHTTP3_CTRL_STREAM_STATE_GOAWAY:
@@ -1013,8 +1044,7 @@ nghttp3_ssize nghttp3_conn_read_control(nghttp3_conn *conn,
       nghttp3_stream_read_state_reset(rstate);
       break;
     default:
-      /* unreachable */
-      assert(0);
+      nghttp3_unreachable();
     }
   }
 
@@ -1022,7 +1052,7 @@ nghttp3_ssize nghttp3_conn_read_control(nghttp3_conn *conn,
 }
 
 static int conn_delete_stream(nghttp3_conn *conn, nghttp3_stream *stream) {
-  int bidi = nghttp3_client_stream_bidi(stream->node.nid.id);
+  int bidi = nghttp3_client_stream_bidi(stream->node.id);
   int rv;
 
   rv = conn_call_deferred_consume(conn, stream,
@@ -1032,16 +1062,21 @@ static int conn_delete_stream(nghttp3_conn *conn, nghttp3_stream *stream) {
   }
 
   if (bidi && conn->callbacks.stream_close) {
-    rv = conn->callbacks.stream_close(conn, stream->node.nid.id,
-                                      stream->error_code, conn->user_data,
-                                      stream->user_data);
+    rv = conn->callbacks.stream_close(conn, stream->node.id, stream->error_code,
+                                      conn->user_data, stream->user_data);
     if (rv != 0) {
       return NGHTTP3_ERR_CALLBACK_FAILURE;
     }
   }
 
-  rv = nghttp3_map_remove(&conn->streams,
-                          (nghttp3_map_key_type)stream->node.nid.id);
+  if (conn->server && nghttp3_client_stream_bidi(stream->node.id)) {
+    assert(conn->remote.bidi.num_streams > 0);
+
+    --conn->remote.bidi.num_streams;
+  }
+
+  rv =
+      nghttp3_map_remove(&conn->streams, (nghttp3_map_key_type)stream->node.id);
 
   assert(0 == rv);
 
@@ -1058,7 +1093,7 @@ static int conn_process_blocked_stream_data(nghttp3_conn *conn,
   int rv;
   size_t len;
 
-  assert(nghttp3_client_stream_bidi(stream->node.nid.id));
+  assert(nghttp3_client_stream_bidi(stream->node.id));
 
   for (;;) {
     len = nghttp3_ringbuf_len(&stream->inq);
@@ -1149,16 +1184,17 @@ static nghttp3_tnode *stream_get_sched_node(nghttp3_stream *stream) {
 }
 
 static int conn_update_stream_priority(nghttp3_conn *conn,
-                                       nghttp3_stream *stream, uint8_t pri) {
-  assert(nghttp3_client_stream_bidi(stream->node.nid.id));
+                                       nghttp3_stream *stream,
+                                       const nghttp3_pri *pri) {
+  assert(nghttp3_client_stream_bidi(stream->node.id));
 
-  if (stream->node.pri == pri) {
+  if (nghttp3_pri_eq(&stream->node.pri, pri)) {
     return 0;
   }
 
   nghttp3_conn_unschedule_stream(conn, stream);
 
-  stream->node.pri = pri;
+  stream->node.pri = *pri;
 
   if (nghttp3_stream_require_schedule(stream)) {
     return nghttp3_conn_schedule_stream(conn, stream);
@@ -1286,8 +1322,7 @@ nghttp3_ssize nghttp3_conn_read_bidi(nghttp3_conn *conn, size_t *pnproc,
           rv = conn_call_begin_trailers(conn, stream);
           break;
         default:
-          /* Unreachable */
-          assert(0);
+          nghttp3_unreachable();
         }
 
         if (rv != 0) {
@@ -1377,9 +1412,7 @@ nghttp3_ssize nghttp3_conn_read_bidi(nghttp3_conn *conn, size_t *pnproc,
         rv = 0;
         break;
       default:
-        /* Unreachable */
-        assert(0);
-        abort();
+        nghttp3_unreachable();
       }
 
       if (rv != 0) {
@@ -1398,7 +1431,7 @@ nghttp3_ssize nghttp3_conn_read_bidi(nghttp3_conn *conn, size_t *pnproc,
             (stream->rx.http.flags & NGHTTP3_HTTP_FLAG_PRIORITY) &&
             !(stream->flags & NGHTTP3_STREAM_FLAG_PRIORITY_UPDATE_RECVED) &&
             !(stream->flags & NGHTTP3_STREAM_FLAG_SERVER_PRIORITY_SET)) {
-          rv = conn_update_stream_priority(conn, stream, stream->rx.http.pri);
+          rv = conn_update_stream_priority(conn, stream, &stream->rx.http.pri);
           if (rv != 0) {
             return rv;
           }
@@ -1412,8 +1445,7 @@ nghttp3_ssize nghttp3_conn_read_bidi(nghttp3_conn *conn, size_t *pnproc,
         rv = conn_call_end_trailers(conn, stream, p == end && fin);
         break;
       default:
-        /* Unreachable */
-        assert(0);
+        nghttp3_unreachable();
       }
 
       if (rv != 0) {
@@ -1505,7 +1537,7 @@ int nghttp3_conn_on_data(nghttp3_conn *conn, nghttp3_stream *stream,
     return 0;
   }
 
-  rv = conn->callbacks.recv_data(conn, stream->node.nid.id, data, datalen,
+  rv = conn->callbacks.recv_data(conn, stream->node.id, data, datalen,
                                  conn->user_data, stream->user_data);
   if (rv != 0) {
     return NGHTTP3_ERR_CALLBACK_FAILURE;
@@ -1515,11 +1547,9 @@ int nghttp3_conn_on_data(nghttp3_conn *conn, nghttp3_stream *stream,
 }
 
 static nghttp3_pq *conn_get_sched_pq(nghttp3_conn *conn, nghttp3_tnode *tnode) {
-  uint32_t urgency = nghttp3_pri_uint8_urgency(tnode->pri);
+  assert(tnode->pri.urgency < NGHTTP3_URGENCY_LEVELS);
 
-  assert(urgency < NGHTTP3_URGENCY_LEVELS);
-
-  return &conn->sched[urgency].spq;
+  return &conn->sched[tnode->pri.urgency].spq;
 }
 
 static nghttp3_ssize conn_decode_headers(nghttp3_conn *conn,
@@ -1552,8 +1582,7 @@ static nghttp3_ssize conn_decode_headers(nghttp3_conn *conn,
     recv_header = conn->callbacks.recv_trailer;
     break;
   default:
-    /* Unreachable */
-    assert(0);
+    nghttp3_unreachable();
   }
   http = &stream->rx.http;
 
@@ -1606,17 +1635,15 @@ static nghttp3_ssize conn_decode_headers(nghttp3_conn *conn,
         break;
       case 0:
         if (recv_header) {
-          rv = recv_header(conn, stream->node.nid.id, nv.token, nv.name,
-                           nv.value, nv.flags, conn->user_data,
-                           stream->user_data);
+          rv = recv_header(conn, stream->node.id, nv.token, nv.name, nv.value,
+                           nv.flags, conn->user_data, stream->user_data);
           if (rv != 0) {
             rv = NGHTTP3_ERR_CALLBACK_FAILURE;
           }
         }
         break;
       default:
-        /* Unreachable */
-        assert(0);
+        nghttp3_unreachable();
       }
 
       nghttp3_rcbuf_decref(nv.name);
@@ -1684,13 +1711,32 @@ int nghttp3_conn_on_settings_entry_received(nghttp3_conn *conn,
     if (!conn->server) {
       break;
     }
-    if (ent->value != 0 && ent->value != 1) {
+
+    switch (ent->value) {
+    case 0:
+      if (dest->enable_connect_protocol) {
+        return NGHTTP3_ERR_H3_SETTINGS_ERROR;
+      }
+
+      break;
+    case 1:
+      break;
+    default:
       return NGHTTP3_ERR_H3_SETTINGS_ERROR;
     }
-    if (ent->value == 0 && dest->enable_connect_protocol) {
+
+    dest->enable_connect_protocol = (uint8_t)ent->value;
+    break;
+  case NGHTTP3_SETTINGS_ID_H3_DATAGRAM:
+    switch (ent->value) {
+    case 0:
+    case 1:
+      break;
+    default:
       return NGHTTP3_ERR_H3_SETTINGS_ERROR;
     }
-    dest->enable_connect_protocol = (int)ent->value;
+
+    dest->h3_datagram = (uint8_t)ent->value;
     break;
   case NGHTTP3_H2_SETTINGS_ID_ENABLE_PUSH:
   case NGHTTP3_H2_SETTINGS_ID_MAX_CONCURRENT_STREAMS:
@@ -1744,7 +1790,7 @@ conn_on_priority_update_stream(nghttp3_conn *conn,
       return rv;
     }
 
-    stream->node.pri = nghttp3_pri_to_uint8(&fr->pri);
+    stream->node.pri = fr->pri;
     stream->flags |= NGHTTP3_STREAM_FLAG_PRIORITY_UPDATE_RECVED;
 
     return 0;
@@ -1756,8 +1802,7 @@ conn_on_priority_update_stream(nghttp3_conn *conn,
 
   stream->flags |= NGHTTP3_STREAM_FLAG_PRIORITY_UPDATE_RECVED;
 
-  return conn_update_stream_priority(conn, stream,
-                                     nghttp3_pri_to_uint8(&fr->pri));
+  return conn_update_stream_priority(conn, stream, &fr->pri);
 }
 
 int nghttp3_conn_on_priority_update(nghttp3_conn *conn,
@@ -1794,7 +1839,7 @@ int nghttp3_conn_create_stream(nghttp3_conn *conn, nghttp3_stream **pstream,
       conn_stream_acked_data,
   };
 
-  rv = nghttp3_stream_new(&stream, stream_id, conn->next_seq, &callbacks,
+  rv = nghttp3_stream_new(&stream, stream_id, &callbacks,
                           &conn->out_chunk_objalloc, &conn->stream_objalloc,
                           conn->mem);
   if (rv != 0) {
@@ -1803,14 +1848,17 @@ int nghttp3_conn_create_stream(nghttp3_conn *conn, nghttp3_stream **pstream,
 
   stream->conn = conn;
 
-  rv = nghttp3_map_insert(&conn->streams,
-                          (nghttp3_map_key_type)stream->node.nid.id, stream);
+  rv = nghttp3_map_insert(&conn->streams, (nghttp3_map_key_type)stream->node.id,
+                          stream);
   if (rv != 0) {
     nghttp3_stream_del(stream);
     return rv;
   }
 
-  ++conn->next_seq;
+  if (conn->server && nghttp3_client_stream_bidi(stream_id)) {
+    ++conn->remote.bidi.num_streams;
+  }
+
   *pstream = stream;
 
   return 0;
@@ -1897,7 +1945,7 @@ static nghttp3_ssize conn_writev_stream(nghttp3_conn *conn, int64_t *pstream_id,
                                         int *pfin, nghttp3_vec *vec,
                                         size_t veccnt, nghttp3_stream *stream) {
   int rv;
-  nghttp3_ssize n;
+  size_t n;
 
   assert(veccnt > 0);
 
@@ -1910,31 +1958,25 @@ static nghttp3_ssize conn_writev_stream(nghttp3_conn *conn, int64_t *pstream_id,
     }
   }
 
-  if (!nghttp3_stream_uni(stream->node.nid.id) && conn->tx.qenc &&
+  if (!nghttp3_stream_uni(stream->node.id) && conn->tx.qenc &&
       !nghttp3_stream_is_blocked(conn->tx.qenc)) {
     n = nghttp3_stream_writev(conn->tx.qenc, pfin, vec, veccnt);
-    if (n < 0) {
-      return n;
-    }
     if (n) {
-      *pstream_id = conn->tx.qenc->node.nid.id;
-      return n;
+      *pstream_id = conn->tx.qenc->node.id;
+      return (nghttp3_ssize)n;
     }
   }
 
   n = nghttp3_stream_writev(stream, pfin, vec, veccnt);
-  if (n < 0) {
-    return n;
-  }
   /* We might just want to write stream fin without sending any stream
      data. */
   if (n == 0 && *pfin == 0) {
     return 0;
   }
 
-  *pstream_id = stream->node.nid.id;
+  *pstream_id = stream->node.id;
 
-  return n;
+  return (nghttp3_ssize)n;
 }
 
 nghttp3_ssize nghttp3_conn_writev_stream(nghttp3_conn *conn,
@@ -1990,7 +2032,7 @@ nghttp3_ssize nghttp3_conn_writev_stream(nghttp3_conn *conn,
     return ncnt;
   }
 
-  if (nghttp3_client_stream_bidi(stream->node.nid.id) &&
+  if (nghttp3_client_stream_bidi(stream->node.id) &&
       !nghttp3_stream_require_schedule(stream)) {
     nghttp3_conn_unschedule_stream(conn, stream);
   }
@@ -2020,20 +2062,16 @@ nghttp3_stream *nghttp3_conn_get_next_tx_stream(nghttp3_conn *conn) {
 int nghttp3_conn_add_write_offset(nghttp3_conn *conn, int64_t stream_id,
                                   size_t n) {
   nghttp3_stream *stream = nghttp3_conn_find_stream(conn, stream_id);
-  int rv;
 
   if (stream == NULL) {
     return 0;
   }
 
-  rv = nghttp3_stream_add_outq_offset(stream, n);
-  if (rv != 0) {
-    return rv;
-  }
+  nghttp3_stream_add_outq_offset(stream, n);
 
   stream->unscheduled_nwrite += n;
 
-  if (!nghttp3_client_stream_bidi(stream->node.nid.id)) {
+  if (!nghttp3_client_stream_bidi(stream->node.id)) {
     return 0;
   }
 
@@ -2065,7 +2103,7 @@ static int conn_submit_headers_data(nghttp3_conn *conn, nghttp3_stream *stream,
                                     const nghttp3_data_reader *dr) {
   int rv;
   nghttp3_nv *nnva;
-  nghttp3_frame_entry frent;
+  nghttp3_frame_entry frent = {0};
 
   rv = nghttp3_nva_copy(&nnva, nva, nvlen, conn->mem);
   if (rv != 0) {
@@ -2235,7 +2273,7 @@ int nghttp3_conn_submit_trailers(nghttp3_conn *conn, int64_t stream_id,
 }
 
 int nghttp3_conn_submit_shutdown_notice(nghttp3_conn *conn) {
-  nghttp3_frame_entry frent;
+  nghttp3_frame_entry frent = {0};
   int rv;
 
   assert(conn->tx.ctrl);
@@ -2258,7 +2296,7 @@ int nghttp3_conn_submit_shutdown_notice(nghttp3_conn *conn) {
 }
 
 int nghttp3_conn_shutdown(nghttp3_conn *conn) {
-  nghttp3_frame_entry frent;
+  nghttp3_frame_entry frent = {0};
   int rv;
 
   assert(conn->tx.ctrl);
@@ -2279,7 +2317,8 @@ int nghttp3_conn_shutdown(nghttp3_conn *conn) {
   }
 
   conn->tx.goaway_id = frent.fr.goaway.id;
-  conn->flags |= NGHTTP3_CONN_FLAG_GOAWAY_QUEUED;
+  conn->flags |=
+      NGHTTP3_CONN_FLAG_GOAWAY_QUEUED | NGHTTP3_CONN_FLAG_SHUTDOWN_COMMENCED;
 
   return 0;
 }
@@ -2305,7 +2344,7 @@ void nghttp3_conn_block_stream(nghttp3_conn *conn, int64_t stream_id) {
   stream->flags |= NGHTTP3_STREAM_FLAG_FC_BLOCKED;
   stream->unscheduled_nwrite = 0;
 
-  if (nghttp3_client_stream_bidi(stream->node.nid.id)) {
+  if (nghttp3_client_stream_bidi(stream->node.id)) {
     nghttp3_conn_unschedule_stream(conn, stream);
   }
 }
@@ -2320,7 +2359,7 @@ void nghttp3_conn_shutdown_stream_write(nghttp3_conn *conn, int64_t stream_id) {
   stream->flags |= NGHTTP3_STREAM_FLAG_SHUT_WR;
   stream->unscheduled_nwrite = 0;
 
-  if (nghttp3_client_stream_bidi(stream->node.nid.id)) {
+  if (nghttp3_client_stream_bidi(stream->node.id)) {
     nghttp3_conn_unschedule_stream(conn, stream);
   }
 }
@@ -2334,7 +2373,7 @@ int nghttp3_conn_unblock_stream(nghttp3_conn *conn, int64_t stream_id) {
 
   stream->flags &= (uint16_t)~NGHTTP3_STREAM_FLAG_FC_BLOCKED;
 
-  if (nghttp3_client_stream_bidi(stream->node.nid.id) &&
+  if (nghttp3_client_stream_bidi(stream->node.id) &&
       nghttp3_stream_require_schedule(stream)) {
     return nghttp3_conn_ensure_stream_scheduled(conn, stream);
   }
@@ -2364,7 +2403,7 @@ int nghttp3_conn_resume_stream(nghttp3_conn *conn, int64_t stream_id) {
 
   stream->flags &= (uint16_t)~NGHTTP3_STREAM_FLAG_READ_DATA_BLOCKED;
 
-  if (nghttp3_client_stream_bidi(stream->node.nid.id) &&
+  if (nghttp3_client_stream_bidi(stream->node.id) &&
       nghttp3_stream_require_schedule(stream)) {
     return nghttp3_conn_ensure_stream_scheduled(conn, stream);
   }
@@ -2381,7 +2420,6 @@ int nghttp3_conn_close_stream(nghttp3_conn *conn, int64_t stream_id,
   }
 
   if (nghttp3_stream_uni(stream_id) &&
-      stream->type != NGHTTP3_STREAM_TYPE_PUSH &&
       stream->type != NGHTTP3_STREAM_TYPE_UNKNOWN) {
     return NGHTTP3_ERR_H3_CLOSED_CRITICAL_STREAM;
   }
@@ -2459,18 +2497,34 @@ int nghttp3_conn_set_stream_user_data(nghttp3_conn *conn, int64_t stream_id,
 
 uint64_t nghttp3_conn_get_frame_payload_left(nghttp3_conn *conn,
                                              int64_t stream_id) {
-  nghttp3_stream *stream = nghttp3_conn_find_stream(conn, stream_id);
+  nghttp3_stream *stream;
+  int uni = 0;
 
+  if (!nghttp3_client_stream_bidi(stream_id)) {
+    uni = conn_remote_stream_uni(conn, stream_id);
+    if (!uni) {
+      return 0;
+    }
+  }
+
+  stream = nghttp3_conn_find_stream(conn, stream_id);
   if (stream == NULL) {
     return 0;
   }
 
+  if (uni && stream->type != NGHTTP3_STREAM_TYPE_CONTROL) {
+    return 0;
+  }
+
   return (uint64_t)stream->rstate.left;
 }
 
-int nghttp3_conn_get_stream_priority(nghttp3_conn *conn, nghttp3_pri *dest,
-                                     int64_t stream_id) {
+int nghttp3_conn_get_stream_priority_versioned(nghttp3_conn *conn,
+                                               int pri_version,
+                                               nghttp3_pri *dest,
+                                               int64_t stream_id) {
   nghttp3_stream *stream;
+  (void)pri_version;
 
   assert(conn->server);
 
@@ -2483,17 +2537,55 @@ int nghttp3_conn_get_stream_priority(nghttp3_conn *conn, nghttp3_pri *dest,
     return NGHTTP3_ERR_STREAM_NOT_FOUND;
   }
 
-  dest->urgency = nghttp3_pri_uint8_urgency(stream->node.pri);
-  dest->inc = nghttp3_pri_uint8_inc(stream->node.pri);
+  *dest = stream->node.pri;
 
   return 0;
 }
 
-int nghttp3_conn_set_stream_priority(nghttp3_conn *conn, int64_t stream_id,
-                                     const nghttp3_pri *pri) {
+int nghttp3_conn_set_client_stream_priority(nghttp3_conn *conn,
+                                            int64_t stream_id,
+                                            const uint8_t *data,
+                                            size_t datalen) {
   nghttp3_stream *stream;
-  nghttp3_frame_entry frent;
+  nghttp3_frame_entry frent = {0};
+  uint8_t *buf = NULL;
 
+  assert(!conn->server);
+
+  if (!nghttp3_client_stream_bidi(stream_id)) {
+    return NGHTTP3_ERR_INVALID_ARGUMENT;
+  }
+
+  stream = nghttp3_conn_find_stream(conn, stream_id);
+  if (stream == NULL) {
+    return NGHTTP3_ERR_STREAM_NOT_FOUND;
+  }
+
+  if (datalen) {
+    buf = nghttp3_mem_malloc(conn->mem, datalen);
+    if (buf == NULL) {
+      return NGHTTP3_ERR_NOMEM;
+    }
+
+    memcpy(buf, data, datalen);
+  }
+
+  frent.fr.hd.type = NGHTTP3_FRAME_PRIORITY_UPDATE;
+  frent.fr.priority_update.pri_elem_id = stream_id;
+  frent.fr.priority_update.data = buf;
+  frent.fr.priority_update.datalen = datalen;
+
+  return nghttp3_stream_frq_add(conn->tx.ctrl, &frent);
+}
+
+int nghttp3_conn_set_server_stream_priority_versioned(nghttp3_conn *conn,
+                                                      int64_t stream_id,
+                                                      int pri_version,
+                                                      const nghttp3_pri *pri) {
+  nghttp3_stream *stream;
+  (void)pri_version;
+
+  assert(conn->server);
   assert(pri->urgency < NGHTTP3_URGENCY_LEVELS);
   assert(pri->inc == 0 || pri->inc == 1);
 
@@ -2506,29 +2598,18 @@ int nghttp3_conn_set_stream_priority(nghttp3_conn *conn, int64_t stream_id,
     return NGHTTP3_ERR_STREAM_NOT_FOUND;
   }
 
-  if (conn->server) {
-    stream->flags |= NGHTTP3_STREAM_FLAG_SERVER_PRIORITY_SET;
+  stream->flags |= NGHTTP3_STREAM_FLAG_SERVER_PRIORITY_SET;
 
-    return conn_update_stream_priority(conn, stream, nghttp3_pri_to_uint8(pri));
-  }
-
-  frent.fr.hd.type = NGHTTP3_FRAME_PRIORITY_UPDATE;
-  frent.fr.priority_update.pri_elem_id = stream_id;
-  frent.fr.priority_update.pri = *pri;
-
-  return nghttp3_stream_frq_add(conn->tx.ctrl, &frent);
+  return conn_update_stream_priority(conn, stream, pri);
 }
 
-int nghttp3_conn_is_remote_qpack_encoder_stream(nghttp3_conn *conn,
-                                                int64_t stream_id) {
-  nghttp3_stream *stream;
+int nghttp3_conn_is_drained(nghttp3_conn *conn) {
+  assert(conn->server);
 
-  if (!conn_remote_stream_uni(conn, stream_id)) {
-    return 0;
-  }
-
-  stream = nghttp3_conn_find_stream(conn, stream_id);
-  return stream && stream->type == NGHTTP3_STREAM_TYPE_QPACK_ENCODER;
+  return (conn->flags & NGHTTP3_CONN_FLAG_SHUTDOWN_COMMENCED) &&
+         conn->remote.bidi.num_streams == 0 &&
+         nghttp3_stream_outq_write_done(conn->tx.ctrl) &&
+         nghttp3_ringbuf_len(&conn->tx.ctrl->frq) == 0;
 }
 
 void nghttp3_settings_default_versioned(int settings_version,
