@@ -30,6 +30,8 @@ class AstTraversalVisitor : public AstVisitor<Subclass> {
  public:
   explicit AstTraversalVisitor(Isolate* isolate, AstNode* root = nullptr);
   explicit AstTraversalVisitor(uintptr_t stack_limit, AstNode* root = nullptr);
+  AstTraversalVisitor(const AstTraversalVisitor&) = delete;
+  AstTraversalVisitor& operator=(const AstTraversalVisitor&) = delete;
 
   void Run() {
     DCHECK_NOT_NULL(root_);
@@ -56,8 +58,6 @@ class AstTraversalVisitor : public AstVisitor<Subclass> {
 
   AstNode* root_;
   int depth_;
-
-  DISALLOW_COPY_AND_ASSIGN(AstTraversalVisitor);
 };
 
 // ----------------------------------------------------------------------------
@@ -294,13 +294,6 @@ void AstTraversalVisitor<Subclass>::VisitNativeFunctionLiteral(
   PROCESS_EXPRESSION(expr);
 }
 
-template <class Subclass>
-void AstTraversalVisitor<Subclass>::VisitDoExpression(DoExpression* expr) {
-  PROCESS_EXPRESSION(expr);
-  RECURSE(VisitBlock(expr->block()));
-  RECURSE(VisitVariableProxy(expr->result()));
-}
-
 template <class Subclass>
 void AstTraversalVisitor<Subclass>::VisitConditional(Conditional* expr) {
   PROCESS_EXPRESSION(expr);
@@ -395,14 +388,6 @@ void AstTraversalVisitor<Subclass>::VisitProperty(Property* expr) {
   RECURSE_EXPRESSION(Visit(expr->key()));
 }
 
-template <class Subclass>
-void AstTraversalVisitor<Subclass>::VisitResolvedProperty(
-    ResolvedProperty* expr) {
-  PROCESS_EXPRESSION(expr);
-  RECURSE_EXPRESSION(VisitVariableProxy(expr->object()));
-  RECURSE_EXPRESSION(VisitVariableProxy(expr->property()));
-}
-
 template <class Subclass>
 void AstTraversalVisitor<Subclass>::VisitCall(Call* expr) {
   PROCESS_EXPRESSION(expr);
@@ -484,13 +469,19 @@ void AstTraversalVisitor<Subclass>::VisitClassLiteral(ClassLiteral* expr) {
     RECURSE_EXPRESSION(Visit(expr->extends()));
   }
   RECURSE_EXPRESSION(Visit(expr->constructor()));
-  if (expr->static_fields_initializer() != nullptr) {
-    RECURSE_EXPRESSION(Visit(expr->static_fields_initializer()));
+  if (expr->static_initializer() != nullptr) {
+    RECURSE_EXPRESSION(Visit(expr->static_initializer()));
   }
   if (expr->instance_members_initializer_function() != nullptr) {
     RECURSE_EXPRESSION(Visit(expr->instance_members_initializer_function()));
   }
-  ZonePtrList<ClassLiteral::Property>* props = expr->properties();
+  ZonePtrList<ClassLiteral::Property>* private_members =
+      expr->private_members();
+  for (int i = 0; i < private_members->length(); ++i) {
+    ClassLiteralProperty* prop = private_members->at(i);
+    RECURSE_EXPRESSION(Visit(prop->value()));
+  }
+  ZonePtrList<ClassLiteral::Property>* props = expr->public_members();
   for (int i = 0; i < props->length(); ++i) {
     ClassLiteralProperty* prop = props->at(i);
     if (!prop->key()->IsLiteral()) {
@@ -515,18 +506,32 @@ void AstTraversalVisitor<Subclass>::VisitInitializeClassMembersStatement(
 }
 
 template <class Subclass>
-void AstTraversalVisitor<Subclass>::VisitSpread(Spread* expr) {
-  PROCESS_EXPRESSION(expr);
-  RECURSE_EXPRESSION(Visit(expr->expression()));
+void AstTraversalVisitor<Subclass>::VisitInitializeClassStaticElementsStatement(
+    InitializeClassStaticElementsStatement* stmt) {
+  PROCESS_NODE(stmt);
+  ZonePtrList<ClassLiteral::StaticElement>* elements = stmt->elements();
+  for (int i = 0; i < elements->length(); ++i) {
+    ClassLiteral::StaticElement* element = elements->at(i);
+    switch (element->kind()) {
+      case ClassLiteral::StaticElement::PROPERTY: {
+        ClassLiteral::Property* prop = element->property();
+        if (!prop->key()->IsLiteral()) {
+          RECURSE(Visit(prop->key()));
+        }
+        RECURSE(Visit(prop->value()));
+        break;
+      }
+      case ClassLiteral::StaticElement::STATIC_BLOCK:
+        RECURSE(Visit(element->static_block()));
+        break;
+    }
+  }
 }
 
 template <class Subclass>
-void AstTraversalVisitor<Subclass>::VisitStoreInArrayLiteral(
-    StoreInArrayLiteral* expr) {
+void AstTraversalVisitor<Subclass>::VisitSpread(Spread* expr) {
   PROCESS_EXPRESSION(expr);
-  RECURSE_EXPRESSION(Visit(expr->array()));
-  RECURSE_EXPRESSION(Visit(expr->index()));
-  RECURSE_EXPRESSION(Visit(expr->value()));
+  RECURSE_EXPRESSION(Visit(expr->expression()));
 }
 
 template <class Subclass>
@@ -554,14 +559,16 @@ template <class Subclass>
 void AstTraversalVisitor<Subclass>::VisitImportCallExpression(
     ImportCallExpression* expr) {
   PROCESS_EXPRESSION(expr);
-  RECURSE_EXPRESSION(Visit(expr->argument()));
+  RECURSE_EXPRESSION(Visit(expr->specifier()));
+  if (expr->import_assertions()) {
+    RECURSE_EXPRESSION(Visit(expr->import_assertions()));
+  }
 }
 
 template <class Subclass>
 void AstTraversalVisitor<Subclass>::VisitSuperPropertyReference(
     SuperPropertyReference* expr) {
   PROCESS_EXPRESSION(expr);
-  RECURSE_EXPRESSION(Visit(expr->home_object()));
 }
 
 template <class Subclass>
