@@ -10,6 +10,10 @@
 
 #include <stdlib.h>
 
+#include <memory>
+
+#include "src/base/platform/wrappers.h"
+
 namespace v8 {
 namespace base {
 
@@ -19,7 +23,7 @@ namespace base {
 // std::unique_ptr<int, base::FreeDeleter> foo_ptr(
 //     static_cast<int*>(malloc(sizeof(int))));
 struct FreeDeleter {
-  inline void operator()(void* ptr) const { free(ptr); }
+  inline void operator()(void* ptr) const { base::Free(ptr); }
 };
 
 }  // namespace base
