@@ -5,6 +5,7 @@
 #ifndef V8_BASE_LOGGING_H_
 #define V8_BASE_LOGGING_H_
 
+#include <cstdint>
 #include <cstring>
 #include <sstream>
 #include <string>
@@ -12,6 +13,7 @@
 #include "src/base/base-export.h"
 #include "src/base/build_config.h"
 #include "src/base/compiler-specific.h"
+#include "src/base/immediate-crash.h"
 #include "src/base/template-utils.h"
 
 V8_BASE_EXPORT V8_NOINLINE void V8_Dcheck(const char* file, int line,
@@ -24,28 +26,25 @@ V8_BASE_EXPORT V8_NOINLINE void V8_Dcheck(const char* file, int line,
     void V8_Fatal(const char* file, int line, const char* format, ...);
 #define FATAL(...) V8_Fatal(__FILE__, __LINE__, __VA_ARGS__)
 
-#elif !defined(OFFICIAL_BUILD)
+#else
+[[noreturn]] PRINTF_FORMAT(1, 2) V8_BASE_EXPORT V8_NOINLINE
+    void V8_Fatal(const char* format, ...);
+#if !defined(OFFICIAL_BUILD)
 // In non-official release, include full error message, but drop file & line
 // numbers. It saves binary size to drop the |file| & |line| as opposed to just
 // passing in "", 0 for them.
-[[noreturn]] PRINTF_FORMAT(1, 2) V8_BASE_EXPORT V8_NOINLINE
-    void V8_Fatal(const char* format, ...);
 #define FATAL(...) V8_Fatal(__VA_ARGS__)
 #else
-// In official builds, include only messages that contain parameters because
-// single-message errors can always be derived from stack traces.
-[[noreturn]] V8_BASE_EXPORT V8_NOINLINE void V8_FatalNoContext();
-[[noreturn]] PRINTF_FORMAT(1, 2) V8_BASE_EXPORT V8_NOINLINE
-    void V8_Fatal(const char* format, ...);
-// FATAL(msg) -> V8_FatalNoContext()
-// FATAL(msg, ...) -> V8_Fatal()
+// FATAL(msg) -> IMMEDIATE_CRASH()
+// FATAL(msg, ...) -> V8_Fatal(msg, ...)
 #define FATAL_HELPER(_7, _6, _5, _4, _3, _2, _1, _0, ...) _0
-#define FATAL_DISCARD_ARG(arg) V8_FatalNoContext()
+#define FATAL_DISCARD_ARG(arg) IMMEDIATE_CRASH()
 #define FATAL(...)                                                            \
   FATAL_HELPER(__VA_ARGS__, V8_Fatal, V8_Fatal, V8_Fatal, V8_Fatal, V8_Fatal, \
-               V8_Fatal, V8_Fatal, FATAL_DISCARD_ARG)                         \
+               V8_Fatal, FATAL_DISCARD_ARG)                                   \
   (__VA_ARGS__)
-#endif
+#endif  // !defined(OFFICIAL_BUILD)
+#endif  // DEBUG
 
 #define UNIMPLEMENTED() FATAL("unimplemented code")
 #define UNREACHABLE() FATAL("unreachable code")
@@ -53,6 +52,8 @@ V8_BASE_EXPORT V8_NOINLINE void V8_Dcheck(const char* file, int line,
 namespace v8 {
 namespace base {
 
+class CheckMessageStream : public std::ostringstream {};
+
 // Overwrite the default function that prints a stack trace.
 V8_BASE_EXPORT void SetPrintStackTrace(void (*print_stack_trace_)());
 
@@ -134,13 +135,37 @@ V8_BASE_EXPORT void SetDcheckFunction(void (*dcheck_Function)(const char*, int,
 
 #endif
 
+namespace detail {
+template <typename... Ts>
+std::string PrintToString(Ts&&... ts) {
+  CheckMessageStream oss;
+  int unused_results[]{((oss << std::forward<Ts>(ts)), 0)...};
+  (void)unused_results;  // Avoid "unused variable" warning.
+  return oss.str();
+}
+
+template <typename T>
+auto GetUnderlyingEnumTypeForPrinting(T val) {
+  using underlying_t = typename std::underlying_type<T>::type;
+  // For single-byte enums, return a 16-bit integer to avoid printing the value
+  // as a character.
+  using int_t = typename std::conditional_t<
+      sizeof(underlying_t) != 1, underlying_t,
+      std::conditional_t<std::is_signed<underlying_t>::value, int16_t,
+                         uint16_t> >;
+  return static_cast<int_t>(static_cast<underlying_t>(val));
+}
+}  // namespace detail
+
 // Define PrintCheckOperand<T> for each T which defines operator<< for ostream.
 template <typename T>
 typename std::enable_if<
     !std::is_function<typename std::remove_pointer<T>::type>::value &&
-    has_output_operator<T>::value>::type
-PrintCheckOperand(std::ostream& os, T val) {
-  os << std::forward<T>(val);
+        !std::is_enum<T>::value &&
+        has_output_operator<T, CheckMessageStream>::value,
+    std::string>::type
+PrintCheckOperand(T val) {
+  return detail::PrintToString(std::forward<T>(val));
 }
 
 // Provide an overload for functions and function pointers. Function pointers
@@ -150,43 +175,59 @@ PrintCheckOperand(std::ostream& os, T val) {
 // pointers, so this is a no-op for MSVC.)
 template <typename T>
 typename std::enable_if<
-    std::is_function<typename std::remove_pointer<T>::type>::value>::type
-PrintCheckOperand(std::ostream& os, T val) {
-  os << reinterpret_cast<const void*>(val);
+    std::is_function<typename std::remove_pointer<T>::type>::value,
+    std::string>::type
+PrintCheckOperand(T val) {
+  return PrintCheckOperand(reinterpret_cast<const void*>(val));
 }
 
-// Define PrintCheckOperand<T> for enums which have no operator<<.
+// Define PrintCheckOperand<T> for enums with an output operator.
 template <typename T>
 typename std::enable_if<std::is_enum<T>::value &&
-                        !has_output_operator<T>::value>::type
-PrintCheckOperand(std::ostream& os, T val) {
-  using underlying_t = typename std::underlying_type<T>::type;
-  // 8-bit types are not printed as number, so extend them to 16 bit.
-  using int_t = typename std::conditional<
-      std::is_same<underlying_t, uint8_t>::value, uint16_t,
-      typename std::conditional<std::is_same<underlying_t, int8_t>::value,
-                                int16_t, underlying_t>::type>::type;
-  PrintCheckOperand(os, static_cast<int_t>(static_cast<underlying_t>(val)));
+                            has_output_operator<T, CheckMessageStream>::value,
+                        std::string>::type
+PrintCheckOperand(T val) {
+  std::string val_str = detail::PrintToString(val);
+  std::string int_str =
+      detail::PrintToString(detail::GetUnderlyingEnumTypeForPrinting(val));
+  // Printing the original enum might have printed a single non-printable
+  // character. Ignore it in that case. Also ignore if it printed the same as
+  // the integral representation.
+  // TODO(clemensb): Can we somehow statically find out if the output operator
+  // is the default one, printing the integral value?
+  if ((val_str.length() == 1 && !std::isprint(val_str[0])) ||
+      val_str == int_str) {
+    return int_str;
+  }
+  return detail::PrintToString(val_str, " (", int_str, ")");
+}
+
+// Define PrintCheckOperand<T> for enums without an output operator.
+template <typename T>
+typename std::enable_if<std::is_enum<T>::value &&
+                            !has_output_operator<T, CheckMessageStream>::value,
+                        std::string>::type
+PrintCheckOperand(T val) {
+  return detail::PrintToString(detail::GetUnderlyingEnumTypeForPrinting(val));
 }
 
 // Define default PrintCheckOperand<T> for non-printable types.
 template <typename T>
-typename std::enable_if<!has_output_operator<T>::value &&
-                        !std::is_enum<T>::value>::type
-PrintCheckOperand(std::ostream& os, T val) {
-  os << "<unprintable>";
+typename std::enable_if<!has_output_operator<T, CheckMessageStream>::value &&
+                            !std::is_enum<T>::value,
+                        std::string>::type
+PrintCheckOperand(T val) {
+  return "<unprintable>";
 }
 
 // Define specializations for character types, defined in logging.cc.
-#define DEFINE_PRINT_CHECK_OPERAND_CHAR(type)                              \
-  template <>                                                              \
-  V8_BASE_EXPORT void PrintCheckOperand<type>(std::ostream & os, type ch); \
-  template <>                                                              \
-  V8_BASE_EXPORT void PrintCheckOperand<type*>(std::ostream & os,          \
-                                               type * cstr);               \
-  template <>                                                              \
-  V8_BASE_EXPORT void PrintCheckOperand<const type*>(std::ostream & os,    \
-                                                     const type* cstr);
+#define DEFINE_PRINT_CHECK_OPERAND_CHAR(type)                       \
+  template <>                                                       \
+  V8_BASE_EXPORT std::string PrintCheckOperand<type>(type ch);      \
+  template <>                                                       \
+  V8_BASE_EXPORT std::string PrintCheckOperand<type*>(type * cstr); \
+  template <>                                                       \
+  V8_BASE_EXPORT std::string PrintCheckOperand<const type*>(const type* cstr);
 
 DEFINE_PRINT_CHECK_OPERAND_CHAR(char)
 DEFINE_PRINT_CHECK_OPERAND_CHAR(signed char)
@@ -198,13 +239,18 @@ DEFINE_PRINT_CHECK_OPERAND_CHAR(unsigned char)
 // be out of line, while the "Impl" code should be inline. Caller
 // takes ownership of the returned string.
 template <typename Lhs, typename Rhs>
-std::string* MakeCheckOpString(Lhs lhs, Rhs rhs, char const* msg) {
-  std::ostringstream ss;
-  ss << msg << " (";
-  PrintCheckOperand<Lhs>(ss, lhs);
-  ss << " vs. ";
-  PrintCheckOperand<Rhs>(ss, rhs);
-  ss << ")";
+V8_NOINLINE std::string* MakeCheckOpString(Lhs lhs, Rhs rhs, char const* msg) {
+  std::string lhs_str = PrintCheckOperand<Lhs>(lhs);
+  std::string rhs_str = PrintCheckOperand<Rhs>(rhs);
+  CheckMessageStream ss;
+  ss << msg;
+  constexpr size_t kMaxInlineLength = 50;
+  if (lhs_str.size() <= kMaxInlineLength &&
+      rhs_str.size() <= kMaxInlineLength) {
+    ss << " (" << lhs_str << " vs. " << rhs_str << ")";
+  } else {
+    ss << "\n   " << lhs_str << "\n vs.\n   " << rhs_str << "\n";
+  }
   return new std::string(ss.str());
 }
 
@@ -213,8 +259,7 @@ std::string* MakeCheckOpString(Lhs lhs, Rhs rhs, char const* msg) {
 #define EXPLICIT_CHECK_OP_INSTANTIATION(type)                                \
   extern template V8_BASE_EXPORT std::string* MakeCheckOpString<type, type>( \
       type, type, char const*);                                              \
-  extern template V8_BASE_EXPORT void PrintCheckOperand<type>(std::ostream&, \
-                                                              type);
+  extern template V8_BASE_EXPORT std::string PrintCheckOperand<type>(type);
 
 EXPLICIT_CHECK_OP_INSTANTIATION(int)
 EXPLICIT_CHECK_OP_INSTANTIATION(long)       // NOLINT(runtime/int)
@@ -268,11 +313,12 @@ struct is_unsigned_vs_signed : public is_signed_vs_unsigned<Rhs, Lhs> {};
 #define MAKE_UNSIGNED(Type, value)         \
   static_cast<typename std::make_unsigned< \
       typename comparison_underlying_type<Type>::type>::type>(value)
-#define DEFINE_SIGNED_MISMATCH_COMP(CHECK, NAME, IMPL)                  \
-  template <typename Lhs, typename Rhs>                                 \
-  V8_INLINE typename std::enable_if<CHECK<Lhs, Rhs>::value, bool>::type \
-      Cmp##NAME##Impl(Lhs lhs, Rhs rhs) {                               \
-    return IMPL;                                                        \
+#define DEFINE_SIGNED_MISMATCH_COMP(CHECK, NAME, IMPL)            \
+  template <typename Lhs, typename Rhs>                           \
+  V8_INLINE constexpr                                             \
+      typename std::enable_if<CHECK<Lhs, Rhs>::value, bool>::type \
+          Cmp##NAME##Impl(Lhs lhs, Rhs rhs) {                     \
+    return IMPL;                                                  \
   }
 DEFINE_SIGNED_MISMATCH_COMP(is_signed_vs_unsigned, EQ,
                             lhs >= 0 && MAKE_UNSIGNED(Lhs, lhs) ==
@@ -302,29 +348,24 @@ DEFINE_SIGNED_MISMATCH_COMP(is_unsigned_vs_signed, GE, CmpLEImpl(rhs, lhs))
 // not integral or their signedness matches (i.e. whenever no specialization is
 // required, see above). Otherwise it is disabled by the enable_if construct,
 // and the compiler will pick a specialization from above.
-#define DEFINE_CHECK_OP_IMPL(NAME, op)                                         \
-  template <typename Lhs, typename Rhs>                                        \
-  V8_INLINE                                                                    \
-      typename std::enable_if<!is_signed_vs_unsigned<Lhs, Rhs>::value &&       \
-                                  !is_unsigned_vs_signed<Lhs, Rhs>::value,     \
-                              bool>::type Cmp##NAME##Impl(Lhs lhs, Rhs rhs) {  \
-    return lhs op rhs;                                                         \
-  }                                                                            \
-  template <typename Lhs, typename Rhs>                                        \
-  V8_INLINE std::string* Check##NAME##Impl(Lhs lhs, Rhs rhs,                   \
-                                           char const* msg) {                  \
-    using LhsPassT = typename pass_value_or_ref<Lhs>::type;                    \
-    using RhsPassT = typename pass_value_or_ref<Rhs>::type;                    \
-    bool cmp = Cmp##NAME##Impl<LhsPassT, RhsPassT>(lhs, rhs);                  \
-    return V8_LIKELY(cmp)                                                      \
-               ? nullptr                                                       \
-               : MakeCheckOpString<LhsPassT, RhsPassT>(lhs, rhs, msg);         \
-  }                                                                            \
-  extern template V8_BASE_EXPORT std::string* Check##NAME##Impl<float, float>( \
-      float lhs, float rhs, char const* msg);                                  \
-  extern template V8_BASE_EXPORT std::string*                                  \
-      Check##NAME##Impl<double, double>(double lhs, double rhs,                \
-                                        char const* msg);
+#define DEFINE_CHECK_OP_IMPL(NAME, op)                                        \
+  template <typename Lhs, typename Rhs>                                       \
+  V8_INLINE constexpr                                                         \
+      typename std::enable_if<!is_signed_vs_unsigned<Lhs, Rhs>::value &&      \
+                                  !is_unsigned_vs_signed<Lhs, Rhs>::value,    \
+                              bool>::type Cmp##NAME##Impl(Lhs lhs, Rhs rhs) { \
+    return lhs op rhs;                                                        \
+  }                                                                           \
+  template <typename Lhs, typename Rhs>                                       \
+  V8_INLINE constexpr std::string* Check##NAME##Impl(Lhs lhs, Rhs rhs,        \
+                                                     char const* msg) {       \
+    using LhsPassT = typename pass_value_or_ref<Lhs>::type;                   \
+    using RhsPassT = typename pass_value_or_ref<Rhs>::type;                   \
+    bool cmp = Cmp##NAME##Impl<LhsPassT, RhsPassT>(lhs, rhs);                 \
+    return V8_LIKELY(cmp)                                                     \
+               ? nullptr                                                      \
+               : MakeCheckOpString<LhsPassT, RhsPassT>(lhs, rhs, msg);        \
+  }
 DEFINE_CHECK_OP_IMPL(EQ, ==)
 DEFINE_CHECK_OP_IMPL(NE, !=)
 DEFINE_CHECK_OP_IMPL(LE, <=)
