@@ -5,8 +5,10 @@
 #ifndef V8_HEAP_GC_TRACER_H_
 #define V8_HEAP_GC_TRACER_H_
 
+#include "include/v8-metrics.h"
 #include "src/base/compiler-specific.h"
-#include "src/base/platform/platform.h"
+#include "src/base/macros.h"
+#include "src/base/optional.h"
 #include "src/base/ring-buffer.h"
 #include "src/common/globals.h"
 #include "src/heap/heap.h"
@@ -25,47 +27,48 @@ inline BytesAndDuration MakeBytesAndDuration(uint64_t bytes, double duration) {
 
 enum ScavengeSpeedMode { kForAllObjects, kForSurvivedObjects };
 
-#define TRACE_GC(tracer, scope_id)                             \
-  GCTracer::Scope::ScopeId gc_tracer_scope_id(scope_id);       \
-  GCTracer::Scope gc_tracer_scope(tracer, gc_tracer_scope_id); \
-  TRACE_EVENT0(TRACE_DISABLED_BY_DEFAULT("v8.gc"),             \
-               GCTracer::Scope::Name(gc_tracer_scope_id))
+#define TRACE_GC_CATEGORIES \
+  "devtools.timeline," TRACE_DISABLED_BY_DEFAULT("v8.gc")
 
-#define TRACE_BACKGROUND_GC(tracer, scope_id)                                 \
-  WorkerThreadRuntimeCallStatsScope runtime_call_stats_scope(                 \
-      tracer->worker_thread_runtime_call_stats());                            \
-  GCTracer::BackgroundScope background_scope(tracer, scope_id,                \
-                                             runtime_call_stats_scope.Get()); \
-  TRACE_EVENT0(TRACE_DISABLED_BY_DEFAULT("v8.gc"),                            \
-               GCTracer::BackgroundScope::Name(scope_id))
+#define TRACE_GC(tracer, scope_id)                                    \
+  GCTracer::Scope UNIQUE_IDENTIFIER(gc_tracer_scope)(                 \
+      tracer, GCTracer::Scope::ScopeId(scope_id), ThreadKind::kMain); \
+  TRACE_EVENT0(TRACE_GC_CATEGORIES,                                   \
+               GCTracer::Scope::Name(GCTracer::Scope::ScopeId(scope_id)))
+
+#define TRACE_GC1(tracer, scope_id, thread_kind)                \
+  GCTracer::Scope UNIQUE_IDENTIFIER(gc_tracer_scope)(           \
+      tracer, GCTracer::Scope::ScopeId(scope_id), thread_kind); \
+  TRACE_EVENT0(TRACE_GC_CATEGORIES,                             \
+               GCTracer::Scope::Name(GCTracer::Scope::ScopeId(scope_id)))
+
+#define TRACE_GC_EPOCH(tracer, scope_id, thread_kind)                     \
+  GCTracer::Scope UNIQUE_IDENTIFIER(gc_tracer_scope)(                     \
+      tracer, GCTracer::Scope::ScopeId(scope_id), thread_kind);           \
+  TRACE_EVENT1(TRACE_GC_CATEGORIES,                                       \
+               GCTracer::Scope::Name(GCTracer::Scope::ScopeId(scope_id)), \
+               "epoch", tracer->CurrentEpoch(scope_id))
+
+using CollectionEpoch = uint32_t;
 
 // GCTracer collects and prints ONE line after each garbage collector
 // invocation IFF --trace_gc is used.
 class V8_EXPORT_PRIVATE GCTracer {
  public:
+  GCTracer(const GCTracer&) = delete;
+  GCTracer& operator=(const GCTracer&) = delete;
+
   struct IncrementalMarkingInfos {
-    IncrementalMarkingInfos() : duration(0), longest_step(0), steps(0) {}
+    V8_INLINE IncrementalMarkingInfos();
+    V8_INLINE void Update(double delta);
+    V8_INLINE void ResetCurrentCycle();
 
-    void Update(double duration) {
-      steps++;
-      this->duration += duration;
-      if (duration > longest_step) {
-        longest_step = duration;
-      }
-    }
-
-    void ResetCurrentCycle() {
-      duration = 0;
-      longest_step = 0;
-      steps = 0;
-    }
-
-    double duration;
-    double longest_step;
+    double duration;      // in ms
+    double longest_step;  // in ms
     int steps;
   };
 
-  class Scope {
+  class V8_EXPORT_PRIVATE V8_NODISCARD Scope {
    public:
     enum ScopeId {
 #define DEFINE_SCOPE(scope) scope,
@@ -78,7 +81,7 @@ class V8_EXPORT_PRIVATE GCTracer {
       FIRST_SCOPE = MC_INCREMENTAL,
       NUMBER_OF_INCREMENTAL_SCOPES =
           LAST_INCREMENTAL_SCOPE - FIRST_INCREMENTAL_SCOPE + 1,
-      FIRST_GENERAL_BACKGROUND_SCOPE = BACKGROUND_ARRAY_BUFFER_FREE,
+      FIRST_GENERAL_BACKGROUND_SCOPE = BACKGROUND_YOUNG_ARRAY_BUFFER_SWEEP,
       LAST_GENERAL_BACKGROUND_SCOPE = BACKGROUND_UNMAPPER,
       FIRST_MC_BACKGROUND_SCOPE = MC_BACKGROUND_EVACUATE_COPY,
       LAST_MC_BACKGROUND_SCOPE = MC_BACKGROUND_SWEEPING,
@@ -86,50 +89,32 @@ class V8_EXPORT_PRIVATE GCTracer {
       LAST_TOP_MC_SCOPE = MC_SWEEP,
       FIRST_MINOR_GC_BACKGROUND_SCOPE = MINOR_MC_BACKGROUND_EVACUATE_COPY,
       LAST_MINOR_GC_BACKGROUND_SCOPE = SCAVENGER_BACKGROUND_SCAVENGE_PARALLEL,
-      FIRST_BACKGROUND_SCOPE = FIRST_GENERAL_BACKGROUND_SCOPE
+      FIRST_BACKGROUND_SCOPE = FIRST_GENERAL_BACKGROUND_SCOPE,
+      LAST_BACKGROUND_SCOPE = LAST_MINOR_GC_BACKGROUND_SCOPE
     };
 
-    Scope(GCTracer* tracer, ScopeId scope);
-    ~Scope();
+    V8_INLINE Scope(GCTracer* tracer, ScopeId scope, ThreadKind thread_kind);
+    V8_INLINE ~Scope();
+    Scope(const Scope&) = delete;
+    Scope& operator=(const Scope&) = delete;
     static const char* Name(ScopeId id);
+    static bool NeedsYoungEpoch(ScopeId id);
+    V8_INLINE static constexpr int IncrementalOffset(ScopeId id);
 
    private:
-    GCTracer* tracer_;
-    ScopeId scope_;
-    double start_time_;
+#if DEBUG
+    void AssertMainThread();
+#endif  // DEBUG
+
+    GCTracer* const tracer_;
+    const ScopeId scope_;
+    const ThreadKind thread_kind_;
+    const double start_time_;
+#ifdef V8_RUNTIME_CALL_STATS
     RuntimeCallTimer timer_;
     RuntimeCallStats* runtime_stats_ = nullptr;
-
-    DISALLOW_COPY_AND_ASSIGN(Scope);
-  };
-
-  class V8_EXPORT_PRIVATE BackgroundScope {
-   public:
-    enum ScopeId {
-#define DEFINE_SCOPE(scope) scope,
-      TRACER_BACKGROUND_SCOPES(DEFINE_SCOPE)
-#undef DEFINE_SCOPE
-          NUMBER_OF_SCOPES,
-      FIRST_GENERAL_BACKGROUND_SCOPE = BACKGROUND_ARRAY_BUFFER_FREE,
-      LAST_GENERAL_BACKGROUND_SCOPE = BACKGROUND_UNMAPPER,
-      FIRST_MC_BACKGROUND_SCOPE = MC_BACKGROUND_EVACUATE_COPY,
-      LAST_MC_BACKGROUND_SCOPE = MC_BACKGROUND_SWEEPING,
-      FIRST_MINOR_GC_BACKGROUND_SCOPE = MINOR_MC_BACKGROUND_EVACUATE_COPY,
-      LAST_MINOR_GC_BACKGROUND_SCOPE = SCAVENGER_BACKGROUND_SCAVENGE_PARALLEL
-    };
-    BackgroundScope(GCTracer* tracer, ScopeId scope,
-                    RuntimeCallStats* runtime_stats);
-    ~BackgroundScope();
-
-    static const char* Name(ScopeId id);
-
-   private:
-    GCTracer* tracer_;
-    ScopeId scope_;
-    double start_time_;
-    RuntimeCallTimer timer_;
-    RuntimeCallStats* runtime_stats_;
-    DISALLOW_COPY_AND_ASSIGN(BackgroundScope);
+    base::Optional<WorkerThreadRuntimeCallStatsScope> runtime_call_stats_scope_;
+#endif  // defined(V8_RUNTIME_CALL_STATS)
   };
 
   class Event {
@@ -142,15 +127,28 @@ class V8_EXPORT_PRIVATE GCTracer {
       START = 4
     };
 
-    Event(Type type, GarbageCollectionReason gc_reason,
+    // Returns true if the event corresponds to a young generation GC.
+    V8_INLINE static constexpr bool IsYoungGenerationEvent(Type type);
+
+    // The state diagram for a GC cycle:
+    //   (NOT_RUNNING) -----(StartCycle)----->
+    //   MARKING       --(StartAtomicPause)-->
+    //   ATOMIC        ---(StopAtomicPause)-->
+    //   SWEEPING      ------(StopCycle)-----> NOT_RUNNING
+    enum class State { NOT_RUNNING, MARKING, ATOMIC, SWEEPING };
+
+    Event(Type type, State state, GarbageCollectionReason gc_reason,
           const char* collector_reason);
 
     // Returns a string describing the event type.
     const char* TypeName(bool short_name) const;
 
-    // Type of event
+    // Type of the event.
     Type type;
 
+    // State of the cycle corresponding to the event.
+    State state;
+
     GarbageCollectionReason gc_reason;
     const char* collector_reason;
 
@@ -192,15 +190,34 @@ class V8_EXPORT_PRIVATE GCTracer {
     // Bytes marked incrementally for INCREMENTAL_MARK_COMPACTOR
     size_t incremental_marking_bytes;
 
-    // Duration of incremental marking steps for INCREMENTAL_MARK_COMPACTOR.
+    // Duration (in ms) of incremental marking steps for
+    // INCREMENTAL_MARK_COMPACTOR.
     double incremental_marking_duration;
 
-    // Amounts of time spent in different scopes during GC.
+    // Amounts of time (in ms) spent in different scopes during GC.
     double scopes[Scope::NUMBER_OF_SCOPES];
 
     // Holds details for incremental marking scopes.
     IncrementalMarkingInfos
-        incremental_marking_scopes[Scope::NUMBER_OF_INCREMENTAL_SCOPES];
+        incremental_scopes[Scope::NUMBER_OF_INCREMENTAL_SCOPES];
+  };
+
+  class RecordGCPhasesInfo {
+   public:
+    RecordGCPhasesInfo(Heap* heap, GarbageCollector collector);
+
+    enum class Mode { None, Scavenger, Finalize };
+
+    Mode mode;
+
+    // The timer used for a given GC type:
+    // - GCScavenger: young generation GC
+    // - GCCompactor: full GC
+    // - GCFinalizeMC: finalization of incremental full GC
+    // - GCFinalizeMCReduceMemory: finalization of incremental full GC with
+    //   memory reduction.
+    TimedHistogram* type_timer;
+    TimedHistogram* type_priority_timer;
   };
 
   static const int kThroughputTimeFrameMs = 5000;
@@ -209,24 +226,60 @@ class V8_EXPORT_PRIVATE GCTracer {
   static double CombineSpeedsInBytesPerMillisecond(double default_speed,
                                                    double optional_speed);
 
-  static RuntimeCallCounterId RCSCounterFromScope(Scope::ScopeId id);
-  static RuntimeCallCounterId RCSCounterFromBackgroundScope(
-      BackgroundScope::ScopeId id);
+#ifdef V8_RUNTIME_CALL_STATS
+  V8_INLINE static RuntimeCallCounterId RCSCounterFromScope(Scope::ScopeId id);
+#endif  // defined(V8_RUNTIME_CALL_STATS)
 
   explicit GCTracer(Heap* heap);
 
-  // Start collecting data.
-  void Start(GarbageCollector collector, GarbageCollectionReason gc_reason,
-             const char* collector_reason);
+  V8_INLINE CollectionEpoch CurrentEpoch(Scope::ScopeId id) const;
 
-  // Stop collecting data and print results.
-  void Stop(GarbageCollector collector);
+  // Start and stop an observable pause.
+  void StartObservablePause();
+  void StopObservablePause();
+
+  // Update the current event if it precedes the start of the observable pause.
+  void UpdateCurrentEvent(GarbageCollectionReason gc_reason,
+                          const char* collector_reason);
+
+  void UpdateStatistics(GarbageCollector collector);
+  void FinalizeCurrentEvent();
+
+  enum class MarkingType { kAtomic, kIncremental };
+
+  // Start and stop a GC cycle (collecting data and reporting results).
+  void StartCycle(GarbageCollector collector, GarbageCollectionReason gc_reason,
+                  const char* collector_reason, MarkingType marking);
+  void StopYoungCycleIfNeeded();
+  void StopFullCycleIfNeeded();
+
+  // Start and stop a cycle's atomic pause.
+  void StartAtomicPause();
+  void StopAtomicPause();
+
+  void StartInSafepoint();
+  void StopInSafepoint();
 
   void NotifySweepingCompleted();
+  void NotifyFullCppGCCompleted();
+
+  void NotifyYoungCppGCRunning();
+  void NotifyYoungCppGCCompleted();
 
   void NotifyYoungGenerationHandling(
       YoungGenerationHandling young_generation_handling);
 
+#ifdef DEBUG
+  V8_INLINE bool IsInObservablePause() const;
+
+  // Checks if the current event is consistent with a collector.
+  V8_INLINE bool IsConsistentWithCollector(GarbageCollector collector) const;
+
+  // Checks if the current event corresponds to a full GC cycle whose sweeping
+  // has not finalized yet.
+  V8_INLINE bool IsSweepingInProgress() const;
+#endif
+
   // Sample and accumulate bytes allocated since the last GC.
   void SampleAllocation(double current_ms, size_t new_space_counter_bytes,
                         size_t old_generation_counter_bytes,
@@ -235,8 +288,6 @@ class V8_EXPORT_PRIVATE GCTracer {
   // Log the accumulated new space allocation bytes.
   void AddAllocation(double current_ms);
 
-  void AddContextDisposalTime(double time);
-
   void AddCompactionEvent(double duration, size_t live_bytes_compacted);
 
   void AddSurvivalRatio(double survival_ratio);
@@ -244,6 +295,9 @@ class V8_EXPORT_PRIVATE GCTracer {
   // Log an incremental marking step.
   void AddIncrementalMarkingStep(double duration, size_t bytes);
 
+  // Log an incremental marking step.
+  void AddIncrementalSweepingStep(double duration);
+
   // Compute the average incremental marking speed in bytes/millisecond.
   // Returns a conservative value if no events have been recorded.
   double IncrementalMarkingSpeedInBytesPerMillisecond() const;
@@ -311,12 +365,6 @@ class V8_EXPORT_PRIVATE GCTracer {
   // Returns 0 if no allocation events have been recorded.
   double CurrentEmbedderAllocationThroughputInBytesPerMillisecond() const;
 
-  // Computes the context disposal rate in milliseconds. It takes the time
-  // frame of the first recorded context disposal to the current time and
-  // divides it by the number of recorded events.
-  // Returns 0 if no events have been recorded.
-  double ContextDisposalRateInMilliseconds() const;
-
   // Computes the average survival ratio based on the last recorded survival
   // events.
   // Returns 0 if no events have been recorded.
@@ -335,25 +383,20 @@ class V8_EXPORT_PRIVATE GCTracer {
   double AverageMarkCompactMutatorUtilization() const;
   double CurrentMarkCompactMutatorUtilization() const;
 
-  V8_INLINE void AddScopeSample(Scope::ScopeId scope, double duration) {
-    DCHECK(scope < Scope::NUMBER_OF_SCOPES);
-    if (scope >= Scope::FIRST_INCREMENTAL_SCOPE &&
-        scope <= Scope::LAST_INCREMENTAL_SCOPE) {
-      incremental_marking_scopes_[scope - Scope::FIRST_INCREMENTAL_SCOPE]
-          .Update(duration);
-    } else {
-      current_.scopes[scope] += duration;
-    }
-  }
+  V8_INLINE void AddScopeSample(Scope::ScopeId id, double duration);
 
-  void AddBackgroundScopeSample(BackgroundScope::ScopeId scope,
-                                double duration);
-
-  void RecordGCPhasesHistograms(TimedHistogram* gc_timer);
+  void RecordGCPhasesHistograms(RecordGCPhasesInfo::Mode mode);
 
   void RecordEmbedderSpeed(size_t bytes, double duration);
 
-  WorkerThreadRuntimeCallStats* worker_thread_runtime_call_stats();
+  // Returns the average time between scheduling and invocation of an
+  // incremental marking task.
+  double AverageTimeToIncrementalMarkingTask() const;
+  void RecordTimeToIncrementalMarkingTask(double time_to_task);
+
+#ifdef V8_RUNTIME_CALL_STATS
+  V8_INLINE WorkerThreadRuntimeCallStats* worker_thread_runtime_call_stats();
+#endif  // defined(V8_RUNTIME_CALL_STATS)
 
  private:
   FRIEND_TEST(GCTracer, AverageSpeed);
@@ -371,7 +414,6 @@ class V8_EXPORT_PRIVATE GCTracer {
   FRIEND_TEST(GCTracerTest, IncrementalScope);
   FRIEND_TEST(GCTracerTest, IncrementalMarkingSpeed);
   FRIEND_TEST(GCTracerTest, MutatorUtilization);
-  FRIEND_TEST(GCTracerTest, RecordGCSumHistograms);
   FRIEND_TEST(GCTracerTest, RecordMarkCompactHistograms);
   FRIEND_TEST(GCTracerTest, RecordScavengerHistograms);
 
@@ -379,6 +421,19 @@ class V8_EXPORT_PRIVATE GCTracer {
     double total_duration_ms;
   };
 
+  void StopCycle(GarbageCollector collector);
+
+  // Statistics for incremental and background scopes are kept out of the
+  // current event and only copied there by FinalizeCurrentEvent, at StopCycle.
+  // This method can be used to access scopes correctly, before this happens.
+  // Note: when accessing a background scope via this method, the caller is
+  // responsible for avoiding data races, e.g., by acquiring
+  // background_counter_mutex_.
+  V8_INLINE constexpr double current_scope(Scope::ScopeId id) const;
+
+  V8_INLINE constexpr const IncrementalMarkingInfos& incremental_scope(
+      Scope::ScopeId id) const;
+
   // Returns the average speed of the events in the buffer.
   // If the buffer is empty, the result is 0.
   // Otherwise, the result is between 1 byte/ms and 1 GB/ms.
@@ -392,10 +447,12 @@ class V8_EXPORT_PRIVATE GCTracer {
   void RecordMutatorUtilization(double mark_compactor_end_time,
                                 double mark_compactor_duration);
 
-  // Overall time spent in mark compact within a given GC cycle. Exact
-  // accounting of events within a GC is not necessary which is why the
-  // recording takes place at the end of the atomic pause.
-  void RecordGCSumCounters(double atomic_pause_duration);
+  // Update counters for an entire full GC cycle. Exact accounting of events
+  // within a GC is not necessary which is why the recording takes place at the
+  // end of the atomic pause.
+  void RecordGCSumCounters();
+
+  V8_INLINE double MonotonicallyIncreasingTimeInMs();
 
   // Print one detailed trace line in name=value format.
   // TODO(ernstm): Move to Heap.
@@ -409,21 +466,16 @@ class V8_EXPORT_PRIVATE GCTracer {
   // it can be included in later crash dumps.
   void PRINTF_FORMAT(2, 3) Output(const char* format, ...) const;
 
-  double TotalExternalTime() const {
-    return current_.scopes[Scope::HEAP_EXTERNAL_WEAK_GLOBAL_HANDLES] +
-           current_.scopes[Scope::HEAP_EXTERNAL_EPILOGUE] +
-           current_.scopes[Scope::HEAP_EXTERNAL_PROLOGUE] +
-           current_.scopes[Scope::MC_INCREMENTAL_EXTERNAL_EPILOGUE] +
-           current_.scopes[Scope::MC_INCREMENTAL_EXTERNAL_PROLOGUE];
-  }
-
-  void FetchBackgroundCounters(int first_global_scope, int last_global_scope,
-                               int first_background_scope,
-                               int last_background_scope);
+  void FetchBackgroundCounters(int first_scope, int last_scope);
   void FetchBackgroundMinorGCCounters();
   void FetchBackgroundMarkCompactCounters();
   void FetchBackgroundGeneralCounters();
 
+  void ReportFullCycleToRecorder();
+  void ReportIncrementalMarkingStepToRecorder(double v8_duration);
+  void ReportIncrementalSweepingStepToRecorder(double v8_duration);
+  void ReportYoungCycleToRecorder();
+
   // Pointer to the heap that owns this tracer.
   Heap* heap_;
 
@@ -434,25 +486,34 @@ class V8_EXPORT_PRIVATE GCTracer {
   // Previous tracer event.
   Event previous_;
 
+  // The starting time of the observable pause or 0.0 if we're not inside it.
+  double start_of_observable_pause_ = 0.0;
+
+  // We need two epochs, since there can be scavenges during incremental
+  // marking.
+  CollectionEpoch epoch_young_ = 0;
+  CollectionEpoch epoch_full_ = 0;
+
   // Size of incremental marking steps (in bytes) accumulated since the end of
   // the last mark compact GC.
   size_t incremental_marking_bytes_;
 
-  // Duration of incremental marking steps since the end of the last mark-
-  // compact event.
+  // Duration (in ms) of incremental marking steps since the end of the last
+  // mark-compact event.
   double incremental_marking_duration_;
 
   double incremental_marking_start_time_;
 
   double recorded_incremental_marking_speed_;
 
+  double average_time_to_incremental_marking_task_ = 0.0;
+
   double recorded_embedder_speed_ = 0.0;
 
   // Incremental scopes carry more information than just the duration. The infos
   // here are merged back upon starting/stopping the GC tracer.
   IncrementalMarkingInfos
-      incremental_marking_scopes_[Scope::NUMBER_OF_INCREMENTAL_SCOPES];
-
+      incremental_scopes_[Scope::NUMBER_OF_INCREMENTAL_SCOPES];
 
   // Timestamp and allocation counter at the last sampled allocation event.
   double allocation_time_ms_;
@@ -460,7 +521,7 @@ class V8_EXPORT_PRIVATE GCTracer {
   size_t old_generation_allocation_counter_bytes_;
   size_t embedder_allocation_counter_bytes_;
 
-  // Accumulated duration and allocated bytes since the last GC.
+  // Accumulated duration (in ms) and allocated bytes since the last GC.
   double allocation_duration_since_gc_;
   size_t new_space_allocation_in_bytes_since_gc_;
   size_t old_generation_allocation_in_bytes_since_gc_;
@@ -485,13 +546,31 @@ class V8_EXPORT_PRIVATE GCTracer {
   base::RingBuffer<BytesAndDuration> recorded_new_generation_allocations_;
   base::RingBuffer<BytesAndDuration> recorded_old_generation_allocations_;
   base::RingBuffer<BytesAndDuration> recorded_embedder_generation_allocations_;
-  base::RingBuffer<double> recorded_context_disposal_times_;
   base::RingBuffer<double> recorded_survival_ratios_;
 
-  base::Mutex background_counter_mutex_;
-  BackgroundCounter background_counter_[BackgroundScope::NUMBER_OF_SCOPES];
+  // A full GC cycle stops only when both v8 and cppgc (if available) GCs have
+  // finished sweeping.
+  bool notified_sweeping_completed_ = false;
+  bool notified_full_cppgc_completed_ = false;
+  // Similar to full GCs, a young GC cycle stops only when both v8 and cppgc GCs
+  // have finished sweeping.
+  bool notified_young_cppgc_completed_ = false;
+  // Keep track whether the young cppgc GC was scheduled (as opposed to full
+  // cycles, for young cycles cppgc is not always scheduled).
+  bool notified_young_cppgc_running_ = false;
 
-  DISALLOW_COPY_AND_ASSIGN(GCTracer);
+  // When a full GC cycle is interrupted by a young generation GC cycle, the
+  // |previous_| event is used as temporary storage for the |current_| event
+  // that corresponded to the full GC cycle, and this field is set to true.
+  bool young_gc_while_full_gc_ = false;
+
+  v8::metrics::GarbageCollectionFullMainThreadBatchedIncrementalMark
+      incremental_mark_batched_events_;
+  v8::metrics::GarbageCollectionFullMainThreadBatchedIncrementalSweep
+      incremental_sweep_batched_events_;
+
+  mutable base::Mutex background_counter_mutex_;
+  BackgroundCounter background_counter_[Scope::NUMBER_OF_SCOPES];
 };
 
 }  // namespace internal
