@@ -7,21 +7,22 @@
 
 #include "src/objects/microtask.h"
 
+#include "src/heap/heap-write-barrier-inl.h"
+#include "src/objects/contexts-inl.h"
+#include "src/objects/foreign-inl.h"
+#include "src/objects/js-objects-inl.h"
+
 // Has to be the last include (doesn't have include guards):
 #include "src/objects/object-macros.h"
 
 namespace v8 {
 namespace internal {
 
-CAST_ACCESSOR(Microtask)
-CAST_ACCESSOR(CallbackTask)
-CAST_ACCESSOR(CallableTask)
+#include "torque-generated/src/objects/microtask-tq-inl.inc"
 
-ACCESSORS(CallableTask, callable, JSReceiver, kCallableOffset)
-ACCESSORS(CallableTask, context, Context, kContextOffset)
-
-ACCESSORS(CallbackTask, callback, Foreign, kCallbackOffset)
-ACCESSORS(CallbackTask, data, Foreign, kDataOffset)
+TQ_OBJECT_CONSTRUCTORS_IMPL(Microtask)
+TQ_OBJECT_CONSTRUCTORS_IMPL(CallbackTask)
+TQ_OBJECT_CONSTRUCTORS_IMPL(CallableTask)
 
 }  // namespace internal
 }  // namespace v8
