@@ -6,12 +6,14 @@
 #define V8_PROFILER_HEAP_SNAPSHOT_GENERATOR_H_
 
 #include <deque>
+#include <memory>
 #include <unordered_map>
 #include <unordered_set>
 #include <vector>
 
 #include "include/v8-profiler.h"
 #include "src/base/platform/time.h"
+#include "src/execution/isolate.h"
 #include "src/objects/fixed-array.h"
 #include "src/objects/hash-table.h"
 #include "src/objects/heap-object.h"
@@ -22,6 +24,10 @@
 #include "src/profiler/strings-storage.h"
 #include "src/strings/string-hasher.h"
 
+#ifdef V8_ENABLE_HEAP_SNAPSHOT_VERIFY
+#include "src/heap/reference-summarizer.h"
+#endif
+
 namespace v8 {
 namespace internal {
 
@@ -37,6 +43,7 @@ class JSGlobalObject;
 class JSGlobalProxy;
 class JSPromise;
 class JSWeakCollection;
+class SafepointScope;
 
 struct SourceLocation {
   SourceLocation(int entry_index, int scriptId, int line, int col)
@@ -82,8 +89,8 @@ class HeapGraphEdge {
   V8_INLINE HeapSnapshot* snapshot() const;
   int from_index() const { return FromIndexField::decode(bit_field_); }
 
-  using TypeField = BitField<Type, 0, 3>;
-  using FromIndexField = BitField<int, 3, 29>;
+  using TypeField = base::BitField<Type, 0, 3>;
+  using FromIndexField = base::BitField<int, 3, 29>;
   uint32_t bit_field_;
   HeapEntry* to_entry_;
   union {
@@ -124,6 +131,7 @@ class HeapEntry {
   void set_name(const char* name) { name_ = name; }
   SnapshotObjectId id() const { return id_; }
   size_t self_size() const { return self_size_; }
+  void add_self_size(size_t size) { self_size_ += size; }
   unsigned trace_node_id() const { return trace_node_id_; }
   int index() const { return index_; }
   V8_INLINE int children_count() const;
@@ -132,25 +140,53 @@ class HeapEntry {
   V8_INLINE HeapGraphEdge* child(int i);
   V8_INLINE Isolate* isolate() const;
 
-  void SetIndexedReference(
-      HeapGraphEdge::Type type, int index, HeapEntry* entry);
-  void SetNamedReference(
-      HeapGraphEdge::Type type, const char* name, HeapEntry* entry);
-  void SetIndexedAutoIndexReference(HeapGraphEdge::Type type,
-                                    HeapEntry* child) {
-    SetIndexedReference(type, children_count_ + 1, child);
+  void set_detachedness(v8::EmbedderGraph::Node::Detachedness value) {
+    detachedness_ = static_cast<uint8_t>(value);
+  }
+  uint8_t detachedness() const { return detachedness_; }
+
+  enum ReferenceVerification {
+    // Verify that the reference can be found via marking, if verification is
+    // enabled.
+    kVerify,
+
+    // Skip verifying that the reference can be found via marking, for any of
+    // the following reasons:
+
+    kEphemeron,
+    kOffHeapPointer,
+    kCustomWeakPointer,
+  };
+
+  void VerifyReference(HeapGraphEdge::Type type, HeapEntry* entry,
+                       HeapSnapshotGenerator* generator,
+                       ReferenceVerification verification);
+  void SetIndexedReference(HeapGraphEdge::Type type, int index,
+                           HeapEntry* entry, HeapSnapshotGenerator* generator,
+                           ReferenceVerification verification = kVerify);
+  void SetNamedReference(HeapGraphEdge::Type type, const char* name,
+                         HeapEntry* entry, HeapSnapshotGenerator* generator,
+                         ReferenceVerification verification = kVerify);
+  void SetIndexedAutoIndexReference(
+      HeapGraphEdge::Type type, HeapEntry* child,
+      HeapSnapshotGenerator* generator,
+      ReferenceVerification verification = kVerify) {
+    SetIndexedReference(type, children_count_ + 1, child, generator,
+                        verification);
   }
   void SetNamedAutoIndexReference(HeapGraphEdge::Type type,
                                   const char* description, HeapEntry* child,
-                                  StringsStorage* strings);
+                                  StringsStorage* strings,
+                                  HeapSnapshotGenerator* generator,
+                                  ReferenceVerification verification = kVerify);
 
   V8_EXPORT_PRIVATE void Print(const char* prefix, const char* edge_name,
-                               int max_depth, int indent);
+                               int max_depth, int indent) const;
 
  private:
   V8_INLINE std::vector<HeapGraphEdge*>::iterator children_begin() const;
   V8_INLINE std::vector<HeapGraphEdge*>::iterator children_end() const;
-  const char* TypeAsString();
+  const char* TypeAsString() const;
 
   unsigned type_: 4;
   unsigned index_ : 28;  // Supports up to ~250M objects.
@@ -160,7 +196,12 @@ class HeapEntry {
     unsigned children_count_;
     unsigned children_end_index_;
   };
+#ifdef V8_TARGET_ARCH_64_BIT
+  size_t self_size_ : 48;
+#else   // !V8_TARGET_ARCH_64_BIT
   size_t self_size_;
+#endif  // !V8_TARGET_ARCH_64_BIT
+  uint8_t detachedness_ = 0;
   HeapSnapshot* snapshot_;
   const char* name_;
   SnapshotObjectId id_;
@@ -175,7 +216,10 @@ class HeapEntry {
 // HeapSnapshotGenerator fills in a HeapSnapshot.
 class HeapSnapshot {
  public:
-  explicit HeapSnapshot(HeapProfiler* profiler);
+  explicit HeapSnapshot(HeapProfiler* profiler, bool global_objects_as_roots,
+                        bool capture_numeric_value);
+  HeapSnapshot(const HeapSnapshot&) = delete;
+  HeapSnapshot& operator=(const HeapSnapshot&) = delete;
   void Delete();
 
   HeapProfiler* profiler() const { return profiler_; }
@@ -185,7 +229,9 @@ class HeapSnapshot {
     return gc_subroot_entries_[static_cast<int>(root)];
   }
   std::deque<HeapEntry>& entries() { return entries_; }
+  const std::deque<HeapEntry>& entries() const { return entries_; }
   std::deque<HeapGraphEdge>& edges() { return edges_; }
+  const std::deque<HeapGraphEdge>& edges() const { return edges_; }
   std::vector<HeapGraphEdge*>& children() { return children_; }
   const std::vector<SourceLocation>& locations() const { return locations_; }
   void RememberLastJSObjectId();
@@ -193,6 +239,10 @@ class HeapSnapshot {
     return max_snapshot_js_object_id_;
   }
   bool is_complete() const { return !children_.empty(); }
+  bool treat_global_objects_as_roots() const {
+    return treat_global_objects_as_roots_;
+  }
+  bool capture_numeric_value() const { return capture_numeric_value_; }
 
   void AddLocation(HeapEntry* entry, int scriptId, int line, int col);
   HeapEntry* AddEntry(HeapEntry::Type type,
@@ -224,8 +274,8 @@ class HeapSnapshot {
   std::unordered_map<SnapshotObjectId, HeapEntry*> entries_by_id_cache_;
   std::vector<SourceLocation> locations_;
   SnapshotObjectId max_snapshot_js_object_id_ = -1;
-
-  DISALLOW_COPY_AND_ASSIGN(HeapSnapshot);
+  bool treat_global_objects_as_roots_;
+  bool capture_numeric_value_;
 };
 
 
@@ -242,6 +292,8 @@ class HeapObjectsMap {
   };
 
   explicit HeapObjectsMap(Heap* heap);
+  HeapObjectsMap(const HeapObjectsMap&) = delete;
+  HeapObjectsMap& operator=(const HeapObjectsMap&) = delete;
 
   Heap* heap() const { return heap_; }
 
@@ -249,11 +301,17 @@ class HeapObjectsMap {
   SnapshotObjectId FindOrAddEntry(Address addr,
                                   unsigned int size,
                                   bool accessed = true);
+  SnapshotObjectId FindMergedNativeEntry(NativeObject addr);
+  void AddMergedNativeEntry(NativeObject addr, Address canonical_addr);
   bool MoveObject(Address from, Address to, int size);
   void UpdateObjectSize(Address addr, int size);
   SnapshotObjectId last_assigned_id() const {
     return next_id_ - kObjectIdStep;
   }
+  SnapshotObjectId get_next_id() {
+    next_id_ += kObjectIdStep;
+    return next_id_ - kObjectIdStep;
+  }
 
   void StopHeapObjectsTracking();
   SnapshotObjectId PushHeapObjectsStats(OutputStream* stream,
@@ -285,9 +343,9 @@ class HeapObjectsMap {
   base::HashMap entries_map_;
   std::vector<EntryInfo> entries_;
   std::vector<TimeInterval> time_intervals_;
+  // Map from NativeObject to EntryInfo index in entries_.
+  std::unordered_map<NativeObject, size_t> merged_native_entries_map_;
   Heap* heap_;
-
-  DISALLOW_COPY_AND_ASSIGN(HeapObjectsMap);
 };
 
 // A typedef for referencing anything that can be snapshotted living
@@ -299,6 +357,7 @@ class HeapEntriesAllocator {
  public:
   virtual ~HeapEntriesAllocator() = default;
   virtual HeapEntry* AllocateEntry(HeapThing ptr) = 0;
+  virtual HeapEntry* AllocateEntry(Smi smi) = 0;
 };
 
 class SnapshottingProgressReportingInterface {
@@ -315,19 +374,25 @@ class V8_EXPORT_PRIVATE V8HeapExplorer : public HeapEntriesAllocator {
                  SnapshottingProgressReportingInterface* progress,
                  v8::HeapProfiler::ObjectNameResolver* resolver);
   ~V8HeapExplorer() override = default;
+  V8HeapExplorer(const V8HeapExplorer&) = delete;
+  V8HeapExplorer& operator=(const V8HeapExplorer&) = delete;
+
+  V8_INLINE Isolate* isolate() { return Isolate::FromHeap(heap_); }
 
   HeapEntry* AllocateEntry(HeapThing ptr) override;
-  int EstimateObjectsCount();
+  HeapEntry* AllocateEntry(Smi smi) override;
+  uint32_t EstimateObjectsCount();
   bool IterateAndExtractReferences(HeapSnapshotGenerator* generator);
-  void TagGlobalObjects();
-  void TagBuiltinCodeObject(Code code, const char* name);
+  void CollectGlobalObjectsTags();
+  void MakeGlobalObjectTagMap(const SafepointScope& safepoint_scope);
+  void TagBuiltinCodeObject(CodeT code, const char* name);
   HeapEntry* AddEntry(Address address,
                       HeapEntry::Type type,
                       const char* name,
                       size_t size);
 
-  static JSFunction GetConstructor(JSReceiver receiver);
-  static String GetConstructorName(JSObject object);
+  static JSFunction GetConstructor(Isolate* isolate, JSReceiver receiver);
+  static String GetConstructorName(Isolate* isolate, JSObject object);
 
  private:
   void MarkVisitedField(int offset);
@@ -337,6 +402,7 @@ class V8_EXPORT_PRIVATE V8HeapExplorer : public HeapEntriesAllocator {
                       const char* name);
 
   const char* GetSystemEntryName(HeapObject object);
+  HeapEntry::Type GetSystemEntryType(HeapObject object);
 
   void ExtractLocation(HeapEntry* entry, HeapObject object);
   void ExtractLocationForJSFunction(HeapEntry* entry, JSFunction func);
@@ -360,17 +426,24 @@ class V8_EXPORT_PRIVATE V8HeapExplorer : public HeapEntriesAllocator {
   void ExtractAccessorPairReferences(HeapEntry* entry, AccessorPair accessors);
   void ExtractCodeReferences(HeapEntry* entry, Code code);
   void ExtractCellReferences(HeapEntry* entry, Cell cell);
+  void ExtractJSWeakRefReferences(HeapEntry* entry, JSWeakRef js_weak_ref);
+  void ExtractWeakCellReferences(HeapEntry* entry, WeakCell weak_cell);
   void ExtractFeedbackCellReferences(HeapEntry* entry,
                                      FeedbackCell feedback_cell);
   void ExtractPropertyCellReferences(HeapEntry* entry, PropertyCell cell);
   void ExtractAllocationSiteReferences(HeapEntry* entry, AllocationSite site);
   void ExtractArrayBoilerplateDescriptionReferences(
       HeapEntry* entry, ArrayBoilerplateDescription value);
+  void ExtractRegExpBoilerplateDescriptionReferences(
+      HeapEntry* entry, RegExpBoilerplateDescription value);
   void ExtractJSArrayBufferReferences(HeapEntry* entry, JSArrayBuffer buffer);
   void ExtractJSPromiseReferences(HeapEntry* entry, JSPromise promise);
   void ExtractJSGeneratorObjectReferences(HeapEntry* entry,
                                           JSGeneratorObject generator);
   void ExtractFixedArrayReferences(HeapEntry* entry, FixedArray array);
+  void ExtractNumberReference(HeapEntry* entry, Object number);
+  void ExtractBytecodeArrayReferences(HeapEntry* entry, BytecodeArray bytecode);
+  void ExtractScopeInfoReferences(HeapEntry* entry, ScopeInfo info);
   void ExtractFeedbackVectorReferences(HeapEntry* entry,
                                        FeedbackVector feedback_vector);
   void ExtractDescriptorArrayReferences(HeapEntry* entry,
@@ -397,10 +470,12 @@ class V8_EXPORT_PRIVATE V8HeapExplorer : public HeapEntriesAllocator {
                             int field_offset = -1);
   void SetHiddenReference(HeapObject parent_obj, HeapEntry* parent_entry,
                           int index, Object child, int field_offset);
-  void SetWeakReference(HeapEntry* parent_entry, const char* reference_name,
-                        Object child_obj, int field_offset);
+  void SetWeakReference(
+      HeapEntry* parent_entry, const char* reference_name, Object child_obj,
+      int field_offset,
+      HeapEntry::ReferenceVerification verification = HeapEntry::kVerify);
   void SetWeakReference(HeapEntry* parent_entry, int index, Object child_obj,
-                        int field_offset);
+                        base::Optional<int> field_offset);
   void SetPropertyReference(HeapEntry* parent_entry, Name reference_name,
                             Object child,
                             const char* name_format_string = nullptr,
@@ -416,7 +491,10 @@ class V8_EXPORT_PRIVATE V8HeapExplorer : public HeapEntriesAllocator {
   void SetGcSubrootReference(Root root, const char* description, bool is_weak,
                              Object child);
   const char* GetStrongGcSubrootName(Object object);
-  void TagObject(Object obj, const char* tag);
+  void TagObject(Object obj, const char* tag,
+                 base::Optional<HeapEntry::Type> type = {});
+  void RecursivelyTagConstantPool(Object obj, const char* tag,
+                                  HeapEntry::Type type, int recursion_limit);
 
   HeapEntry* GetEntry(Object obj);
 
@@ -426,7 +504,10 @@ class V8_EXPORT_PRIVATE V8HeapExplorer : public HeapEntriesAllocator {
   HeapObjectsMap* heap_object_map_;
   SnapshottingProgressReportingInterface* progress_;
   HeapSnapshotGenerator* generator_ = nullptr;
-  std::unordered_map<JSGlobalObject, const char*, Object::Hasher> objects_tags_;
+  std::vector<std::pair<Handle<JSGlobalObject>, const char*>>
+      global_object_tag_pairs_;
+  std::unordered_map<JSGlobalObject, const char*, Object::Hasher>
+      global_object_tag_map_;
   std::unordered_map<Object, const char*, Object::Hasher>
       strong_gc_subroot_names_;
   std::unordered_set<JSGlobalObject, Object::Hasher> user_roots_;
@@ -436,8 +517,6 @@ class V8_EXPORT_PRIVATE V8HeapExplorer : public HeapEntriesAllocator {
 
   friend class IndexedReferencesExtractor;
   friend class RootsReferencesExtractor;
-
-  DISALLOW_COPY_AND_ASSIGN(V8HeapExplorer);
 };
 
 // An implementation of retained native objects extractor.
@@ -445,14 +524,21 @@ class NativeObjectsExplorer {
  public:
   NativeObjectsExplorer(HeapSnapshot* snapshot,
                         SnapshottingProgressReportingInterface* progress);
+  NativeObjectsExplorer(const NativeObjectsExplorer&) = delete;
+  NativeObjectsExplorer& operator=(const NativeObjectsExplorer&) = delete;
   bool IterateAndExtractReferences(HeapSnapshotGenerator* generator);
 
  private:
+  // Returns an entry for a given node, where node may be a V8 node or an
+  // embedder node. Returns the coresponding wrapper node if present.
   HeapEntry* EntryForEmbedderGraphNode(EmbedderGraph::Node* node);
+  void MergeNodeIntoEntry(HeapEntry* entry, EmbedderGraph::Node* original_node,
+                          EmbedderGraph::Node* wrapper_node);
 
   Isolate* isolate_;
   HeapSnapshot* snapshot_;
   StringsStorage* names_;
+  HeapObjectsMap* heap_object_map_;
   std::unique_ptr<HeapEntriesAllocator> embedder_graph_entries_allocator_;
   // Used during references extraction.
   HeapSnapshotGenerator* generator_ = nullptr;
@@ -460,20 +546,25 @@ class NativeObjectsExplorer {
   static HeapThing const kNativesRootObject;
 
   friend class GlobalHandlesExtractor;
-
-  DISALLOW_COPY_AND_ASSIGN(NativeObjectsExplorer);
 };
 
+class HeapEntryVerifier;
+
 class HeapSnapshotGenerator : public SnapshottingProgressReportingInterface {
  public:
   // The HeapEntriesMap instance is used to track a mapping between
   // real heap objects and their representations in heap snapshots.
   using HeapEntriesMap = std::unordered_map<HeapThing, HeapEntry*>;
+  // The SmiEntriesMap instance is used to track a mapping between smi and
+  // their representations in heap snapshots.
+  using SmiEntriesMap = std::unordered_map<int, HeapEntry*>;
 
   HeapSnapshotGenerator(HeapSnapshot* snapshot,
                         v8::ActivityControl* control,
                         v8::HeapProfiler::ObjectNameResolver* resolver,
                         Heap* heap);
+  HeapSnapshotGenerator(const HeapSnapshotGenerator&) = delete;
+  HeapSnapshotGenerator& operator=(const HeapSnapshotGenerator&) = delete;
   bool GenerateSnapshot();
 
   HeapEntry* FindEntry(HeapThing ptr) {
@@ -481,8 +572,41 @@ class HeapSnapshotGenerator : public SnapshottingProgressReportingInterface {
     return it != entries_map_.end() ? it->second : nullptr;
   }
 
+  HeapEntry* FindEntry(Smi smi) {
+    auto it = smis_map_.find(smi.value());
+    return it != smis_map_.end() ? it->second : nullptr;
+  }
+
   HeapEntry* AddEntry(HeapThing ptr, HeapEntriesAllocator* allocator) {
-    return entries_map_.emplace(ptr, allocator->AllocateEntry(ptr))
+    HeapEntry* result =
+        entries_map_.emplace(ptr, allocator->AllocateEntry(ptr)).first->second;
+#ifdef V8_ENABLE_HEAP_SNAPSHOT_VERIFY
+    if (FLAG_heap_snapshot_verify) {
+      reverse_entries_map_.emplace(result, ptr);
+    }
+#endif
+    return result;
+  }
+
+#ifdef V8_ENABLE_HEAP_SNAPSHOT_VERIFY
+  HeapThing FindHeapThingForHeapEntry(HeapEntry* entry) {
+    // The reverse lookup map is only populated if the verification flag is
+    // enabled.
+    DCHECK(FLAG_heap_snapshot_verify);
+
+    auto it = reverse_entries_map_.find(entry);
+    return it == reverse_entries_map_.end() ? nullptr : it->second;
+  }
+
+  HeapEntryVerifier* verifier() const { return verifier_; }
+  void set_verifier(HeapEntryVerifier* verifier) {
+    DCHECK_IMPLIES(verifier_, !verifier);
+    verifier_ = verifier;
+  }
+#endif
+
+  HeapEntry* AddEntry(Smi smi, HeapEntriesAllocator* allocator) {
+    return smis_map_.emplace(smi.value(), allocator->AllocateEntry(smi))
         .first->second;
   }
 
@@ -491,6 +615,13 @@ class HeapSnapshotGenerator : public SnapshottingProgressReportingInterface {
     return entry != nullptr ? entry : AddEntry(ptr, allocator);
   }
 
+  HeapEntry* FindOrAddEntry(Smi smi, HeapEntriesAllocator* allocator) {
+    HeapEntry* entry = FindEntry(smi);
+    return entry != nullptr ? entry : AddEntry(smi, allocator);
+  }
+
+  Heap* heap() const { return heap_; }
+
  private:
   bool FillReferences();
   void ProgressStep() override;
@@ -503,12 +634,16 @@ class HeapSnapshotGenerator : public SnapshottingProgressReportingInterface {
   NativeObjectsExplorer dom_explorer_;
   // Mapping from HeapThing pointers to HeapEntry indices.
   HeapEntriesMap entries_map_;
+  SmiEntriesMap smis_map_;
   // Used during snapshot generation.
-  int progress_counter_;
-  int progress_total_;
+  uint32_t progress_counter_;
+  uint32_t progress_total_;
   Heap* heap_;
 
-  DISALLOW_COPY_AND_ASSIGN(HeapSnapshotGenerator);
+#ifdef V8_ENABLE_HEAP_SNAPSHOT_VERIFY
+  std::unordered_map<HeapEntry*, HeapThing> reverse_entries_map_;
+  HeapEntryVerifier* verifier_ = nullptr;
+#endif
 };
 
 class OutputStreamWriter;
@@ -521,6 +656,9 @@ class HeapSnapshotJSONSerializer {
         next_node_id_(1),
         next_string_id_(1),
         writer_(nullptr) {}
+  HeapSnapshotJSONSerializer(const HeapSnapshotJSONSerializer&) = delete;
+  HeapSnapshotJSONSerializer& operator=(const HeapSnapshotJSONSerializer&) =
+      delete;
   void Serialize(v8::OutputStream* stream);
 
  private:
@@ -560,8 +698,6 @@ class HeapSnapshotJSONSerializer {
 
   friend class HeapSnapshotJSONSerializerEnumerator;
   friend class HeapSnapshotJSONSerializerIterator;
-
-  DISALLOW_COPY_AND_ASSIGN(HeapSnapshotJSONSerializer);
 };
 
 
