@@ -21,13 +21,31 @@
 'use strict';
 
 const {
+  ArrayPrototypeIndexOf,
+  ArrayPrototypeJoin,
+  ArrayPrototypePush,
+  ArrayPrototypeShift,
+  ArrayPrototypeSlice,
   Error,
+  ErrorCaptureStackTrace,
+  FunctionPrototypeBind,
+  NumberIsNaN,
   ObjectAssign,
   ObjectIs,
   ObjectKeys,
   ObjectPrototypeIsPrototypeOf,
-  Map,
-  RegExpPrototypeTest,
+  ReflectApply,
+  RegExpPrototypeExec,
+  RegExpPrototypeSymbolReplace,
+  SafeMap,
+  String,
+  StringPrototypeCharCodeAt,
+  StringPrototypeIncludes,
+  StringPrototypeIndexOf,
+  StringPrototypeReplace,
+  StringPrototypeSlice,
+  StringPrototypeSplit,
+  StringPrototypeStartsWith,
 } = primordials;
 
 const { Buffer } = require('buffer');
@@ -39,6 +57,7 @@ const {
     ERR_INVALID_RETURN_VALUE,
     ERR_MISSING_ARGS,
   },
+  isErrorStackTraceLimitWritable,
   overrideStackTrace,
 } = require('internal/errors');
 const AssertionError = require('internal/assert/assertion_error');
@@ -46,15 +65,20 @@ const { openSync, closeSync, readSync } = require('fs');
 const { inspect } = require('internal/util/inspect');
 const { isPromise, isRegExp } = require('internal/util/types');
 const { EOL } = require('internal/constants');
-const { NativeModule } = require('internal/bootstrap/loaders');
+const { BuiltinModule } = require('internal/bootstrap/realm');
+const { isError } = require('internal/util');
 
-const errorCache = new Map();
+const errorCache = new SafeMap();
 const CallTracker = require('internal/assert/calltracker');
+const {
+  validateFunction,
+} = require('internal/validators');
 
 let isDeepEqual;
 let isDeepStrictEqual;
 let parseExpressionAt;
 let findNodeAround;
+let tokenizer;
 let decoder;
 
 function lazyLoadComparison() {
@@ -74,10 +98,10 @@ const meta = [
   '\\u000f', '\\u0010', '\\u0011', '\\u0012', '\\u0013',
   '\\u0014', '\\u0015', '\\u0016', '\\u0017', '\\u0018',
   '\\u0019', '\\u001a', '\\u001b', '\\u001c', '\\u001d',
-  '\\u001e', '\\u001f'
+  '\\u001e', '\\u001f',
 ];
 
-const escapeFn = (str) => meta[str.charCodeAt(0)];
+const escapeFn = (str) => meta[StringPrototypeCharCodeAt(str, 0)];
 
 let warned = false;
 
@@ -101,6 +125,13 @@ function innerFail(obj) {
   throw new AssertionError(obj);
 }
 
+/**
+ * @param {any} actual
+ * @param {any} expected
+ * @param {string | Error} [message]
+ * @param {string} [operator]
+ * @param {Function} [stackStartFn]
+ */
 function fail(actual, expected, message, operator, stackStartFn) {
   const argsLen = arguments.length;
 
@@ -118,7 +149,7 @@ function fail(actual, expected, message, operator, stackStartFn) {
         'assert.fail() with more than one argument is deprecated. ' +
           'Please use assert.strictEqual() instead or only pass a message.',
         'DeprecationWarning',
-        'DEP0094'
+        'DEP0094',
       );
     }
     if (argsLen === 2)
@@ -132,7 +163,7 @@ function fail(actual, expected, message, operator, stackStartFn) {
     expected,
     operator: operator === undefined ? 'fail' : operator,
     stackStartFn: stackStartFn || fail,
-    message
+    message,
   };
   const err = new AssertionError(errArgs);
   if (internalMessage) {
@@ -184,7 +215,7 @@ function getCode(fd, line, column) {
   let lines = 0;
   // Prevent blocking the event loop by limiting the maximum amount of
   // data that may be read.
-  let maxReads = 32; // bytesPerRead * maxReads = 512 kb
+  let maxReads = 32; // bytesPerRead * maxReads = 512 KiB
   const bytesPerRead = 16384;
   // Use a single buffer up front that is reused until the call site is found.
   let buffer = Buffer.allocUnsafe(bytesPerRead);
@@ -213,71 +244,60 @@ function getCode(fd, line, column) {
 function parseCode(code, offset) {
   // Lazy load acorn.
   if (parseExpressionAt === undefined) {
-    const acorn = require('internal/deps/acorn/acorn/dist/acorn');
-    const privateMethods =
-      require('internal/deps/acorn-plugins/acorn-private-methods/index');
-    const classFields =
-      require('internal/deps/acorn-plugins/acorn-class-fields/index');
-    const numericSeparator =
-      require('internal/deps/acorn-plugins/acorn-numeric-separator/index');
-    const staticClassFeatures =
-      require('internal/deps/acorn-plugins/acorn-static-class-features/index');
-
+    const Parser = require('internal/deps/acorn/acorn/dist/acorn').Parser;
     ({ findNodeAround } = require('internal/deps/acorn/acorn-walk/dist/walk'));
 
-    const Parser = acorn.Parser.extend(
-      privateMethods,
-      classFields,
-      numericSeparator,
-      staticClassFeatures
-    );
-    parseExpressionAt = Parser.parseExpressionAt.bind(Parser);
+    parseExpressionAt = FunctionPrototypeBind(Parser.parseExpressionAt, Parser);
+    tokenizer = FunctionPrototypeBind(Parser.tokenizer, Parser);
   }
   let node;
-  let start = 0;
+  let start;
   // Parse the read code until the correct expression is found.
-  do {
+  for (const token of tokenizer(code, { ecmaVersion: 'latest' })) {
+    start = token.start;
+    if (start > offset) {
+      // No matching expression found. This could happen if the assert
+      // expression is bigger than the provided buffer.
+      break;
+    }
     try {
-      node = parseExpressionAt(code, start, { ecmaVersion: 11 });
-      start = node.end + 1 || start;
+      node = parseExpressionAt(code, start, { ecmaVersion: 'latest' });
       // Find the CallExpression in the tree.
       node = findNodeAround(node, offset, 'CallExpression');
-    } catch (err) {
-      // Unexpected token error and the like.
-      start += err.raisedAt || 1;
-      if (start > offset) {
-        // No matching expression found. This could happen if the assert
-        // expression is bigger than the provided buffer.
-        // eslint-disable-next-line no-throw-literal
-        throw null;
+      if (node?.node.end >= offset) {
+        return [
+          node.node.start,
+          StringPrototypeReplace(StringPrototypeSlice(code,
+                                                      node.node.start, node.node.end),
+                                 escapeSequencesRegExp, escapeFn),
+        ];
       }
+    // eslint-disable-next-line no-unused-vars
+    } catch (err) {
+      continue;
     }
-  } while (node === undefined || node.node.end < offset);
-
-  return [
-    node.node.start,
-    code.slice(node.node.start, node.node.end)
-        .replace(escapeSequencesRegExp, escapeFn)
-  ];
+  }
+  // eslint-disable-next-line no-throw-literal
+  throw null;
 }
 
 function getErrMessage(message, fn) {
   const tmpLimit = Error.stackTraceLimit;
+  const errorStackTraceLimitIsWritable = isErrorStackTraceLimitWritable();
   // Make sure the limit is set to 1. Otherwise it could fail (<= 0) or it
   // does to much work.
-  Error.stackTraceLimit = 1;
+  if (errorStackTraceLimitIsWritable) Error.stackTraceLimit = 1;
   // We only need the stack trace. To minimize the overhead use an object
   // instead of an error.
   const err = {};
-  // eslint-disable-next-line no-restricted-syntax
-  Error.captureStackTrace(err, fn);
-  Error.stackTraceLimit = tmpLimit;
+  ErrorCaptureStackTrace(err, fn);
+  if (errorStackTraceLimitIsWritable) Error.stackTraceLimit = tmpLimit;
 
   overrideStackTrace.set(err, (_, stack) => stack);
   const call = err.stack[0];
 
   const filename = call.getFileName();
-  let line = call.getLineNumber() - 1;
+  const line = call.getLineNumber() - 1;
   let column = call.getColumnNumber() - 1;
   let identifier;
   let code;
@@ -286,21 +306,13 @@ function getErrMessage(message, fn) {
     identifier = `${filename}${line}${column}`;
 
     // Skip Node.js modules!
-    if (filename.endsWith('.js') &&
-        NativeModule.exists(filename.slice(0, -3))) {
+    if (StringPrototypeStartsWith(filename, 'node:') &&
+        BuiltinModule.exists(StringPrototypeSlice(filename, 5))) {
       errorCache.set(identifier, undefined);
       return;
     }
   } else {
-    const fn = call.getFunction();
-    if (!fn) {
-      return message;
-    }
-    code = String(fn);
-    // For functions created with the Function constructor, V8 does not count
-    // the lines containing the function header.
-    line += 2;
-    identifier = `${code}${line}${column}`;
+    return message;
   }
 
   if (errorCache.has(identifier)) {
@@ -311,7 +323,7 @@ function getErrMessage(message, fn) {
   try {
     // Set the stack trace limit to zero. This makes sure unexpected token
     // errors are handled faster.
-    Error.stackTraceLimit = 0;
+    if (errorStackTraceLimitIsWritable) Error.stackTraceLimit = 0;
 
     if (filename) {
       if (decoder === undefined) {
@@ -320,28 +332,29 @@ function getErrMessage(message, fn) {
       }
       fd = openSync(filename, 'r', 0o666);
       // Reset column and message.
-      [column, message] = getCode(fd, line, column);
+      ({ 0: column, 1: message } = getCode(fd, line, column));
       // Flush unfinished multi byte characters.
       decoder.end();
     } else {
       for (let i = 0; i < line; i++) {
-        code = code.slice(code.indexOf('\n') + 1);
+        code = StringPrototypeSlice(code,
+                                    StringPrototypeIndexOf(code, '\n') + 1);
       }
-      [column, message] = parseCode(code, column);
+      ({ 0: column, 1: message } = parseCode(code, column));
     }
     // Always normalize indentation, otherwise the message could look weird.
-    if (message.includes('\n')) {
+    if (StringPrototypeIncludes(message, '\n')) {
       if (EOL === '\r\n') {
-        message = message.replace(/\r\n/g, '\n');
+        message = RegExpPrototypeSymbolReplace(/\r\n/g, message, '\n');
       }
-      const frames = message.split('\n');
-      message = frames.shift();
+      const frames = StringPrototypeSplit(message, '\n');
+      message = ArrayPrototypeShift(frames);
       for (const frame of frames) {
         let pos = 0;
         while (pos < column && (frame[pos] === ' ' || frame[pos] === '\t')) {
           pos++;
         }
-        message += `\n  ${frame.slice(pos)}`;
+        message += `\n  ${StringPrototypeSlice(frame, pos)}`;
       }
     }
     message = `The expression evaluated to a falsy value:\n\n  ${message}\n`;
@@ -355,7 +368,7 @@ function getErrMessage(message, fn) {
     errorCache.set(identifier, undefined);
   } finally {
     // Reset limit.
-    Error.stackTraceLimit = tmpLimit;
+    if (errorStackTraceLimitIsWritable) Error.stackTraceLimit = tmpLimit;
     if (fd !== undefined)
       closeSync(fd);
   }
@@ -380,57 +393,79 @@ function innerOk(fn, argLen, value, message) {
       expected: true,
       message,
       operator: '==',
-      stackStartFn: fn
+      stackStartFn: fn,
     });
     err.generatedMessage = generatedMessage;
     throw err;
   }
 }
 
-// Pure assertion tests whether a value is truthy, as determined
-// by !!value.
+/**
+ * Pure assertion tests whether a value is truthy, as determined
+ * by !!value.
+ * @param {...any} args
+ * @returns {void}
+ */
 function ok(...args) {
   innerOk(ok, args.length, ...args);
 }
 assert.ok = ok;
 
-// The equality assertion tests shallow, coercive equality with ==.
+/**
+ * The equality assertion tests shallow, coercive equality with ==.
+ * @param {any} actual
+ * @param {any} expected
+ * @param {string | Error} [message]
+ * @returns {void}
+ */
 /* eslint-disable no-restricted-properties */
 assert.equal = function equal(actual, expected, message) {
   if (arguments.length < 2) {
     throw new ERR_MISSING_ARGS('actual', 'expected');
   }
   // eslint-disable-next-line eqeqeq
-  if (actual != expected) {
+  if (actual != expected && (!NumberIsNaN(actual) || !NumberIsNaN(expected))) {
     innerFail({
       actual,
       expected,
       message,
       operator: '==',
-      stackStartFn: equal
+      stackStartFn: equal,
     });
   }
 };
 
-// The non-equality assertion tests for whether two objects are not
-// equal with !=.
+/**
+ * The non-equality assertion tests for whether two objects are not
+ * equal with !=.
+ * @param {any} actual
+ * @param {any} expected
+ * @param {string | Error} [message]
+ * @returns {void}
+ */
 assert.notEqual = function notEqual(actual, expected, message) {
   if (arguments.length < 2) {
     throw new ERR_MISSING_ARGS('actual', 'expected');
   }
   // eslint-disable-next-line eqeqeq
-  if (actual == expected) {
+  if (actual == expected || (NumberIsNaN(actual) && NumberIsNaN(expected))) {
     innerFail({
       actual,
       expected,
       message,
       operator: '!=',
-      stackStartFn: notEqual
+      stackStartFn: notEqual,
     });
   }
 };
 
-// The equivalence assertion tests a deep equality relation.
+/**
+ * The deep equivalence assertion tests a deep equality relation.
+ * @param {any} actual
+ * @param {any} expected
+ * @param {string | Error} [message]
+ * @returns {void}
+ */
 assert.deepEqual = function deepEqual(actual, expected, message) {
   if (arguments.length < 2) {
     throw new ERR_MISSING_ARGS('actual', 'expected');
@@ -442,12 +477,18 @@ assert.deepEqual = function deepEqual(actual, expected, message) {
       expected,
       message,
       operator: 'deepEqual',
-      stackStartFn: deepEqual
+      stackStartFn: deepEqual,
     });
   }
 };
 
-// The non-equivalence assertion tests for any deep inequality.
+/**
+ * The deep non-equivalence assertion tests for any deep inequality.
+ * @param {any} actual
+ * @param {any} expected
+ * @param {string | Error} [message]
+ * @returns {void}
+ */
 assert.notDeepEqual = function notDeepEqual(actual, expected, message) {
   if (arguments.length < 2) {
     throw new ERR_MISSING_ARGS('actual', 'expected');
@@ -459,12 +500,20 @@ assert.notDeepEqual = function notDeepEqual(actual, expected, message) {
       expected,
       message,
       operator: 'notDeepEqual',
-      stackStartFn: notDeepEqual
+      stackStartFn: notDeepEqual,
     });
   }
 };
 /* eslint-enable */
 
+/**
+ * The deep strict equivalence assertion tests a deep strict equality
+ * relation.
+ * @param {any} actual
+ * @param {any} expected
+ * @param {string | Error} [message]
+ * @returns {void}
+ */
 assert.deepStrictEqual = function deepStrictEqual(actual, expected, message) {
   if (arguments.length < 2) {
     throw new ERR_MISSING_ARGS('actual', 'expected');
@@ -476,11 +525,19 @@ assert.deepStrictEqual = function deepStrictEqual(actual, expected, message) {
       expected,
       message,
       operator: 'deepStrictEqual',
-      stackStartFn: deepStrictEqual
+      stackStartFn: deepStrictEqual,
     });
   }
 };
 
+/**
+ * The deep strict non-equivalence assertion tests for any deep strict
+ * inequality.
+ * @param {any} actual
+ * @param {any} expected
+ * @param {string | Error} [message]
+ * @returns {void}
+ */
 assert.notDeepStrictEqual = notDeepStrictEqual;
 function notDeepStrictEqual(actual, expected, message) {
   if (arguments.length < 2) {
@@ -493,11 +550,18 @@ function notDeepStrictEqual(actual, expected, message) {
       expected,
       message,
       operator: 'notDeepStrictEqual',
-      stackStartFn: notDeepStrictEqual
+      stackStartFn: notDeepStrictEqual,
     });
   }
 }
 
+/**
+ * The strict equivalence assertion tests a strict equality relation.
+ * @param {any} actual
+ * @param {any} expected
+ * @param {string | Error} [message]
+ * @returns {void}
+ */
 assert.strictEqual = function strictEqual(actual, expected, message) {
   if (arguments.length < 2) {
     throw new ERR_MISSING_ARGS('actual', 'expected');
@@ -508,11 +572,18 @@ assert.strictEqual = function strictEqual(actual, expected, message) {
       expected,
       message,
       operator: 'strictEqual',
-      stackStartFn: strictEqual
+      stackStartFn: strictEqual,
     });
   }
 };
 
+/**
+ * The strict non-equivalence assertion tests for any strict inequality.
+ * @param {any} actual
+ * @param {any} expected
+ * @param {string | Error} [message]
+ * @returns {void}
+ */
 assert.notStrictEqual = function notStrictEqual(actual, expected, message) {
   if (arguments.length < 2) {
     throw new ERR_MISSING_ARGS('actual', 'expected');
@@ -523,7 +594,7 @@ assert.notStrictEqual = function notStrictEqual(actual, expected, message) {
       expected,
       message,
       operator: 'notStrictEqual',
-      stackStartFn: notStrictEqual
+      stackStartFn: notStrictEqual,
     });
   }
 };
@@ -535,7 +606,7 @@ class Comparison {
         if (actual !== undefined &&
             typeof actual[key] === 'string' &&
             isRegExp(obj[key]) &&
-            RegExpPrototypeTest(obj[key], actual[key])) {
+            RegExpPrototypeExec(obj[key], actual[key]) !== null) {
           this[key] = actual[key];
         } else {
           this[key] = obj[key];
@@ -556,7 +627,7 @@ function compareExceptionKey(actual, expected, key, message, keys, fn) {
         actual: a,
         expected: b,
         operator: 'deepStrictEqual',
-        stackStartFn: fn
+        stackStartFn: fn,
       });
       err.actual = actual;
       err.expected = expected;
@@ -568,7 +639,7 @@ function compareExceptionKey(actual, expected, key, message, keys, fn) {
       expected,
       message,
       operator: fn.name,
-      stackStartFn: fn
+      stackStartFn: fn,
     });
   }
 }
@@ -581,7 +652,7 @@ function expectedException(actual, expected, message, fn) {
     // Handle regular expressions.
     if (isRegExp(expected)) {
       const str = String(actual);
-      if (RegExpPrototypeTest(expected, str))
+      if (RegExpPrototypeExec(expected, str) !== null)
         return;
 
       if (!message) {
@@ -597,7 +668,7 @@ function expectedException(actual, expected, message, fn) {
         expected,
         message,
         operator: 'deepStrictEqual',
-        stackStartFn: fn
+        stackStartFn: fn,
       });
       err.operator = fn.name;
       throw err;
@@ -607,7 +678,7 @@ function expectedException(actual, expected, message, fn) {
       // Special handle errors to make sure the name and the message are
       // compared as well.
       if (expected instanceof Error) {
-        keys.push('name', 'message');
+        ArrayPrototypePush(keys, 'name', 'message');
       } else if (keys.length === 0) {
         throw new ERR_INVALID_ARG_VALUE('error',
                                         expected, 'may not be an empty object');
@@ -616,7 +687,7 @@ function expectedException(actual, expected, message, fn) {
       for (const key of keys) {
         if (typeof actual[key] === 'string' &&
             isRegExp(expected[key]) &&
-            RegExpPrototypeTest(expected[key], actual[key])) {
+            RegExpPrototypeExec(expected[key], actual[key]) !== null) {
           continue;
         }
         compareExceptionKey(actual, expected, key, message, keys, fn);
@@ -628,12 +699,41 @@ function expectedException(actual, expected, message, fn) {
   } else if (expected.prototype !== undefined && actual instanceof expected) {
     return;
   } else if (ObjectPrototypeIsPrototypeOf(Error, expected)) {
-    throw actual;
+    if (!message) {
+      generatedMessage = true;
+      message = 'The error is expected to be an instance of ' +
+        `"${expected.name}". Received `;
+      if (isError(actual)) {
+        const name = (actual.constructor && actual.constructor.name) ||
+                     actual.name;
+        if (expected.name === name) {
+          message += 'an error with identical name but a different prototype.';
+        } else {
+          message += `"${name}"`;
+        }
+        if (actual.message) {
+          message += `\n\nError message:\n\n${actual.message}`;
+        }
+      } else {
+        message += `"${inspect(actual, { depth: -1 })}"`;
+      }
+    }
+    throwError = true;
   } else {
     // Check validation functions return value.
-    const res = expected.call({}, actual);
+    const res = ReflectApply(expected, {}, [actual]);
     if (res !== true) {
-      throw actual;
+      if (!message) {
+        generatedMessage = true;
+        const name = expected.name ? `"${expected.name}" ` : '';
+        message = `The ${name}validation function is expected to return` +
+          ` "true". Received ${inspect(res)}`;
+
+        if (isError(actual)) {
+          message += `\n\nCaught error:\n\n${actual}`;
+        }
+      }
+      throwError = true;
     }
   }
 
@@ -643,7 +743,7 @@ function expectedException(actual, expected, message, fn) {
       expected,
       message,
       operator: fn.name,
-      stackStartFn: fn
+      stackStartFn: fn,
     });
     err.generatedMessage = generatedMessage;
     throw err;
@@ -651,9 +751,7 @@ function expectedException(actual, expected, message, fn) {
 }
 
 function getActual(fn) {
-  if (typeof fn !== 'function') {
-    throw new ERR_INVALID_ARG_TYPE('fn', 'Function', fn);
-  }
+  validateFunction(fn, 'fn');
   try {
     fn();
   } catch (e) {
@@ -708,13 +806,13 @@ function expectsError(stackStartFn, actual, error, message) {
       if (actual.message === error) {
         throw new ERR_AMBIGUOUS_ARGUMENT(
           'error/message',
-          `The error message "${actual.message}" is identical to the message.`
+          `The error message "${actual.message}" is identical to the message.`,
         );
       }
     } else if (actual === error) {
       throw new ERR_AMBIGUOUS_ARGUMENT(
         'error/message',
-        `The error "${actual}" is identical to the message.`
+        `The error "${actual}" is identical to the message.`,
       );
     }
     message = error;
@@ -733,13 +831,13 @@ function expectsError(stackStartFn, actual, error, message) {
       details += ` (${error.name})`;
     }
     details += message ? `: ${message}` : '.';
-    const fnType = stackStartFn.name === 'rejects' ? 'rejection' : 'exception';
+    const fnType = stackStartFn === assert.rejects ? 'rejection' : 'exception';
     innerFail({
       actual: undefined,
       expected: error,
       operator: stackStartFn.name,
       message: `Missing expected ${fnType}${details}`,
-      stackStartFn
+      stackStartFn,
     });
   }
 
@@ -753,20 +851,20 @@ function hasMatchingError(actual, expected) {
   if (typeof expected !== 'function') {
     if (isRegExp(expected)) {
       const str = String(actual);
-      return RegExpPrototypeTest(expected, str);
+      return RegExpPrototypeExec(expected, str) !== null;
     }
     throw new ERR_INVALID_ARG_TYPE(
-      'expected', ['Function', 'RegExp'], expected
+      'expected', ['Function', 'RegExp'], expected,
     );
   }
   // Guard instanceof against arrow functions as they don't have a prototype.
   if (expected.prototype !== undefined && actual instanceof expected) {
     return true;
   }
-  if (Error.isPrototypeOf(expected)) {
+  if (ObjectPrototypeIsPrototypeOf(Error, expected)) {
     return false;
   }
-  return expected.call({}, actual) === true;
+  return ReflectApply(expected, {}, [actual]) === true;
 }
 
 function expectsNoError(stackStartFn, actual, error, message) {
@@ -780,7 +878,7 @@ function expectsNoError(stackStartFn, actual, error, message) {
 
   if (!error || hasMatchingError(actual, error)) {
     const details = message ? `: ${message}` : '.';
-    const fnType = stackStartFn.name === 'doesNotReject' ?
+    const fnType = stackStartFn === assert.doesNotReject ?
       'rejection' : 'exception';
     innerFail({
       actual,
@@ -788,28 +886,57 @@ function expectsNoError(stackStartFn, actual, error, message) {
       operator: stackStartFn.name,
       message: `Got unwanted ${fnType}${details}\n` +
                `Actual message: "${actual && actual.message}"`,
-      stackStartFn
+      stackStartFn,
     });
   }
   throw actual;
 }
 
+/**
+ * Expects the function `promiseFn` to throw an error.
+ * @param {() => any} promiseFn
+ * @param {...any} [args]
+ * @returns {void}
+ */
 assert.throws = function throws(promiseFn, ...args) {
   expectsError(throws, getActual(promiseFn), ...args);
 };
 
+/**
+ * Expects `promiseFn` function or its value to reject.
+ * @param {() => Promise<any>} promiseFn
+ * @param {...any} [args]
+ * @returns {Promise<void>}
+ */
 assert.rejects = async function rejects(promiseFn, ...args) {
   expectsError(rejects, await waitForActual(promiseFn), ...args);
 };
 
+/**
+ * Asserts that the function `fn` does not throw an error.
+ * @param {() => any} fn
+ * @param {...any} [args]
+ * @returns {void}
+ */
 assert.doesNotThrow = function doesNotThrow(fn, ...args) {
   expectsNoError(doesNotThrow, getActual(fn), ...args);
 };
 
+/**
+ * Expects `fn` or its value to not reject.
+ * @param {() => Promise<any>} fn
+ * @param {...any} [args]
+ * @returns {Promise<void>}
+ */
 assert.doesNotReject = async function doesNotReject(fn, ...args) {
   expectsNoError(doesNotReject, await waitForActual(fn), ...args);
 };
 
+/**
+ * Throws `value` if the value is not `null` or `undefined`.
+ * @param {any} err
+ * @returns {void}
+ */
 assert.ifError = function ifError(err) {
   if (err !== null && err !== undefined) {
     let message = 'ifError got unwanted exception: ';
@@ -828,7 +955,7 @@ assert.ifError = function ifError(err) {
       expected: null,
       operator: 'ifError',
       message,
-      stackStartFn: ifError
+      stackStartFn: ifError,
     });
 
     // Make sure we actually have a stack trace!
@@ -838,20 +965,27 @@ assert.ifError = function ifError(err) {
       // This will remove any duplicated frames from the error frames taken
       // from within `ifError` and add the original error frames to the newly
       // created ones.
-      const tmp2 = origStack.split('\n');
-      tmp2.shift();
-      // Filter all frames existing in err.stack.
-      let tmp1 = newErr.stack.split('\n');
-      for (const errFrame of tmp2) {
-        // Find the first occurrence of the frame.
-        const pos = tmp1.indexOf(errFrame);
-        if (pos !== -1) {
-          // Only keep new frames.
-          tmp1 = tmp1.slice(0, pos);
-          break;
+      const origStackStart = StringPrototypeIndexOf(origStack, '\n    at');
+      if (origStackStart !== -1) {
+        const originalFrames = StringPrototypeSplit(
+          StringPrototypeSlice(origStack, origStackStart + 1),
+          '\n',
+        );
+        // Filter all frames existing in err.stack.
+        let newFrames = StringPrototypeSplit(newErr.stack, '\n');
+        for (const errFrame of originalFrames) {
+          // Find the first occurrence of the frame.
+          const pos = ArrayPrototypeIndexOf(newFrames, errFrame);
+          if (pos !== -1) {
+            // Only keep new frames.
+            newFrames = ArrayPrototypeSlice(newFrames, 0, pos);
+            break;
+          }
         }
+        const stackStart = ArrayPrototypeJoin(newFrames, '\n');
+        const stackEnd = ArrayPrototypeJoin(originalFrames, '\n');
+        newErr.stack = `${stackStart}\n${stackEnd}`;
       }
-      newErr.stack = `${tmp1.join('\n')}\n${tmp2.join('\n')}`;
     }
 
     throw newErr;
@@ -861,12 +995,12 @@ assert.ifError = function ifError(err) {
 function internalMatch(string, regexp, message, fn) {
   if (!isRegExp(regexp)) {
     throw new ERR_INVALID_ARG_TYPE(
-      'regexp', 'RegExp', regexp
+      'regexp', 'RegExp', regexp,
     );
   }
-  const match = fn.name === 'match';
+  const match = fn === assert.match;
   if (typeof string !== 'string' ||
-      RegExpPrototypeTest(regexp, string) !== match) {
+      RegExpPrototypeExec(regexp, string) !== null !== match) {
     if (message instanceof Error) {
       throw message;
     }
@@ -886,31 +1020,51 @@ function internalMatch(string, regexp, message, fn) {
       expected: regexp,
       message,
       operator: fn.name,
-      stackStartFn: fn
+      stackStartFn: fn,
     });
     err.generatedMessage = generatedMessage;
     throw err;
   }
 }
 
+/**
+ * Expects the `string` input to match the regular expression.
+ * @param {string} string
+ * @param {RegExp} regexp
+ * @param {string | Error} [message]
+ * @returns {void}
+ */
 assert.match = function match(string, regexp, message) {
   internalMatch(string, regexp, message, match);
 };
 
+/**
+ * Expects the `string` input not to match the regular expression.
+ * @param {string} string
+ * @param {RegExp} regexp
+ * @param {string | Error} [message]
+ * @returns {void}
+ */
 assert.doesNotMatch = function doesNotMatch(string, regexp, message) {
   internalMatch(string, regexp, message, doesNotMatch);
 };
 
 assert.CallTracker = CallTracker;
 
-// Expose a strict only variant of assert
+/**
+ * Expose a strict only variant of assert.
+ * @param {...any} args
+ * @returns {void}
+ */
 function strict(...args) {
   innerOk(strict, args.length, ...args);
 }
+
 assert.strict = ObjectAssign(strict, assert, {
   equal: assert.strictEqual,
   deepEqual: assert.deepStrictEqual,
   notEqual: assert.notStrictEqual,
-  notDeepEqual: assert.notDeepStrictEqual
+  notDeepEqual: assert.notDeepStrictEqual,
 });
+
 assert.strict.strict = assert.strict;
