@@ -18,8 +18,7 @@ const {
 const {
   ERR_ASYNC_CALLBACK,
   ERR_ASYNC_TYPE,
-  ERR_INVALID_ARG_TYPE,
-  ERR_INVALID_ASYNC_ID
+  ERR_INVALID_ASYNC_ID,
 } = require('internal/errors').codes;
 const { kEmptyObject } = require('internal/util');
 const {
@@ -59,7 +58,7 @@ const {
 const {
   async_id_symbol, trigger_async_id_symbol,
   init_symbol, before_symbol, after_symbol, destroy_symbol,
-  promise_resolve_symbol
+  promise_resolve_symbol,
 } = internal_async_hooks.symbols;
 
 // Get constants
@@ -252,7 +251,7 @@ class AsyncResource {
         enumerable: true,
         value: this,
         writable: true,
-      }
+      },
     });
     return bound;
   }
@@ -271,25 +270,21 @@ const storageHook = createHook({
     for (let i = 0; i < storageList.length; ++i) {
       storageList[i]._propagate(resource, currentResource, type);
     }
-  }
+  },
 });
 
 class AsyncLocalStorage {
-  constructor(options = kEmptyObject) {
-    if (typeof options !== 'object' || options === null) {
-      throw new ERR_INVALID_ARG_TYPE('options', 'Object', options);
-    }
-
-    const { onPropagate = null } = options;
-    if (onPropagate !== null && typeof onPropagate !== 'function') {
-      throw new ERR_INVALID_ARG_TYPE('options.onPropagate',
-                                     'function',
-                                     onPropagate);
-    }
-
+  constructor() {
     this.kResourceStore = Symbol('kResourceStore');
     this.enabled = false;
-    this._onPropagate = onPropagate;
+  }
+
+  static bind(fn) {
+    return AsyncResource.bind(fn);
+  }
+
+  static snapshot() {
+    return AsyncLocalStorage.bind((cb, ...args) => cb(...args));
   }
 
   disable() {
@@ -316,9 +311,7 @@ class AsyncLocalStorage {
   _propagate(resource, triggerResource, type) {
     const store = triggerResource[this.kResourceStore];
     if (this.enabled) {
-      if (this._onPropagate === null || this._onPropagate(type, store)) {
-        resource[this.kResourceStore] = store;
-      }
+      resource[this.kResourceStore] = store;
     }
   }
 
