@@ -44,9 +44,13 @@ const {
   StringPrototypeTrim,
   SymbolSpecies,
   SymbolToPrimitive,
+  TypedArrayPrototypeGetBuffer,
   TypedArrayPrototypeGetByteLength,
+  TypedArrayPrototypeGetByteOffset,
   TypedArrayPrototypeFill,
+  TypedArrayPrototypeGetLength,
   TypedArrayPrototypeSet,
+  TypedArrayPrototypeSlice,
   Uint8Array,
   Uint8ArrayPrototype,
 } = primordials;
@@ -57,6 +61,8 @@ const {
   compareOffset,
   createFromString,
   fill: bindingFill,
+  isAscii: bindingIsAscii,
+  isUtf8: bindingIsUtf8,
   indexOfBuffer,
   indexOfNumber,
   indexOfString,
@@ -64,29 +70,31 @@ const {
   swap32: _swap32,
   swap64: _swap64,
   kMaxLength,
-  kStringMaxLength
+  kStringMaxLength,
 } = internalBinding('buffer');
 const {
-  getOwnNonIndexProperties,
-  propertyFilter: {
+  constants: {
     ALL_PROPERTIES,
-    ONLY_ENUMERABLE
+    ONLY_ENUMERABLE,
   },
+  getOwnNonIndexProperties,
 } = internalBinding('util');
 const {
   customInspectSymbol,
   isInsideNodeModules,
   lazyDOMException,
   normalizeEncoding,
-  kIsEncodingSymbol
+  kIsEncodingSymbol,
+  defineLazyProperties,
 } = require('internal/util');
 const {
   isAnyArrayBuffer,
   isArrayBufferView,
-  isUint8Array
+  isUint8Array,
+  isTypedArray,
 } = require('internal/util/types');
 const {
-  inspect: utilInspect
+  inspect: utilInspect,
 } = require('internal/util/inspect');
 const { encodings } = internalBinding('string_decoder');
 
@@ -98,7 +106,7 @@ const {
     ERR_INVALID_BUFFER_SIZE,
     ERR_OUT_OF_RANGE,
     ERR_MISSING_ARGS,
-    ERR_UNKNOWN_ENCODING
+    ERR_UNKNOWN_ENCODING,
   },
   genericNodeError,
   hideStackFrames,
@@ -108,7 +116,7 @@ const {
   validateBuffer,
   validateNumber,
   validateInteger,
-  validateString
+  validateString,
 } = require('internal/validators');
 // Provide validateInteger() but with kMaxLength as the default maximum value.
 const validateOffset = (value, name, min = 0, max = kMaxLength) =>
@@ -118,14 +126,9 @@ const {
   FastBuffer,
   markAsUntransferable,
   addBufferPrototypeMethods,
-  createUnsafeBuffer
+  createUnsafeBuffer,
 } = require('internal/buffer');
 
-const {
-  Blob,
-  resolveObjectURL,
-} = require('internal/blob');
-
 FastBuffer.prototype.constructor = Buffer;
 Buffer.prototype = FastBuffer.prototype;
 addBufferPrototypeMethods(Buffer.prototype);
@@ -135,14 +138,14 @@ const constants = ObjectDefineProperties({}, {
     __proto__: null,
     value: kMaxLength,
     writable: false,
-    enumerable: true
+    enumerable: true,
   },
   MAX_STRING_LENGTH: {
     __proto__: null,
     value: kStringMaxLength,
     writable: false,
-    enumerable: true
-  }
+    enumerable: true,
+  },
 });
 
 Buffer.poolSize = 8 * 1024;
@@ -220,8 +223,8 @@ function _copy(source, target, targetStart, sourceStart, sourceEnd) {
     sourceStart = 0;
   } else {
     sourceStart = toInteger(sourceStart, 0);
-    if (sourceStart < 0)
-      throw new ERR_OUT_OF_RANGE('sourceStart', '>= 0', sourceStart);
+    if (sourceStart < 0 || sourceStart > source.length)
+      throw new ERR_OUT_OF_RANGE('sourceStart', `>= 0 && <= ${source.length}`, sourceStart);
   }
 
   if (sourceEnd === undefined) {
@@ -235,12 +238,6 @@ function _copy(source, target, targetStart, sourceStart, sourceEnd) {
   if (targetStart >= target.length || sourceStart >= sourceEnd)
     return 0;
 
-  if (sourceStart > source.length) {
-    throw new ERR_OUT_OF_RANGE('sourceStart',
-                               `<= ${source.length}`,
-                               sourceStart);
-  }
-
   return _copyActual(source, target, targetStart, sourceStart, sourceEnd);
 }
 
@@ -287,7 +284,7 @@ ObjectDefineProperty(Buffer, SymbolSpecies, {
   __proto__: null,
   enumerable: false,
   configurable: true,
-  get() { return FastBuffer; }
+  get() { return FastBuffer; },
 });
 
 /**
@@ -328,10 +325,52 @@ Buffer.from = function from(value, encodingOrOffset, length) {
   throw new ERR_INVALID_ARG_TYPE(
     'first argument',
     ['string', 'Buffer', 'ArrayBuffer', 'Array', 'Array-like Object'],
-    value
+    value,
   );
 };
 
+/**
+ * Creates the Buffer as a copy of the underlying ArrayBuffer of the view
+ * rather than the contents of the view.
+ * @param {TypedArray} view
+ * @param {number} [offset]
+ * @param {number} [length]
+ * @returns {Buffer}
+ */
+Buffer.copyBytesFrom = function copyBytesFrom(view, offset, length) {
+  if (!isTypedArray(view)) {
+    throw new ERR_INVALID_ARG_TYPE('view', [ 'TypedArray' ], view);
+  }
+
+  const viewLength = TypedArrayPrototypeGetLength(view);
+  if (viewLength === 0) {
+    return Buffer.alloc(0);
+  }
+
+  if (offset !== undefined || length !== undefined) {
+    if (offset !== undefined) {
+      validateInteger(offset, 'offset', 0);
+      if (offset >= viewLength) return Buffer.alloc(0);
+    } else {
+      offset = 0;
+    }
+    let end;
+    if (length !== undefined) {
+      validateInteger(length, 'length', 0);
+      end = offset + length;
+    } else {
+      end = viewLength;
+    }
+
+    view = TypedArrayPrototypeSlice(view, offset, end);
+  }
+
+  return fromArrayLike(new Uint8Array(
+    TypedArrayPrototypeGetBuffer(view),
+    TypedArrayPrototypeGetByteOffset(view),
+    TypedArrayPrototypeGetByteLength(view)));
+};
+
 // Identical to the built-in %TypedArray%.of(), but avoids using the deprecated
 // Buffer() constructor. Must use arrow function syntax to avoid automatically
 // adding a `prototype` property and making the function a constructor.
@@ -598,7 +637,7 @@ const encodingOps = {
     write: (buf, string, offset, len) => buf.utf8Write(string, offset, len),
     slice: (buf, start, end) => buf.utf8Slice(start, end),
     indexOf: (buf, val, byteOffset, dir) =>
-      indexOfString(buf, val, byteOffset, encodingsMap.utf8, dir)
+      indexOfString(buf, val, byteOffset, encodingsMap.utf8, dir),
   },
   ucs2: {
     encoding: 'ucs2',
@@ -607,7 +646,7 @@ const encodingOps = {
     write: (buf, string, offset, len) => buf.ucs2Write(string, offset, len),
     slice: (buf, start, end) => buf.ucs2Slice(start, end),
     indexOf: (buf, val, byteOffset, dir) =>
-      indexOfString(buf, val, byteOffset, encodingsMap.utf16le, dir)
+      indexOfString(buf, val, byteOffset, encodingsMap.utf16le, dir),
   },
   utf16le: {
     encoding: 'utf16le',
@@ -616,7 +655,7 @@ const encodingOps = {
     write: (buf, string, offset, len) => buf.ucs2Write(string, offset, len),
     slice: (buf, start, end) => buf.ucs2Slice(start, end),
     indexOf: (buf, val, byteOffset, dir) =>
-      indexOfString(buf, val, byteOffset, encodingsMap.utf16le, dir)
+      indexOfString(buf, val, byteOffset, encodingsMap.utf16le, dir),
   },
   latin1: {
     encoding: 'latin1',
@@ -625,7 +664,7 @@ const encodingOps = {
     write: (buf, string, offset, len) => buf.latin1Write(string, offset, len),
     slice: (buf, start, end) => buf.latin1Slice(start, end),
     indexOf: (buf, val, byteOffset, dir) =>
-      indexOfString(buf, val, byteOffset, encodingsMap.latin1, dir)
+      indexOfString(buf, val, byteOffset, encodingsMap.latin1, dir),
   },
   ascii: {
     encoding: 'ascii',
@@ -638,7 +677,7 @@ const encodingOps = {
                     fromStringFast(val, encodingOps.ascii),
                     byteOffset,
                     encodingsMap.ascii,
-                    dir)
+                    dir),
   },
   base64: {
     encoding: 'base64',
@@ -651,7 +690,7 @@ const encodingOps = {
                     fromStringFast(val, encodingOps.base64),
                     byteOffset,
                     encodingsMap.base64,
-                    dir)
+                    dir),
   },
   base64url: {
     encoding: 'base64url',
@@ -665,7 +704,7 @@ const encodingOps = {
                     fromStringFast(val, encodingOps.base64url),
                     byteOffset,
                     encodingsMap.base64url,
-                    dir)
+                    dir),
   },
   hex: {
     encoding: 'hex',
@@ -678,8 +717,8 @@ const encodingOps = {
                     fromStringFast(val, encodingOps.hex),
                     byteOffset,
                     encodingsMap.hex,
-                    dir)
-  }
+                    dir),
+  },
 };
 function getEncodingOps(encoding) {
   encoding += '';
@@ -738,7 +777,7 @@ function byteLength(string, encoding) {
     }
 
     throw new ERR_INVALID_ARG_TYPE(
-      'string', ['string', 'Buffer', 'ArrayBuffer'], string
+      'string', ['string', 'Buffer', 'ArrayBuffer'], string,
     );
   }
 
@@ -765,7 +804,7 @@ ObjectDefineProperty(Buffer.prototype, 'parent', {
     if (!(this instanceof Buffer))
       return undefined;
     return this.buffer;
-  }
+  },
 });
 ObjectDefineProperty(Buffer.prototype, 'offset', {
   __proto__: null,
@@ -774,7 +813,7 @@ ObjectDefineProperty(Buffer.prototype, 'offset', {
     if (!(this instanceof Buffer))
       return undefined;
     return this.byteOffset;
-  }
+  },
 });
 
 Buffer.prototype.copy =
@@ -861,7 +900,7 @@ Buffer.prototype[customInspectSymbol] = function inspect(recurseTimes, ctx) {
       str += StringPrototypeSlice(utilInspect(obj, {
         ...ctx,
         breakLength: Infinity,
-        compact: true
+        compact: true,
       }), 27, -2);
     }
   }
@@ -959,7 +998,7 @@ function bidirectionalIndexOf(buffer, val, byteOffset, encoding, dir) {
   }
 
   throw new ERR_INVALID_ARG_TYPE(
-    'value', ['number', 'string', 'Buffer', 'Uint8Array'], val
+    'value', ['number', 'string', 'Buffer', 'Uint8Array'], val,
   );
 }
 
@@ -1194,7 +1233,7 @@ let transcode;
 if (internalBinding('config').hasIntl) {
   const {
     icuErrName,
-    transcode: _transcode
+    transcode: _transcode,
   } = internalBinding('icu');
 
   // Transcodes the Buffer from one encoding to another, returning a new
@@ -1215,7 +1254,7 @@ if (internalBinding('config').hasIntl) {
     const code = icuErrName(result);
     const err = genericNodeError(
       `Unable to transcode Buffer [${code}]`,
-      { code: code, errno: result }
+      { code: code, errno: result },
     );
     throw err;
   };
@@ -1256,6 +1295,8 @@ const kForgivingBase64AllowedChars = [
   0x2F, // /
   0x3D, // =
 ];
+const kEqualSignIndex = ArrayPrototypeIndexOf(kForgivingBase64AllowedChars,
+                                              0x3D);
 
 function atob(input) {
   // The implementation here has not been performance optimized in any way and
@@ -1267,6 +1308,7 @@ function atob(input) {
 
   input = `${input}`;
   let nonAsciiWhitespaceCharCount = 0;
+  let equalCharCount = 0;
 
   for (let n = 0; n < input.length; n++) {
     const index = ArrayPrototypeIndexOf(
@@ -1277,13 +1319,36 @@ function atob(input) {
       // The first 5 elements of `kForgivingBase64AllowedChars` are
       // ASCII whitespace char codes.
       nonAsciiWhitespaceCharCount++;
+
+      if (index === kEqualSignIndex) {
+        equalCharCount++;
+      } else if (equalCharCount) {
+        // The `=` char is only allowed at the end.
+        throw lazyDOMException('Invalid character', 'InvalidCharacterError');
+      }
+
+      if (equalCharCount > 2) {
+        // Only one more `=` is permitted after the first equal sign.
+        throw lazyDOMException('Invalid character', 'InvalidCharacterError');
+      }
     } else if (index === -1) {
       throw lazyDOMException('Invalid character', 'InvalidCharacterError');
     }
   }
 
+  let reminder = nonAsciiWhitespaceCharCount % 4;
+
+  // See #2, #3, #4 - https://infra.spec.whatwg.org/#forgiving-base64
+  if (!reminder) {
+    // Remove all trailing `=` characters and get the new reminder.
+    reminder = (nonAsciiWhitespaceCharCount - equalCharCount) % 4;
+  } else if (equalCharCount) {
+    // `=` should not in the input if there's a reminder.
+    throw lazyDOMException('Invalid character', 'InvalidCharacterError');
+  }
+
   // See #3 - https://infra.spec.whatwg.org/#forgiving-base64
-  if (nonAsciiWhitespaceCharCount % 4 === 1) {
+  if (reminder === 1) {
     throw lazyDOMException(
       'The string to be decoded is not correctly encoded.',
       'InvalidCharacterError');
@@ -1292,12 +1357,29 @@ function atob(input) {
   return Buffer.from(input, 'base64').toString('latin1');
 }
 
+function isUtf8(input) {
+  if (isTypedArray(input) || isAnyArrayBuffer(input)) {
+    return bindingIsUtf8(input);
+  }
+
+  throw new ERR_INVALID_ARG_TYPE('input', ['TypedArray', 'Buffer'], input);
+}
+
+function isAscii(input) {
+  if (isTypedArray(input) || isAnyArrayBuffer(input)) {
+    return bindingIsAscii(input);
+  }
+
+  throw new ERR_INVALID_ARG_TYPE('input', ['ArrayBuffer', 'Buffer', 'TypedArray'], input);
+}
+
 module.exports = {
-  Blob,
-  resolveObjectURL,
   Buffer,
   SlowBuffer,
   transcode,
+  isUtf8,
+  isAscii,
+
   // Legacy
   kMaxLength,
   kStringMaxLength,
@@ -1310,13 +1392,24 @@ ObjectDefineProperties(module.exports, {
     __proto__: null,
     configurable: false,
     enumerable: true,
-    value: constants
+    value: constants,
   },
   INSPECT_MAX_BYTES: {
     __proto__: null,
     configurable: true,
     enumerable: true,
     get() { return INSPECT_MAX_BYTES; },
-    set(val) { INSPECT_MAX_BYTES = val; }
-  }
+    set(val) { INSPECT_MAX_BYTES = val; },
+  },
 });
+
+defineLazyProperties(
+  module.exports,
+  'internal/blob',
+  ['Blob', 'resolveObjectURL'],
+);
+defineLazyProperties(
+  module.exports,
+  'internal/file',
+  ['File'],
+);
