@@ -39,8 +39,10 @@ const {
   ObjectPrototypeHasOwnProperty,
   RegExpPrototypeExec,
   SafeSet,
+  StringPrototypeIncludes,
   StringPrototypeSlice,
   StringPrototypeToUpperCase,
+  SymbolDispose,
 } = primordials;
 
 const {
@@ -55,7 +57,7 @@ let debug = require('internal/util/debuglog').debuglog(
   'child_process',
   (fn) => {
     debug = fn;
-  }
+  },
 );
 const { Buffer } = require('buffer');
 const { Pipe, constants: PipeConstants } = internalBinding('pipe_wrap');
@@ -88,17 +90,20 @@ const {
   getValidStdio,
   setupChannel,
   ChildProcess,
-  stdioStringToArray
+  stdioStringToArray,
 } = child_process;
 
 const MAX_BUFFER = 1024 * 1024;
 
+const isZOS = process.platform === 'os390';
+let addAbortListener;
+
 /**
  * Spawns a new Node.js process + fork.
  * @param {string|URL} modulePath
  * @param {string[]} [args]
  * @param {{
- *   cwd?: string;
+ *   cwd?: string | URL;
  *   detached?: boolean;
  *   env?: Record<string, string>;
  *   execPath?: string;
@@ -133,11 +138,13 @@ function fork(modulePath, args = [], options) {
   if (options != null) {
     validateObject(options, 'options');
   }
-  options = { ...options, shell: false };
+  options = { __proto__: null, ...options, shell: false };
   options.execPath = options.execPath || process.execPath;
+  validateArgumentNullCheck(options.execPath, 'options.execPath');
 
   // Prepare arguments for fork:
   execArgv = options.execArgv || process.execArgv;
+  validateArgumentsNullCheck(execArgv, 'options.execArgv');
 
   if (execArgv === process.execArgv && process._eval != null) {
     const index = ArrayPrototypeLastIndexOf(execArgv, process._eval);
@@ -180,19 +187,22 @@ function _forkChild(fd, serializationMode) {
 }
 
 function normalizeExecArgs(command, options, callback) {
+  validateString(command, 'command');
+  validateArgumentNullCheck(command, 'command');
+
   if (typeof options === 'function') {
     callback = options;
     options = undefined;
   }
 
   // Make a shallow copy so we don't clobber the user's options object.
-  options = { ...options };
+  options = { __proto__: null, ...options };
   options.shell = typeof options.shell === 'string' ? options.shell : true;
 
   return {
     file: command,
     options: options,
-    callback: callback
+    callback: callback,
   };
 }
 
@@ -247,7 +257,7 @@ const customPromiseExecFunction = (orig) => {
 ObjectDefineProperty(exec, promisify.custom, {
   __proto__: null,
   enumerable: false,
-  value: customPromiseExecFunction(exec)
+  value: customPromiseExecFunction(exec),
 });
 
 function normalizeExecFileArgs(file, args, options, callback) {
@@ -284,6 +294,7 @@ function normalizeExecFileArgs(file, args, options, callback) {
   // Validate argv0, if present.
   if (options.argv0 != null) {
     validateString(options.argv0, 'options.argv0');
+    validateArgumentNullCheck(options.argv0, 'options.argv0');
   }
 
   return { file, args, options, callback };
@@ -294,7 +305,7 @@ function normalizeExecFileArgs(file, args, options, callback) {
  * @param {string} file
  * @param {string[]} [args]
  * @param {{
- *   cwd?: string;
+ *   cwd?: string | URL;
  *   env?: Record<string, string>;
  *   encoding?: string;
  *   timeout?: number;
@@ -318,6 +329,7 @@ function execFile(file, args, options, callback) {
   ({ file, args, options, callback } = normalizeExecFileArgs(file, args, options, callback));
 
   options = {
+    __proto__: null,
     encoding: 'utf8',
     timeout: 0,
     maxBuffer: MAX_BUFFER,
@@ -325,7 +337,7 @@ function execFile(file, args, options, callback) {
     cwd: null,
     env: null,
     shell: false,
-    ...options
+    ...options,
   };
 
   // Validate the timeout, if present.
@@ -344,7 +356,7 @@ function execFile(file, args, options, callback) {
     signal: options.signal,
     uid: options.uid,
     windowsHide: !!options.windowsHide,
-    windowsVerbatimArguments: !!options.windowsVerbatimArguments
+    windowsVerbatimArguments: !!options.windowsVerbatimArguments,
   });
 
   let encoding;
@@ -410,7 +422,7 @@ function execFile(file, args, options, callback) {
       ex = genericNodeError(`Command failed: ${cmd}\n${stderr}`, {
         code: code < 0 ? getSystemErrorName(code) : code,
         killed: child.killed || killed,
-        signal: signal
+        signal: signal,
       });
     }
 
@@ -521,11 +533,20 @@ function execFile(file, args, options, callback) {
 ObjectDefineProperty(execFile, promisify.custom, {
   __proto__: null,
   enumerable: false,
-  value: customPromiseExecFunction(execFile)
+  value: customPromiseExecFunction(execFile),
 });
 
+function copyProcessEnvToEnv(env, name, optionEnv) {
+  if (process.env[name] &&
+      (!optionEnv ||
+       !ObjectPrototypeHasOwnProperty(optionEnv, name))) {
+    env[name] = process.env[name];
+  }
+}
+
 function normalizeSpawnArguments(file, args, options) {
   validateString(file, 'file');
+  validateArgumentNullCheck(file, 'file');
 
   if (file.length === 0)
     throw new ERR_INVALID_ARG_VALUE('file', file, 'cannot be empty');
@@ -541,6 +562,8 @@ function normalizeSpawnArguments(file, args, options) {
     args = [];
   }
 
+  validateArgumentsNullCheck(args, 'args');
+
   if (options === undefined)
     options = kEmptyObject;
   else
@@ -579,6 +602,7 @@ function normalizeSpawnArguments(file, args, options) {
   // Validate argv0, if present.
   if (options.argv0 != null) {
     validateString(options.argv0, 'options.argv0');
+    validateArgumentNullCheck(options.argv0, 'options.argv0');
   }
 
   // Validate windowsHide, if present.
@@ -594,6 +618,7 @@ function normalizeSpawnArguments(file, args, options) {
   }
 
   if (options.shell) {
+    validateArgumentNullCheck(options.shell, 'options.shell');
     const command = ArrayPrototypeJoin([file, ...args], ' ');
     // Set the shell, switches, and commands.
     if (process.platform === 'win32') {
@@ -630,9 +655,19 @@ function normalizeSpawnArguments(file, args, options) {
 
   // process.env.NODE_V8_COVERAGE always propagates, making it possible to
   // collect coverage for programs that spawn with white-listed environment.
-  if (process.env.NODE_V8_COVERAGE &&
-      !ObjectPrototypeHasOwnProperty(options.env || {}, 'NODE_V8_COVERAGE')) {
-    env.NODE_V8_COVERAGE = process.env.NODE_V8_COVERAGE;
+  copyProcessEnvToEnv(env, 'NODE_V8_COVERAGE', options.env);
+
+  if (isZOS) {
+    // The following environment variables must always propagate if set.
+    copyProcessEnvToEnv(env, '_BPXK_AUTOCVT', options.env);
+    copyProcessEnvToEnv(env, '_CEE_RUNOPTS', options.env);
+    copyProcessEnvToEnv(env, '_TAG_REDIR_ERR', options.env);
+    copyProcessEnvToEnv(env, '_TAG_REDIR_IN', options.env);
+    copyProcessEnvToEnv(env, '_TAG_REDIR_OUT', options.env);
+    copyProcessEnvToEnv(env, 'STEPLIB', options.env);
+    copyProcessEnvToEnv(env, 'LIBPATH', options.env);
+    copyProcessEnvToEnv(env, '_EDC_SIG_DFLT', options.env);
+    copyProcessEnvToEnv(env, '_EDC_SUSV3', options.env);
   }
 
   let envKeys = [];
@@ -654,19 +689,22 @@ function normalizeSpawnArguments(file, args, options) {
         }
         sawKey.add(uppercaseKey);
         return true;
-      }
+      },
     );
   }
 
   for (const key of envKeys) {
     const value = env[key];
     if (value !== undefined) {
+      validateArgumentNullCheck(key, `options.env['${key}']`);
+      validateArgumentNullCheck(value, `options.env['${key}']`);
       ArrayPrototypePush(envPairs, `${key}=${value}`);
     }
   }
 
   return {
     // Make a shallow copy so we don't clobber the user's options object.
+    __proto__: null,
     ...options,
     args,
     cwd,
@@ -678,12 +716,12 @@ function normalizeSpawnArguments(file, args, options) {
   };
 }
 
-function abortChildProcess(child, killSignal) {
+function abortChildProcess(child, killSignal, reason) {
   if (!child)
     return;
   try {
     if (child.kill(killSignal)) {
-      child.emit('error', new AbortError());
+      child.emit('error', new AbortError(undefined, { cause: reason }));
     }
   } catch (err) {
     child.emit('error', err);
@@ -695,7 +733,7 @@ function abortChildProcess(child, killSignal) {
  * @param {string} file
  * @param {string[]} [args]
  * @param {{
- *   cwd?: string;
+ *   cwd?: string | URL;
  *   env?: Record<string, string>;
  *   argv0?: string;
  *   stdio?: Array | string;
@@ -747,13 +785,13 @@ function spawn(file, args, options) {
     if (signal.aborted) {
       process.nextTick(onAbortListener);
     } else {
-      signal.addEventListener('abort', onAbortListener, { once: true });
-      child.once('exit',
-                 () => signal.removeEventListener('abort', onAbortListener));
+      addAbortListener ??= require('events').addAbortListener;
+      const disposable = addAbortListener(signal, onAbortListener);
+      child.once('exit', disposable[SymbolDispose]);
     }
 
     function onAbortListener() {
-      abortChildProcess(child, killSignal);
+      abortChildProcess(child, killSignal, options.signal.reason);
     }
   }
 
@@ -765,7 +803,7 @@ function spawn(file, args, options) {
  * @param {string} file
  * @param {string[]} [args]
  * @param {{
- *   cwd?: string;
+ *   cwd?: string | URL;
  *   input?: string | Buffer | TypedArray | DataView;
  *   argv0?: string;
  *   stdio?: string | Array;
@@ -792,8 +830,9 @@ function spawn(file, args, options) {
  */
 function spawnSync(file, args, options) {
   options = {
+    __proto__: null,
     maxBuffer: MAX_BUFFER,
-    ...normalizeSpawnArguments(file, args, options)
+    ...normalizeSpawnArguments(file, args, options),
   };
 
   debug('spawnSync', options);
@@ -858,7 +897,7 @@ function checkExecSyncError(ret, args, cmd) {
  * @param {string} file
  * @param {string[]} [args]
  * @param {{
- *   cwd?: string;
+ *   cwd?: string | URL;
  *   input?: string | Buffer | TypedArray | DataView;
  *   stdio?: string | Array;
  *   env?: Record<string, string>;
@@ -896,7 +935,7 @@ function execFileSync(file, args, options) {
  * Spawns a shell executing the given `command` synchronously.
  * @param {string} command
  * @param {{
- *   cwd?: string;
+ *   cwd?: string | URL;
  *   input?: string | Buffer | TypedArray | DataView;
  *   stdio?: string | Array;
  *   env?: Record<string, string>;
@@ -920,7 +959,7 @@ function execSync(command, options) {
   if (inheritStderr && ret.stderr)
     process.stderr.write(ret.stderr);
 
-  const err = checkExecSyncError(ret, opts.args, command);
+  const err = checkExecSyncError(ret, undefined, command);
 
   if (err)
     throw err;
@@ -929,6 +968,20 @@ function execSync(command, options) {
 }
 
 
+function validateArgumentNullCheck(arg, propName) {
+  if (typeof arg === 'string' && StringPrototypeIncludes(arg, '\u0000')) {
+    throw new ERR_INVALID_ARG_VALUE(propName, arg, 'must be a string without null bytes');
+  }
+}
+
+
+function validateArgumentsNullCheck(args, propName) {
+  for (let i = 0; i < args.length; ++i) {
+    validateArgumentNullCheck(args[i], `${propName}[${i}]`);
+  }
+}
+
+
 function validateTimeout(timeout) {
   if (timeout != null && !(NumberIsInteger(timeout) && timeout >= 0)) {
     throw new ERR_OUT_OF_RANGE('timeout', 'an unsigned integer', timeout);
@@ -964,5 +1017,5 @@ module.exports = {
   execSync,
   fork,
   spawn,
-  spawnSync
+  spawnSync,
 };
