@@ -30,6 +30,8 @@ const {
   ObjectDefineProperty,
   ObjectSetPrototypeOf,
   ReflectApply,
+  SymbolAsyncDispose,
+  SymbolDispose,
 } = primordials;
 
 const errors = require('internal/errors');
@@ -49,7 +51,7 @@ const {
   ERR_SOCKET_DGRAM_IS_CONNECTED,
   ERR_SOCKET_DGRAM_NOT_CONNECTED,
   ERR_SOCKET_DGRAM_NOT_RUNNING,
-  ERR_INVALID_FD_TYPE
+  ERR_INVALID_FD_TYPE,
 } = errors.codes;
 const {
   isInt32,
@@ -59,19 +61,19 @@ const {
   validatePort,
 } = require('internal/validators');
 const { Buffer } = require('buffer');
-const { deprecate } = require('internal/util');
+const { deprecate, promisify } = require('internal/util');
 const { isArrayBufferView } = require('internal/util/types');
 const EventEmitter = require('events');
 const {
   defaultTriggerAsyncIdScope,
-  symbols: { async_id_symbol, owner_symbol }
+  symbols: { async_id_symbol, owner_symbol },
 } = require('internal/async_hooks');
 const { UV_UDP_REUSEADDR } = internalBinding('constants').os;
 
 const {
   constants: { UV_UDP_IPV6ONLY },
   UDP,
-  SendWrap
+  SendWrap,
 } = internalBinding('udp_wrap');
 
 const dc = require('diagnostics_channel');
@@ -132,20 +134,20 @@ function Socket(type, listener) {
     reuseAddr: options && options.reuseAddr, // Use UV_UDP_REUSEADDR if true.
     ipv6Only: options && options.ipv6Only,
     recvBufferSize,
-    sendBufferSize
+    sendBufferSize,
   };
 
   if (options?.signal !== undefined) {
     const { signal } = options;
     validateAbortSignal(signal, 'options.signal');
     const onAborted = () => {
-      this.close();
+      if (this[kStateSymbol].handle) this.close();
     };
     if (signal.aborted) {
       onAborted();
     } else {
-      signal.addEventListener('abort', onAborted);
-      this.once('close', () => signal.removeEventListener('abort', onAborted));
+      const disposable = EventEmitter.addAbortListener(signal, onAborted);
+      this.once('close', disposable[SymbolDispose]);
     }
   }
   if (udpSocketChannel.hasSubscribers) {
@@ -184,7 +186,10 @@ function startListening(socket) {
 function replaceHandle(self, newHandle) {
   const state = self[kStateSymbol];
   const oldHandle = state.handle;
-
+  // Sync the old handle state to new handle
+  if (!oldHandle.hasRef() && typeof newHandle.unref === 'function') {
+    newHandle.unref();
+  }
   // Set up the handle that we got from primary.
   newHandle.lookup = oldHandle.lookup;
   newHandle.bind = oldHandle.bind;
@@ -279,7 +284,7 @@ Socket.prototype.bind = function(port_, address_ /* , callback */) {
         port: null,
         addressType: this.type,
         fd,
-        flags: null
+        flags: null,
       }, (err) => {
         // Callback to handle error.
         const ex = errnoException(err, 'open');
@@ -343,7 +348,7 @@ Socket.prototype.bind = function(port_, address_ /* , callback */) {
         port: port,
         addressType: this.type,
         fd: -1,
-        flags: flags
+        flags: flags,
       }, (err) => {
         // Callback to handle error.
         const ex = exceptionWithHostPort(err, 'bind', ip, port);
@@ -409,7 +414,7 @@ function _connect(port, address, callback) {
     defaultTriggerAsyncIdScope(
       this[async_id_symbol],
       doConnect,
-      ex, this, ip, address, port, callback
+      ex, this, ip, address, port, callback,
     );
   };
 
@@ -659,7 +664,7 @@ Socket.prototype.send = function(buffer,
     defaultTriggerAsyncIdScope(
       this[async_id_symbol],
       doSend,
-      ex, this, ip, list, address, port, callback
+      ex, this, ip, list, address, port, callback,
     );
   };
 
@@ -749,6 +754,13 @@ Socket.prototype.close = function(callback) {
   return this;
 };
 
+Socket.prototype[SymbolAsyncDispose] = async function() {
+  if (!this[kStateSymbol].handle) {
+    return;
+  }
+  return FunctionPrototypeCall(promisify(this.close), this);
+};
+
 
 function socketCloseNT(self) {
   self.emit('close');
@@ -992,7 +1004,7 @@ ObjectDefineProperty(Socket.prototype, '_handle', {
   }, 'Socket.prototype._handle is deprecated', 'DEP0112'),
   set: deprecate(function(val) {
     this[kStateSymbol].handle = val;
-  }, 'Socket.prototype._handle is deprecated', 'DEP0112')
+  }, 'Socket.prototype._handle is deprecated', 'DEP0112'),
 });
 
 
@@ -1003,7 +1015,7 @@ ObjectDefineProperty(Socket.prototype, '_receiving', {
   }, 'Socket.prototype._receiving is deprecated', 'DEP0112'),
   set: deprecate(function(val) {
     this[kStateSymbol].receiving = val;
-  }, 'Socket.prototype._receiving is deprecated', 'DEP0112')
+  }, 'Socket.prototype._receiving is deprecated', 'DEP0112'),
 });
 
 
@@ -1014,7 +1026,7 @@ ObjectDefineProperty(Socket.prototype, '_bindState', {
   }, 'Socket.prototype._bindState is deprecated', 'DEP0112'),
   set: deprecate(function(val) {
     this[kStateSymbol].bindState = val;
-  }, 'Socket.prototype._bindState is deprecated', 'DEP0112')
+  }, 'Socket.prototype._bindState is deprecated', 'DEP0112'),
 });
 
 
@@ -1025,7 +1037,7 @@ ObjectDefineProperty(Socket.prototype, '_queue', {
   }, 'Socket.prototype._queue is deprecated', 'DEP0112'),
   set: deprecate(function(val) {
     this[kStateSymbol].queue = val;
-  }, 'Socket.prototype._queue is deprecated', 'DEP0112')
+  }, 'Socket.prototype._queue is deprecated', 'DEP0112'),
 });
 
 
@@ -1036,7 +1048,7 @@ ObjectDefineProperty(Socket.prototype, '_reuseAddr', {
   }, 'Socket.prototype._reuseAddr is deprecated', 'DEP0112'),
   set: deprecate(function(val) {
     this[kStateSymbol].reuseAddr = val;
-  }, 'Socket.prototype._reuseAddr is deprecated', 'DEP0112')
+  }, 'Socket.prototype._reuseAddr is deprecated', 'DEP0112'),
 });
 
 
@@ -1055,7 +1067,7 @@ Socket.prototype._stopReceiving = deprecate(function() {
 ObjectDefineProperty(UDP.prototype, 'owner', {
   __proto__: null,
   get() { return this[owner_symbol]; },
-  set(v) { return this[owner_symbol] = v; }
+  set(v) { return this[owner_symbol] = v; },
 });
 
 
@@ -1063,8 +1075,8 @@ module.exports = {
   _createSocketHandle: deprecate(
     _createSocketHandle,
     'dgram._createSocketHandle() is deprecated',
-    'DEP0112'
+    'DEP0112',
   ),
   createSocket,
-  Socket
+  Socket,
 };
