@@ -31,6 +31,7 @@ const {
   JSONStringify,
   ObjectAssign,
   ObjectSetPrototypeOf,
+  ReflectApply,
   ReflectConstruct,
 } = primordials;
 
@@ -43,6 +44,7 @@ assertCrypto();
 const tls = require('tls');
 const { Agent: HttpAgent } = require('_http_agent');
 const {
+  httpServerPreClose,
   Server: HttpServer,
   setupConnectionsTracking,
   storeHTTPOptions,
@@ -52,7 +54,7 @@ const { ClientRequest } = require('_http_client');
 let debug = require('internal/util/debuglog').debuglog('https', (fn) => {
   debug = fn;
 });
-const { URL, urlToHttpOptions, searchParamsSymbol } = require('internal/url');
+const { URL, urlToHttpOptions, isURL } = require('internal/url');
 
 function Server(opts, requestListener) {
   if (!(this instanceof Server)) return new Server(opts, requestListener);
@@ -86,8 +88,9 @@ function Server(opts, requestListener) {
 
   this.timeout = 0;
   this.maxHeadersCount = null;
-  setupConnectionsTracking(this);
+  this.on('listening', setupConnectionsTracking);
 }
+
 ObjectSetPrototypeOf(Server.prototype, tls.Server.prototype);
 ObjectSetPrototypeOf(Server, tls.Server);
 
@@ -97,6 +100,11 @@ Server.prototype.closeIdleConnections = HttpServer.prototype.closeIdleConnection
 
 Server.prototype.setTimeout = HttpServer.prototype.setTimeout;
 
+Server.prototype.close = function() {
+  httpServerPreClose(this);
+  ReflectApply(tls.Server.prototype.close, this, arguments);
+};
+
 /**
  * Creates a new `https.Server` instance.
  * @param {{
@@ -142,7 +150,7 @@ function createConnection(port, host, options) {
       debug('reuse session for %j', options._agentKey);
       options = {
         session,
-        ...options
+        ...options,
       };
     }
   }
@@ -193,7 +201,7 @@ function Agent(options) {
 
   this._sessionCache = {
     map: {},
-    list: []
+    list: [],
   };
 }
 ObjectSetPrototypeOf(Agent.prototype, HttpAgent.prototype);
@@ -344,9 +352,7 @@ function request(...args) {
   if (typeof args[0] === 'string') {
     const urlStr = ArrayPrototypeShift(args);
     options = urlToHttpOptions(new URL(urlStr));
-  } else if (args[0] && args[0][searchParamsSymbol] &&
-             args[0][searchParamsSymbol][searchParamsSymbol]) {
-    // url.URL instance
+  } else if (isURL(args[0])) {
     options = urlToHttpOptions(ArrayPrototypeShift(args));
   }
 
@@ -402,5 +408,5 @@ module.exports = {
   Server,
   createServer,
   get,
-  request
+  request,
 };
