@@ -3,7 +3,7 @@
 const {
   JSONParse,
   JSONStringify,
-  Map,
+  SafeMap,
   Symbol,
 } = primordials;
 
@@ -16,8 +16,6 @@ const {
   ERR_INSPECTOR_NOT_CONNECTED,
   ERR_INSPECTOR_NOT_ACTIVE,
   ERR_INSPECTOR_NOT_WORKER,
-  ERR_INVALID_ARG_TYPE,
-  ERR_INVALID_CALLBACK
 } = require('internal/errors').codes;
 
 const { hasInspector } = internalBinding('config');
@@ -26,8 +24,15 @@ if (!hasInspector)
 
 const EventEmitter = require('events');
 const { queueMicrotask } = require('internal/process/task_queues');
-const { validateString } = require('internal/validators');
+const {
+  isUint32,
+  validateFunction,
+  validateInt32,
+  validateObject,
+  validateString,
+} = require('internal/validators');
 const { isMainThread } = require('worker_threads');
+const { _debugEnd } = internalBinding('process_methods');
 
 const {
   Connection,
@@ -35,7 +40,8 @@ const {
   open,
   url,
   isEnabled,
-  waitForDebugger
+  waitForDebugger,
+  console,
 } = internalBinding('inspector');
 
 const connectionSymbol = Symbol('connectionProperty');
@@ -48,9 +54,13 @@ class Session extends EventEmitter {
     super();
     this[connectionSymbol] = null;
     this[nextIdSymbol] = 1;
-    this[messageCallbacksSymbol] = new Map();
+    this[messageCallbacksSymbol] = new SafeMap();
   }
 
+  /**
+   * Connects the session to the inspector back-end.
+   * @returns {void}
+   */
   connect() {
     if (this[connectionSymbol])
       throw new ERR_INSPECTOR_ALREADY_CONNECTED('The inspector session');
@@ -58,6 +68,11 @@ class Session extends EventEmitter {
       new Connection((message) => this[onMessageSymbol](message));
   }
 
+  /**
+   * Connects the session to the main thread
+   * inspector back-end.
+   * @returns {void}
+   */
   connectToMainThread() {
     if (isMainThread)
       throw new ERR_INSPECTOR_NOT_WORKER();
@@ -91,17 +106,24 @@ class Session extends EventEmitter {
     }
   }
 
+  /**
+   * Posts a message to the inspector back-end.
+   * @param {string} method
+   * @param {Record<unknown, unknown>} [params]
+   * @param {Function} [callback]
+   * @returns {void}
+   */
   post(method, params, callback) {
     validateString(method, 'method');
     if (!callback && typeof params === 'function') {
       callback = params;
       params = null;
     }
-    if (params && typeof params !== 'object') {
-      throw new ERR_INVALID_ARG_TYPE('params', 'Object', params);
+    if (params) {
+      validateObject(params, 'params');
     }
-    if (callback && typeof callback !== 'function') {
-      throw new ERR_INVALID_CALLBACK(callback);
+    if (callback) {
+      validateFunction(callback, 'callback');
     }
 
     if (!this[connectionSymbol]) {
@@ -118,6 +140,12 @@ class Session extends EventEmitter {
     this[connectionSymbol].dispatch(JSONStringify(message));
   }
 
+  /**
+   * Immediately closes the session, all pending
+   * message callbacks will be called with an
+   * error.
+   * @returns {void}
+   */
   disconnect() {
     if (!this[connectionSymbol])
       return;
@@ -132,15 +160,35 @@ class Session extends EventEmitter {
   }
 }
 
+/**
+ * Activates inspector on host and port.
+ * @param {number} [port]
+ * @param {string} [host]
+ * @param {boolean} [wait]
+ * @returns {void}
+ */
 function inspectorOpen(port, host, wait) {
   if (isEnabled()) {
     throw new ERR_INSPECTOR_ALREADY_ACTIVATED();
   }
+  // inspectorOpen() currently does not typecheck its arguments and adding
+  // such checks would be a potentially breaking change. However, the native
+  // open() function requires the port to fit into a 16-bit unsigned integer,
+  // causing an integer overflow otherwise, so we at least need to prevent that.
+  if (isUint32(port)) {
+    validateInt32(port, 'port', 0, 65535);
+  }
   open(port, host);
   if (wait)
     waitForDebugger();
 }
 
+/**
+ * Blocks until a client (existing or connected later)
+ * has sent the `Runtime.runIfWaitingForDebugger`
+ * command.
+ * @returns {void}
+ */
 function inspectorWaitForDebugger() {
   if (!waitForDebugger())
     throw new ERR_INSPECTOR_NOT_ACTIVE();
@@ -148,11 +196,9 @@ function inspectorWaitForDebugger() {
 
 module.exports = {
   open: inspectorOpen,
-  close: process._debugEnd,
-  url: url,
+  close: _debugEnd,
+  url,
   waitForDebugger: inspectorWaitForDebugger,
-  // This is dynamically added during bootstrap,
-  // where the console from the VM is still available
-  console: require('internal/util/inspector').consoleFromVM,
-  Session
+  console,
+  Session,
 };
