@@ -2,9 +2,17 @@
 
 const {
   ArrayIsArray,
+  ArrayPrototypePush,
+  ArrayPrototypeReduce,
+  ArrayPrototypeSlice,
+  FunctionPrototype,
+  FunctionPrototypeCall,
   ObjectDefineProperty,
   ObjectSetPrototypeOf,
+  ReflectApply,
+  StringPrototypeSlice,
   Symbol,
+  SymbolDispose,
   Uint8Array,
 } = primordials;
 
@@ -12,17 +20,22 @@ const {
   errnoException,
   codes: {
     ERR_INVALID_ARG_TYPE,
+    ERR_INVALID_ARG_VALUE,
     ERR_INVALID_HANDLE_TYPE,
-    ERR_INVALID_OPT_VALUE,
     ERR_INVALID_SYNC_FORK_INPUT,
     ERR_IPC_CHANNEL_CLOSED,
     ERR_IPC_DISCONNECTED,
     ERR_IPC_ONE_PIPE,
     ERR_IPC_SYNC_FORK,
-    ERR_MISSING_ARGS
-  }
+    ERR_MISSING_ARGS,
+  },
 } = require('internal/errors');
-const { validateString } = require('internal/validators');
+const {
+  validateArray,
+  validateObject,
+  validateOneOf,
+  validateString,
+} = require('internal/validators');
 const EventEmitter = require('events');
 const net = require('net');
 const dgram = require('dgram');
@@ -35,7 +48,7 @@ const {
   kReadBytesOrError,
   kArrayBufferOffset,
   kLastWriteWasAsync,
-  streamBaseState
+  streamBaseState,
 } = internalBinding('stream_wrap');
 const { Pipe, constants: PipeConstants } = internalBinding('pipe_wrap');
 const { TCP } = internalBinding('tcp_wrap');
@@ -43,7 +56,7 @@ const { TTY } = internalBinding('tty_wrap');
 const { UDP } = internalBinding('udp_wrap');
 const SocketList = require('internal/socket_list');
 const { owner_symbol } = require('internal/async_hooks').symbols;
-const { convertToValidSignal } = require('internal/util');
+const { convertToValidSignal, deprecate } = require('internal/util');
 const { isArrayBufferView } = require('internal/util/types');
 const spawn_sync = internalBinding('spawn_sync');
 const { kStateSymbol } = require('internal/dgram');
@@ -56,7 +69,7 @@ const {
   UV_ENFILE,
   UV_ENOENT,
   UV_ENOSYS,
-  UV_ESRCH
+  UV_ESRCH,
 } = internalBinding('uv');
 
 const { SocketListSend, SocketListReceive } = SocketList;
@@ -67,7 +80,9 @@ let freeParser;
 let HTTPParser;
 
 const MAX_HANDLE_RETRANSMISSIONS = 3;
+const kChannelHandle = Symbol('kChannelHandle');
 const kIsUsedAsStdio = Symbol('kIsUsedAsStdio');
+const kPendingMessages = Symbol('kPendingMessages');
 
 // This object contain function to convert TCP objects to native handle objects
 // and back again.
@@ -81,7 +96,7 @@ const handleConversion = {
 
     got(message, handle, emit) {
       emit(handle);
-    }
+    },
   },
 
   'net.Server': {
@@ -96,7 +111,7 @@ const handleConversion = {
       server.listen(handle, () => {
         emit(server);
       });
-    }
+    },
   },
 
   'net.Socket': {
@@ -109,7 +124,7 @@ const handleConversion = {
         // The worker should keep track of the socket
         message.key = socket.server._connectionKey;
 
-        const firstTime = !this.channel.sockets.send[message.key];
+        const firstTime = !this[kChannelHandle].sockets.send[message.key];
         const socketList = getSocketList('send', this, message.key);
 
         // The server should no longer expose a .connection property
@@ -170,7 +185,7 @@ const handleConversion = {
       const socket = new net.Socket({
         handle: handle,
         readable: true,
-        writable: true
+        writable: true,
       });
 
       // If the socket was created by net.Server we will track the socket
@@ -179,12 +194,12 @@ const handleConversion = {
         // Add socket to connections list
         const socketList = getSocketList('got', this, message.key);
         socketList.add({
-          socket: socket
+          socket: socket,
         });
       }
 
       emit(socket);
-    }
+    },
   },
 
   'dgram.Native': {
@@ -196,7 +211,7 @@ const handleConversion = {
 
     got(message, handle, emit) {
       emit(handle);
-    }
+    },
   },
 
   'dgram.Socket': {
@@ -214,8 +229,8 @@ const handleConversion = {
       socket.bind(handle, () => {
         emit(socket);
       });
-    }
-  }
+    },
+  },
 };
 
 function stdioStringToArray(stdio, channel) {
@@ -223,19 +238,20 @@ function stdioStringToArray(stdio, channel) {
 
   switch (stdio) {
     case 'ignore':
-    case 'pipe': options.push(stdio, stdio, stdio); break;
-    case 'inherit': options.push(0, 1, 2); break;
+    case 'overlapped':
+    case 'pipe': ArrayPrototypePush(options, stdio, stdio, stdio); break;
+    case 'inherit': ArrayPrototypePush(options, 0, 1, 2); break;
     default:
-      throw new ERR_INVALID_OPT_VALUE('stdio', stdio);
+      throw new ERR_INVALID_ARG_VALUE('stdio', stdio);
   }
 
-  if (channel) options.push(channel);
+  if (channel) ArrayPrototypePush(options, channel);
 
   return options;
 }
 
 function ChildProcess() {
-  EventEmitter.call(this);
+  FunctionPrototypeCall(EventEmitter, this);
 
   this._closesNeeded = 1;
   this._closesGot = 0;
@@ -270,7 +286,7 @@ function ChildProcess() {
       if (this.spawnfile)
         err.path = this.spawnfile;
 
-      err.spawnargs = this.spawnargs.slice(1);
+      err.spawnargs = ArrayPrototypeSlice(this.spawnargs, 1);
       this.emit('error', err);
     } else {
       this.emit('exit', this.exitCode, this.signalCode);
@@ -301,9 +317,7 @@ function flushStdio(subprocess) {
     // TODO(addaleax): This doesn't necessarily account for all the ways in
     // which data can be read from a stream, e.g. being consumed on the
     // native layer directly as a StreamBase.
-    if (!stream || !stream.readable ||
-        stream._readableState.readableListening ||
-        stream[kIsUsedAsStdio]) {
+    if (!stream || !stream.readable || stream[kIsUsedAsStdio]) {
       continue;
     }
     stream.resume();
@@ -312,7 +326,7 @@ function flushStdio(subprocess) {
 
 
 function createSocket(pipe, readable) {
-  return net.Socket({ handle: pipe, readable, writable: !readable });
+  return net.Socket({ handle: pipe, readable });
 }
 
 
@@ -334,9 +348,7 @@ function closePendingHandle(target) {
 ChildProcess.prototype.spawn = function(options) {
   let i = 0;
 
-  if (options === null || typeof options !== 'object') {
-    throw new ERR_INVALID_ARG_TYPE('options', 'Object', options);
-  }
+  validateObject(options, 'options');
 
   // If no `stdio` option was given - use default
   let stdio = options.stdio || 'pipe';
@@ -347,38 +359,32 @@ ChildProcess.prototype.spawn = function(options) {
   const ipcFd = stdio.ipcFd;
   stdio = options.stdio = stdio.stdio;
 
-  if (options.serialization !== undefined &&
-      options.serialization !== 'json' &&
-      options.serialization !== 'advanced') {
-    throw new ERR_INVALID_OPT_VALUE('options.serialization',
-                                    options.serialization);
-  }
 
+  validateOneOf(options.serialization, 'options.serialization',
+                [undefined, 'json', 'advanced']);
   const serialization = options.serialization || 'json';
 
   if (ipc !== undefined) {
     // Let child process know about opened IPC channel
     if (options.envPairs === undefined)
       options.envPairs = [];
-    else if (!ArrayIsArray(options.envPairs)) {
-      throw new ERR_INVALID_ARG_TYPE('options.envPairs',
-                                     'Array',
-                                     options.envPairs);
-    }
+    else
+      validateArray(options.envPairs, 'options.envPairs');
 
-    options.envPairs.push(`NODE_CHANNEL_FD=${ipcFd}`);
-    options.envPairs.push(`NODE_CHANNEL_SERIALIZATION_MODE=${serialization}`);
+    ArrayPrototypePush(options.envPairs, `NODE_CHANNEL_FD=${ipcFd}`);
+    ArrayPrototypePush(options.envPairs,
+                       `NODE_CHANNEL_SERIALIZATION_MODE=${serialization}`);
   }
 
   validateString(options.file, 'options.file');
   this.spawnfile = options.file;
 
-  if (ArrayIsArray(options.args))
-    this.spawnargs = options.args;
-  else if (options.args === undefined)
+  if (options.args === undefined) {
     this.spawnargs = [];
-  else
-    throw new ERR_INVALID_ARG_TYPE('options.args', 'Array', options.args);
+  } else {
+    validateArray(options.args, 'options.args');
+    this.spawnargs = options.args;
+  }
 
   const err = this._handle.spawn(options);
 
@@ -406,6 +412,8 @@ ChildProcess.prototype.spawn = function(options) {
     this._handle.close();
     this._handle = null;
     throw errnoException(err, 'spawn');
+  } else {
+    process.nextTick(onSpawnNT, this);
   }
 
   this.pid = this._handle.pid;
@@ -433,8 +441,6 @@ ChildProcess.prototype.spawn = function(options) {
     }
 
     if (stream.handle) {
-      // When i === 0 - we're dealing with stdin
-      // (which is the only one writable pipe).
       stream.socket = createSocket(this.pid !== 0 ?
         stream.handle : null, i > 0);
 
@@ -457,7 +463,8 @@ ChildProcess.prototype.spawn = function(options) {
   this.stdio = [];
 
   for (i = 0; i < stdio.length; i++)
-    this.stdio.push(stdio[i].socket === undefined ? null : stdio[i].socket);
+    ArrayPrototypePush(this.stdio,
+                       stdio[i].socket === undefined ? null : stdio[i].socket);
 
   // Add .send() method and start listening for IPC data
   if (ipc !== undefined) setupChannel(this, ipc, serialization);
@@ -471,6 +478,11 @@ function onErrorNT(self, err) {
 }
 
 
+function onSpawnNT(self) {
+  self.emit('spawn');
+}
+
+
 ChildProcess.prototype.kill = function(sig) {
 
   const signal = sig === 0 ? sig :
@@ -498,6 +510,12 @@ ChildProcess.prototype.kill = function(sig) {
   return false;
 };
 
+ChildProcess.prototype[SymbolDispose] = function() {
+  if (!this.killed) {
+    this.kill();
+  }
+};
+
 
 ChildProcess.prototype.ref = function() {
   if (this._handle) this._handle.ref();
@@ -509,46 +527,78 @@ ChildProcess.prototype.unref = function() {
 };
 
 class Control extends EventEmitter {
+  #channel = null;
+  #refs = 0;
+  #refExplicitlySet = false;
+
   constructor(channel) {
     super();
-    this.channel = channel;
-    this.refs = 0;
+    this.#channel = channel;
+    this[kPendingMessages] = [];
   }
-  ref() {
-    if (++this.refs === 1) {
-      this.channel.ref();
+
+  // The methods keeping track of the counter are being used to track the
+  // listener count on the child process object as well as when writes are
+  // in progress. Once the user has explicitly requested a certain state, these
+  // methods become no-ops in order to not interfere with the user's intentions.
+  refCounted() {
+    if (++this.#refs === 1 && !this.#refExplicitlySet) {
+      this.#channel.ref();
     }
   }
-  unref() {
-    if (--this.refs === 0) {
-      this.channel.unref();
+
+  unrefCounted() {
+    if (--this.#refs === 0 && !this.#refExplicitlySet) {
+      this.#channel.unref();
       this.emit('unref');
     }
   }
+
+  ref() {
+    this.#refExplicitlySet = true;
+    this.#channel.ref();
+  }
+
+  unref() {
+    this.#refExplicitlySet = true;
+    this.#channel.unref();
+  }
+
+  get fd() {
+    return this.#channel ? this.#channel.fd : undefined;
+  }
 }
 
+const channelDeprecationMsg = '_channel is deprecated. ' +
+                              'Use ChildProcess.channel instead.';
+
 let serialization;
 function setupChannel(target, channel, serializationMode) {
-  target.channel = channel;
+  const control = new Control(channel);
+  target.channel = control;
+  target[kChannelHandle] = channel;
 
-  // _channel can be deprecated in version 8
   ObjectDefineProperty(target, '_channel', {
-    get() { return target.channel; },
-    set(val) { target.channel = val; },
-    enumerable: true
+    __proto__: null,
+    get: deprecate(() => {
+      return target.channel;
+    }, channelDeprecationMsg, 'DEP0129'),
+    set: deprecate((val) => {
+      target.channel = val;
+    }, channelDeprecationMsg, 'DEP0129'),
+    configurable: true,
+    enumerable: false,
   });
 
   target._handleQueue = null;
   target._pendingMessage = null;
 
-  const control = new Control(channel);
-
   if (serialization === undefined)
     serialization = require('internal/child_process/serialization');
   const {
     initMessageChannel,
     parseChannelMessages,
-    writeChannelMessage
+    writeChannelMessage,
   } = serialization[serializationMode];
 
   let pendingHandle = null;
@@ -659,6 +709,24 @@ function setupChannel(target, channel, serializationMode) {
     });
   });
 
+  target.on('newListener', function() {
+
+    process.nextTick(() => {
+      if (!target.channel || !target.listenerCount('message'))
+        return;
+
+      const messages = target.channel[kPendingMessages];
+      const { length } = messages;
+      if (!length) return;
+
+      for (let i = 0; i < length; i++) {
+        ReflectApply(target.emit, target, messages[i]);
+      }
+
+      target.channel[kPendingMessages] = [];
+    });
+  });
+
   target.send = function(message, handle, options, callback) {
     if (typeof handle === 'function') {
       callback = handle;
@@ -667,9 +735,8 @@ function setupChannel(target, channel, serializationMode) {
     } else if (typeof options === 'function') {
       callback = options;
       options = undefined;
-    } else if (options !== undefined &&
-               (options === null || typeof options !== 'object')) {
-      throw new ERR_INVALID_ARG_TYPE('options', 'Object', options);
+    } else if (options !== undefined) {
+      validateObject(options, 'options');
     }
 
     options = { swallowErrors: false, ...options };
@@ -695,7 +762,7 @@ function setupChannel(target, channel, serializationMode) {
     // Non-serializable messages should not reach the remote
     // end point; as any failure in the stringification there
     // will result in error message that is weakly consumable.
-    // So perform a sanity check on message prior to sending.
+    // So perform a final check on message prior to sending.
     if (typeof message !== 'string' &&
         typeof message !== 'object' &&
         typeof message !== 'number' &&
@@ -717,7 +784,7 @@ function setupChannel(target, channel, serializationMode) {
       message = {
         cmd: 'NODE_HANDLE',
         type: null,
-        msg: message
+        msg: message,
       };
 
       if (handle instanceof net.Socket) {
@@ -736,7 +803,7 @@ function setupChannel(target, channel, serializationMode) {
 
       // Queue-up message and handle if we haven't received ACK yet.
       if (this._handleQueue) {
-        this._handleQueue.push({
+        ArrayPrototypePush(this._handleQueue, {
           callback: callback,
           handle: handle,
           options: options,
@@ -748,10 +815,8 @@ function setupChannel(target, channel, serializationMode) {
       obj = handleConversion[message.type];
 
       // convert TCP object to native handle object
-      handle = handleConversion[message.type].send.call(target,
-                                                        message,
-                                                        handle,
-                                                        options);
+      handle = ReflectApply(handleConversion[message.type].send,
+                            target, [message, handle, options]);
 
       // If handle was sent twice, or it is impossible to get native handle
       // out of it - just send a text without the handle.
@@ -766,7 +831,7 @@ function setupChannel(target, channel, serializationMode) {
                !(message && (message.cmd === 'NODE_HANDLE_ACK' ||
                              message.cmd === 'NODE_HANDLE_NACK'))) {
       // Queue request anyway to avoid out-of-order messages.
-      this._handleQueue.push({
+      ArrayPrototypePush(this._handleQueue, {
         callback: callback,
         handle: null,
         options: options,
@@ -790,11 +855,11 @@ function setupChannel(target, channel, serializationMode) {
 
       if (wasAsyncWrite) {
         req.oncomplete = () => {
-          control.unref();
+          control.unrefCounted();
           if (typeof callback === 'function')
             callback(null);
         };
-        control.ref();
+        control.refCounted();
       } else if (typeof callback === 'function') {
         process.nextTick(callback, null);
       }
@@ -813,7 +878,7 @@ function setupChannel(target, channel, serializationMode) {
       }
     }
 
-    /* If the master is > 2 read() calls behind, please stop sending. */
+    /* If the primary is > 2 read() calls behind, please stop sending. */
     return channel.writeQueueSize < (65536 * 2);
   };
 
@@ -849,6 +914,7 @@ function setupChannel(target, channel, serializationMode) {
 
     // This marks the fact that the channel is actually disconnected.
     this.channel = null;
+    this[kChannelHandle] = null;
 
     if (this._pendingMessage)
       closePendingHandle(this);
@@ -874,7 +940,15 @@ function setupChannel(target, channel, serializationMode) {
   };
 
   function emit(event, message, handle) {
-    target.emit(event, message, handle);
+    if ('internalMessage' === event || target.listenerCount('message')) {
+      target.emit(event, message, handle);
+      return;
+    }
+
+    ArrayPrototypePush(
+      target.channel[kPendingMessages],
+      [event, message, handle],
+    );
   }
 
   function handleMessage(message, handle, internal) {
@@ -896,10 +970,11 @@ function isInternal(message) {
           typeof message === 'object' &&
           typeof message.cmd === 'string' &&
           message.cmd.length > INTERNAL_PREFIX.length &&
-          message.cmd.slice(0, INTERNAL_PREFIX.length) === INTERNAL_PREFIX);
+          StringPrototypeSlice(message.cmd, 0, INTERNAL_PREFIX.length) ===
+            INTERNAL_PREFIX);
 }
 
-function nop() { }
+const nop = FunctionPrototype;
 
 function getValidStdio(stdio, sync) {
   let ipc;
@@ -909,18 +984,18 @@ function getValidStdio(stdio, sync) {
   if (typeof stdio === 'string') {
     stdio = stdioStringToArray(stdio);
   } else if (!ArrayIsArray(stdio)) {
-    throw new ERR_INVALID_OPT_VALUE('stdio', stdio);
+    throw new ERR_INVALID_ARG_VALUE('stdio', stdio);
   }
 
   // At least 3 stdio will be created
   // Don't concat() a new Array() because it would be sparse, and
   // stdio.reduce() would skip the sparse elements of stdio.
   // See https://stackoverflow.com/a/5501711/3561
-  while (stdio.length < 3) stdio.push(undefined);
+  while (stdio.length < 3) ArrayPrototypePush(stdio, undefined);
 
   // Translate stdio into C++-readable form
   // (i.e. PipeWraps or fds)
-  stdio = stdio.reduce((acc, stdio, i) => {
+  stdio = ArrayPrototypeReduce(stdio, (acc, stdio, i) => {
     function cleanup() {
       for (let i = 0; i < acc.length; i++) {
         if ((acc[i].type === 'pipe' || acc[i].type === 'ipc') && acc[i].handle)
@@ -934,18 +1009,19 @@ function getValidStdio(stdio, sync) {
     }
 
     if (stdio === 'ignore') {
-      acc.push({ type: 'ignore' });
-    } else if (stdio === 'pipe' || (typeof stdio === 'number' && stdio < 0)) {
+      ArrayPrototypePush(acc, { type: 'ignore' });
+    } else if (stdio === 'pipe' || stdio === 'overlapped' ||
+               (typeof stdio === 'number' && stdio < 0)) {
       const a = {
-        type: 'pipe',
+        type: stdio === 'overlapped' ? 'overlapped' : 'pipe',
         readable: i === 0,
-        writable: i !== 0
+        writable: i !== 0,
       };
 
       if (!sync)
         a.handle = new Pipe(PipeConstants.SOCKET);
 
-      acc.push(a);
+      ArrayPrototypePush(acc, a);
     } else if (stdio === 'ipc') {
       if (sync || ipc !== undefined) {
         // Cleanup previously created pipes
@@ -959,20 +1035,20 @@ function getValidStdio(stdio, sync) {
       ipc = new Pipe(PipeConstants.IPC);
       ipcFd = i;
 
-      acc.push({
+      ArrayPrototypePush(acc, {
         type: 'pipe',
         handle: ipc,
-        ipc: true
+        ipc: true,
       });
     } else if (stdio === 'inherit') {
-      acc.push({
+      ArrayPrototypePush(acc, {
         type: 'inherit',
-        fd: i
+        fd: i,
       });
     } else if (typeof stdio === 'number' || typeof stdio.fd === 'number') {
-      acc.push({
+      ArrayPrototypePush(acc, {
         type: 'fd',
-        fd: typeof stdio === 'number' ? stdio : stdio.fd
+        fd: typeof stdio === 'number' ? stdio : stdio.fd,
       });
     } else if (getHandleWrapType(stdio) || getHandleWrapType(stdio.handle) ||
                getHandleWrapType(stdio._handle)) {
@@ -980,11 +1056,11 @@ function getValidStdio(stdio, sync) {
         stdio :
         getHandleWrapType(stdio.handle) ? stdio.handle : stdio._handle;
 
-      acc.push({
+      ArrayPrototypePush(acc, {
         type: 'wrap',
         wrapType: getHandleWrapType(handle),
         handle: handle,
-        _stdio: stdio
+        _stdio: stdio,
       });
     } else if (isArrayBufferView(stdio) || typeof stdio === 'string') {
       if (!sync) {
@@ -994,7 +1070,7 @@ function getValidStdio(stdio, sync) {
     } else {
       // Cleanup
       cleanup();
-      throw new ERR_INVALID_OPT_VALUE('stdio', stdio);
+      throw new ERR_INVALID_ARG_VALUE('stdio', stdio);
     }
 
     return acc;
@@ -1005,7 +1081,7 @@ function getValidStdio(stdio, sync) {
 
 
 function getSocketList(type, worker, key) {
-  const sockets = worker.channel.sockets[type];
+  const sockets = worker[kChannelHandle].sockets[type];
   let socketList = sockets[key];
   if (!socketList) {
     const Construct = type === 'send' ? SocketListSend : SocketListReceive;
@@ -1023,8 +1099,7 @@ function maybeClose(subprocess) {
   }
 }
 
-function spawnSync(opts) {
-  const options = opts.options;
+function spawnSync(options) {
   const result = spawn_sync.spawn(options);
 
   if (result.output && options.encoding && options.encoding !== 'buffer') {
@@ -1039,9 +1114,9 @@ function spawnSync(opts) {
   result.stderr = result.output && result.output[2];
 
   if (result.error) {
-    result.error = errnoException(result.error, 'spawnSync ' + opts.file);
-    result.error.path = opts.file;
-    result.error.spawnargs = opts.args.slice(1);
+    result.error = errnoException(result.error, 'spawnSync ' + options.file);
+    result.error.path = options.file;
+    result.error.spawnargs = ArrayPrototypeSlice(options.args, 1);
   }
 
   return result;
@@ -1049,8 +1124,9 @@ function spawnSync(opts) {
 
 module.exports = {
   ChildProcess,
+  kChannelHandle,
   setupChannel,
   getValidStdio,
   stdioStringToArray,
-  spawnSync
+  spawnSync,
 };
