@@ -13,6 +13,7 @@
 // in the global console prototype chain anymore.
 
 const {
+  FunctionPrototypeBind,
   ObjectCreate,
   ReflectDefineProperty,
   ReflectGetOwnPropertyDescriptor,
@@ -21,8 +22,6 @@ const {
 
 const {
   Console,
-  kBindStreamsLazy,
-  kBindProperties
 } = require('internal/console/constructor');
 
 const globalConsole = ObjectCreate({});
@@ -37,15 +36,12 @@ for (const prop of ReflectOwnKeys(Console.prototype)) {
   const desc = ReflectGetOwnPropertyDescriptor(Console.prototype, prop);
   if (typeof desc.value === 'function') { // fix the receiver
     const name = desc.value.name;
-    desc.value = desc.value.bind(globalConsole);
-    ReflectDefineProperty(desc.value, 'name', { value: name });
+    desc.value = FunctionPrototypeBind(desc.value, globalConsole);
+    ReflectDefineProperty(desc.value, 'name', { __proto__: null, value: name });
   }
   ReflectDefineProperty(globalConsole, prop, desc);
 }
 
-globalConsole[kBindStreamsLazy](process);
-globalConsole[kBindProperties](true, 'auto');
-
 // This is a legacy feature - the Console constructor is exposed on
 // the global console instance.
 globalConsole.Console = Console;
