@@ -19,7 +19,6 @@ const {
   ArrayPrototypeIndexOf,
   ArrayPrototypeJoin,
   ArrayPrototypeMap,
-  ArrayPrototypePop,
   ArrayPrototypePush,
   ArrayPrototypeSlice,
   ArrayPrototypeSplice,
@@ -191,6 +190,12 @@ function lazyInternalUtilInspect() {
   return internalUtilInspect;
 }
 
+let utilColors;
+function lazyUtilColors() {
+  utilColors ??= require('internal/util/colors');
+  return utilColors;
+}
+
 let buffer;
 function lazyBuffer() {
   buffer ??= require('buffer').Buffer;
@@ -200,7 +205,7 @@ function lazyBuffer() {
 function isErrorStackTraceLimitWritable() {
   // Do no touch Error.stackTraceLimit as V8 would attempt to install
   // it again during deserialization.
-  if (require('v8').startupSnapshot.isBuildingSnapshot()) {
+  if (require('internal/v8/startup_snapshot').namespace.isBuildingSnapshot()) {
     return false;
   }
 
@@ -322,7 +327,7 @@ class SystemError extends Error {
             lazyBuffer().from(value.toString()) : undefined;
         },
         enumerable: true,
-        configurable: true
+        configurable: true,
       });
     }
 
@@ -338,7 +343,7 @@ class SystemError extends Error {
             lazyBuffer().from(value.toString()) : undefined;
         },
         enumerable: true,
-        configurable: true
+        configurable: true,
       });
     }
   }
@@ -351,7 +356,7 @@ class SystemError extends Error {
     return lazyInternalUtilInspect().inspect(this, {
       ...ctx,
       getters: true,
-      customInspect: false
+      customInspect: false,
     });
   }
 }
@@ -446,7 +451,7 @@ function getMessage(key, args, self) {
     assert(
       msg.length <= args.length, // Default options do not count.
       `Code: ${key}; The provided arguments length (${args.length}) does not ` +
-        `match the required ones (${msg.length}).`
+        `match the required ones (${msg.length}).`,
     );
     return ReflectApply(msg, self, args);
   }
@@ -457,7 +462,7 @@ function getMessage(key, args, self) {
   assert(
     expectedLength === args.length,
     `Code: ${key}; The provided arguments length (${args.length}) does not ` +
-      `match the required ones (${expectedLength}).`
+      `match the required ones (${expectedLength}).`,
   );
   if (args.length === 0)
     return msg;
@@ -500,7 +505,6 @@ const captureLargerStackTrace = hideStackFrames(
  * function UVException using a context object with data assembled in C++.
  * The goal is to migrate them to ERR_* errors later when compatibility is
  * not a concern.
- *
  * @param {object} ctx
  * @returns {Error}
  */
@@ -552,7 +556,6 @@ const uvException = hideStackFrames(function uvException(ctx) {
  * This creates an error compatible with errors produced in the C++
  * This function should replace the deprecated
  * `exceptionWithHostPort()` function.
- *
  * @param {number} err - A libuv error number
  * @param {string} syscall
  * @param {string} address
@@ -592,7 +595,6 @@ const uvExceptionWithHostPort = hideStackFrames(
 
 /**
  * This used to be util._errnoException().
- *
  * @param {number} err - A libuv error number
  * @param {string} syscall
  * @param {string} [original]
@@ -726,7 +728,6 @@ let maxStack_ErrorMessage;
  * Returns true if `err.name` and `err.message` are equal to engine-specific
  * values indicating max call stack size has been exceeded.
  * "Maximum call stack size exceeded" in V8.
- *
  * @param {Error} err
  * @returns {boolean}
  */
@@ -786,7 +787,7 @@ const fatalExceptionStackEnhancers = {
     // However, fatal error are handled differently and we cannot easily
     // highlight them. On Windows, detecting whether a console supports
     // ANSI escape sequences is not reliable.
-    if (process.platform === 'win32') {
+    if (isWindows) {
       const info = internalBinding('os').getOSInformation();
       const ver = ArrayPrototypeMap(StringPrototypeSplit(info[2], '.'),
                                     Number);
@@ -797,29 +798,26 @@ const fatalExceptionStackEnhancers = {
     const {
       inspect,
       inspectDefaultOptions: {
-        colors: defaultColors
-      }
+        colors: defaultColors,
+      },
     } = lazyInternalUtilInspect();
-    const colors = useColors &&
-                   ((internalBinding('util').guessHandleType(2) === 'TTY' &&
-                   require('internal/tty').hasColors()) ||
-                   defaultColors);
+    const colors = useColors && (lazyUtilColors().shouldColorize(process.stderr) || defaultColors);
     try {
       return inspect(error, {
         colors,
         customInspect: false,
-        depth: MathMax(inspect.defaultOptions.depth, 5)
+        depth: MathMax(inspect.defaultOptions.depth, 5),
       });
     } catch {
       return originalStack;
     }
-  }
+  },
 };
 
 const {
   privateSymbols: {
     arrow_message_private_symbol,
-  }
+  },
 } = internalBinding('util');
 // Ensures the printed error line is from user code.
 function setArrowMessage(err, arrowMessage) {
@@ -834,7 +832,7 @@ function hideInternalStackFrames(error) {
       frames = ArrayPrototypeFilter(
         stackFrames,
         (frm) => !StringPrototypeStartsWith(frm.getFileName() || '',
-                                            'node:internal')
+                                            'node:internal'),
       );
     }
     ArrayPrototypeUnshift(frames, error);
@@ -858,7 +856,6 @@ class AbortError extends Error {
 
 /**
  * This creates a generic Node.js error.
- *
  * @param {string} message The error message.
  * @param {object} errorProperties Object with additional properties to be added to the error.
  * @returns {Error}
@@ -895,6 +892,19 @@ function determineSpecificType(value) {
   return `type ${typeof value} (${inspected})`;
 }
 
+/**
+ * Create a list string in the form like 'A and B' or 'A, B, ..., and Z'.
+ * We cannot use Intl.ListFormat because it's not available in
+ * --without-intl builds.
+ * @param {string[]} array An array of strings.
+ * @param {string} [type] The list type to be inserted before the last element.
+ * @returns {string}
+ */
+function formatList(array, type = 'and') {
+  return array.length < 3 ? ArrayPrototypeJoin(array, ` ${type} `) :
+    `${ArrayPrototypeJoin(ArrayPrototypeSlice(array, 0, -1), ', ')}, ${type} ${array[array.length - 1]}`;
+}
+
 module.exports = {
   AbortError,
   aggregateTwoErrors,
@@ -909,6 +919,7 @@ module.exports = {
   errnoException,
   exceptionWithHostPort,
   fatalExceptionStackEnhancers,
+  formatList,
   genericNodeError,
   getMessage,
   hideInternalStackFrames,
@@ -946,11 +957,12 @@ module.exports = {
 //
 // Note: Node.js specific errors must begin with the prefix ERR_
 
+E('ERR_ACCESS_DENIED',
+  'Access to this API has been restricted. Permission: %s',
+  Error);
 E('ERR_AMBIGUOUS_ARGUMENT', 'The "%s" argument is ambiguous. %s', TypeError);
 E('ERR_ARG_NOT_ITERABLE', '%s must be iterable', TypeError);
 E('ERR_ASSERTION', '%s', Error);
-E('ERR_ASSERT_SNAPSHOT_NOT_SUPPORTED',
-  'Snapshot is not supported in this context ', TypeError);
 E('ERR_ASYNC_CALLBACK', '%s must be a function', TypeError);
 E('ERR_ASYNC_TYPE', 'Invalid name for async "type": %s', TypeError);
 E('ERR_BROTLI_INVALID_PARAM', '%s is not a valid Brotli parameter', RangeError);
@@ -1148,6 +1160,8 @@ E('ERR_HTTP2_TRAILERS_NOT_READY',
   'Trailing headers cannot be sent until after the wantTrailers event is ' +
   'emitted', Error);
 E('ERR_HTTP2_UNSUPPORTED_PROTOCOL', 'protocol "%s" is unsupported.', Error);
+E('ERR_HTTP_BODY_NOT_ALLOWED',
+  'Adding content for this request method or response status is not allowed.', Error);
 E('ERR_HTTP_CONTENT_LENGTH_MISMATCH',
   'Response body\'s content-length of %s byte(s) does not match the content-length of %s byte(s) set in header', Error);
 E('ERR_HTTP_HEADERS_SENT',
@@ -1156,17 +1170,24 @@ E('ERR_HTTP_INVALID_HEADER_VALUE',
   'Invalid value "%s" for header "%s"', TypeError);
 E('ERR_HTTP_INVALID_STATUS_CODE', 'Invalid status code: %s', RangeError);
 E('ERR_HTTP_REQUEST_TIMEOUT', 'Request timeout', Error);
+E('ERR_HTTP_SOCKET_ASSIGNED',
+  'ServerResponse has an already assigned socket', Error);
 E('ERR_HTTP_SOCKET_ENCODING',
   'Changing the socket encoding is not allowed per RFC7230 Section 3.', Error);
 E('ERR_HTTP_TRAILER_INVALID',
   'Trailers are invalid with this transfer encoding', Error);
 E('ERR_ILLEGAL_CONSTRUCTOR', 'Illegal constructor', TypeError);
+// TODO(aduh95): change the error to mention import attributes instead of import assertions.
 E('ERR_IMPORT_ASSERTION_TYPE_FAILED',
   'Module "%s" is not of type "%s"', TypeError);
+// TODO(aduh95): change the error to mention import attributes instead of import assertions.
 E('ERR_IMPORT_ASSERTION_TYPE_MISSING',
-  'Module "%s" needs an import assertion of type "%s"', TypeError);
+  'Module "%s" needs an import attribute of type "%s"', TypeError);
+// TODO(aduh95): change the error to mention import attributes instead of import assertions.
 E('ERR_IMPORT_ASSERTION_TYPE_UNSUPPORTED',
-  'Import assertion type "%s" is unsupported', TypeError);
+  'Import attribute type "%s" is unsupported', TypeError);
+E('ERR_IMPORT_ATTRIBUTE_UNSUPPORTED',
+  'Import attribute "%s" with value "%s" is not supported', TypeError);
 E('ERR_INCOMPATIBLE_OPTION_PAIR',
   'Option "%s" cannot be used in combination with option "%s"', TypeError);
 E('ERR_INPUT_TYPE_NOT_ALLOWED', '--input-type can only be used with string ' +
@@ -1240,39 +1261,20 @@ E('ERR_INVALID_ARG_TYPE',
     }
 
     if (types.length > 0) {
-      if (types.length > 2) {
-        const last = ArrayPrototypePop(types);
-        msg += `one of type ${ArrayPrototypeJoin(types, ', ')}, or ${last}`;
-      } else if (types.length === 2) {
-        msg += `one of type ${types[0]} or ${types[1]}`;
-      } else {
-        msg += `of type ${types[0]}`;
-      }
+      msg += `${types.length > 1 ? 'one of type' : 'of type'} ${formatList(types, 'or')}`;
       if (instances.length > 0 || other.length > 0)
         msg += ' or ';
     }
 
     if (instances.length > 0) {
-      if (instances.length > 2) {
-        const last = ArrayPrototypePop(instances);
-        msg +=
-          `an instance of ${ArrayPrototypeJoin(instances, ', ')}, or ${last}`;
-      } else {
-        msg += `an instance of ${instances[0]}`;
-        if (instances.length === 2) {
-          msg += ` or ${instances[1]}`;
-        }
-      }
+      msg += `an instance of ${formatList(instances, 'or')}`;
       if (other.length > 0)
         msg += ' or ';
     }
 
     if (other.length > 0) {
-      if (other.length > 2) {
-        const last = ArrayPrototypePop(other);
-        msg += `one of ${ArrayPrototypeJoin(other, ', ')}, or ${last}`;
-      } else if (other.length === 2) {
-        msg += `one of ${other[0]} or ${other[1]}`;
+      if (other.length > 1) {
+        msg += `one of ${formatList(other, 'or')}`;
       } else {
         if (StringPrototypeToLowerCase(other[0]) !== other[0])
           msg += 'an ';
@@ -1403,7 +1405,7 @@ E(
   '"%s" did not call the next hook in its chain and did not' +
   ' explicitly signal a short circuit. If this is intentional, include' +
   ' `shortCircuit: true` in the hook\'s return.',
-  Error
+  Error,
 );
 E('ERR_MANIFEST_ASSERT_INTEGRITY',
   (moduleURL, realIntegrities) => {
@@ -1414,7 +1416,7 @@ E('ERR_MANIFEST_ASSERT_INTEGRITY',
       const sri = ArrayPrototypeJoin(
         ArrayFrom(realIntegrities.entries(),
                   ({ 0: alg, 1: dgs }) => `${alg}-${dgs}`),
-        ' '
+        ' ',
       );
       msg += ` Integrities found are: ${sri}`;
     } else {
@@ -1450,25 +1452,18 @@ E('ERR_MISSING_ARGS',
       args,
       (a) => (ArrayIsArray(a) ?
         ArrayPrototypeJoin(ArrayPrototypeMap(a, wrap), ' or ') :
-        wrap(a))
+        wrap(a)),
     );
-    switch (len) {
-      case 1:
-        msg += `${args[0]} argument`;
-        break;
-      case 2:
-        msg += `${args[0]} and ${args[1]} arguments`;
-        break;
-      default:
-        msg += ArrayPrototypeJoin(ArrayPrototypeSlice(args, 0, len - 1), ', ');
-        msg += `, and ${args[len - 1]} arguments`;
-        break;
-    }
+    msg += `${formatList(args)} argument${len > 1 ? 's' : ''}`;
     return `${msg} must be specified`;
   }, TypeError);
 E('ERR_MISSING_OPTION', '%s is required', TypeError);
-E('ERR_MODULE_NOT_FOUND', (path, base, type = 'package') => {
-  return `Cannot find ${type} '${path}' imported from ${base}`;
+E('ERR_MODULE_NOT_FOUND', function(path, base, exactUrl) {
+  if (exactUrl) {
+    lazyInternalUtil().setOwnProperty(this, 'url', `${exactUrl}`);
+  }
+  return `Cannot find ${
+    exactUrl ? 'module' : 'package'} '${path}' imported from ${base}`;
 }, Error);
 E('ERR_MULTIPLE_CALLBACK', 'Callback called multiple times', Error);
 E('ERR_NAPI_CONS_FUNCTION', 'Constructor must be a function', TypeError);
@@ -1556,7 +1551,7 @@ E('ERR_REQUIRE_ESM',
     msg += `\n${basename} is treated as an ES module file as it is a .js ` +
       'file whose nearest parent package.json contains "type": "module" ' +
       'which declares all .js files in that package scope as ES modules.' +
-      `\nInstead rename ${basename} to end in .cjs, change the requiring ` +
+      `\nInstead either rename ${basename} to end in .cjs, change the requiring ` +
       'code to use dynamic import() which is available in all CommonJS ' +
       'modules, or change "type": "module" to "type": "commonjs" in ' +
       `${packageJsonPath} to treat all .js files as CommonJS (using .mjs for ` +
@@ -1583,6 +1578,9 @@ E('ERR_SOCKET_BUFFER_SIZE',
   'Could not get or set buffer size',
   SystemError);
 E('ERR_SOCKET_CLOSED', 'Socket is closed', Error);
+E('ERR_SOCKET_CLOSED_BEFORE_CONNECTION',
+  'Socket closed before the connection was established',
+  Error);
 E('ERR_SOCKET_DGRAM_IS_CONNECTED', 'Already connected', Error);
 E('ERR_SOCKET_DGRAM_NOT_CONNECTED', 'Not connected', Error);
 E('ERR_SOCKET_DGRAM_NOT_RUNNING', 'Not running', Error);
@@ -1620,8 +1618,8 @@ E('ERR_TAP_VALIDATION_ERROR', function(errorMsg) {
 }, Error);
 E('ERR_TEST_FAILURE', function(error, failureType) {
   hideInternalStackFrames(this);
-  assert(typeof failureType === 'string',
-         "The 'failureType' argument must be of type string.");
+  assert(typeof failureType === 'string' || typeof failureType === 'symbol',
+         "The 'failureType' argument must be of type string or symbol.");
 
   let msg = error?.message ?? error;
 
@@ -1633,6 +1631,16 @@ E('ERR_TEST_FAILURE', function(error, failureType) {
   this.cause = error;
   return msg;
 }, Error);
+E('ERR_TLS_ALPN_CALLBACK_INVALID_RESULT', (value, protocols) => {
+  return `ALPN callback returned a value (${
+    value
+  }) that did not match any of the client's offered protocols (${
+    protocols.join(', ')
+  })`;
+}, TypeError);
+E('ERR_TLS_ALPN_CALLBACK_WITH_PROTOCOLS',
+  'The ALPNCallback and ALPNProtocols TLS options are mutually exclusive',
+  TypeError);
 E('ERR_TLS_CERT_ALTNAME_FORMAT', 'Invalid subject alternative name string',
   SyntaxError);
 E('ERR_TLS_CERT_ALTNAME_INVALID', function(reason, host, cert) {
@@ -1683,20 +1691,17 @@ E('ERR_UNHANDLED_ERROR',
 E('ERR_UNKNOWN_BUILTIN_MODULE', 'No such built-in module: %s', Error);
 E('ERR_UNKNOWN_CREDENTIAL', '%s identifier does not exist: %s', Error);
 E('ERR_UNKNOWN_ENCODING', 'Unknown encoding: %s', TypeError);
-E('ERR_UNKNOWN_FILE_EXTENSION', (ext, path, suggestion) => {
-  let msg = `Unknown file extension "${ext}" for ${path}`;
-  if (suggestion) {
-    msg += `. ${suggestion}`;
-  }
-  return msg;
-}, TypeError);
+E('ERR_UNKNOWN_FILE_EXTENSION', 'Unknown file extension "%s" for %s', TypeError);
 E('ERR_UNKNOWN_MODULE_FORMAT', 'Unknown module format: %s for URL %s',
   RangeError);
 E('ERR_UNKNOWN_SIGNAL', 'Unknown signal: %s', TypeError);
-E('ERR_UNSUPPORTED_DIR_IMPORT', "Directory import '%s' is not supported " +
-'resolving ES modules imported from %s', Error);
+E('ERR_UNSUPPORTED_DIR_IMPORT', function(path, base, exactUrl) {
+  lazyInternalUtil().setOwnProperty(this, 'url', exactUrl);
+  return `Directory import '${path}' is not supported ` +
+    `resolving ES modules imported from ${base}`;
+}, Error);
 E('ERR_UNSUPPORTED_ESM_URL_SCHEME', (url, supported) => {
-  let msg = `Only URLs with a scheme in: ${ArrayPrototypeJoin(supported, ', ')} are supported by the default ESM loader`;
+  let msg = `Only URLs with a scheme in: ${formatList(supported)} are supported by the default ESM loader`;
   if (isWindows && url.protocol.length === 2) {
     msg +=
       '. On Windows, absolute paths must be valid file:// URLs';
@@ -1711,6 +1716,9 @@ E('ERR_VALID_PERFORMANCE_ENTRY_TYPE',
   'At least one valid performance entry type is required', Error);
 E('ERR_VM_DYNAMIC_IMPORT_CALLBACK_MISSING',
   'A dynamic import callback was not specified.', TypeError);
+E('ERR_VM_DYNAMIC_IMPORT_CALLBACK_MISSING_FLAG',
+  'A dynamic import callback was invoked without --experimental-vm-modules',
+  TypeError);
 E('ERR_VM_MODULE_ALREADY_LINKED', 'Module has already been linked', Error);
 E('ERR_VM_MODULE_CANNOT_CREATE_CACHED_DATA',
   'Cached data cannot be created for a module which has been evaluated', Error);
