@@ -32,7 +32,7 @@ const {
     ERR_EVENT_RECURSION,
     ERR_MISSING_ARGS,
     ERR_INVALID_THIS,
-  }
+  },
 } = require('internal/errors');
 const { validateObject, validateString } = require('internal/validators');
 
@@ -42,6 +42,7 @@ const {
   kEnumerableProperty,
 } = require('internal/util');
 const { inspect } = require('util');
+const webidl = require('internal/webidl');
 
 const kIsEventTarget = SymbolFor('nodejs.event_target');
 const kIsNodeEventTarget = Symbol('kIsNodeEventTarget');
@@ -56,10 +57,12 @@ const kEvents = Symbol('kEvents');
 const kIsBeingDispatched = Symbol('kIsBeingDispatched');
 const kStop = Symbol('kStop');
 const kTarget = Symbol('kTarget');
-const kHandlers = Symbol('khandlers');
+const kHandlers = Symbol('kHandlers');
 const kWeakHandler = Symbol('kWeak');
+const kResistStopPropagation = Symbol('kResistStopPropagation');
 
 const kHybridDispatch = SymbolFor('nodejs.internal.kHybridDispatch');
+const kRemoveWeakListenerHelper = Symbol('nodejs.internal.removeWeakListenerHelper');
 const kCreateEvent = Symbol('kCreateEvent');
 const kNewListener = Symbol('kNewListener');
 const kRemoveListener = Symbol('kRemoveListener');
@@ -75,9 +78,16 @@ const isTrustedSet = new SafeWeakSet();
 const isTrusted = ObjectGetOwnPropertyDescriptor({
   get isTrusted() {
     return isTrustedSet.has(this);
-  }
+  },
 }, 'isTrusted').get;
 
+const isTrustedDescriptor = {
+  __proto__: null,
+  configurable: false,
+  enumerable: true,
+  get: isTrusted,
+};
+
 function isEvent(value) {
   return typeof value?.[kType] === 'string';
 }
@@ -112,13 +122,6 @@ class Event {
       isTrustedSet.add(this);
     }
 
-    // isTrusted is special (LegacyUnforgeable)
-    ObjectDefineProperty(this, 'isTrusted', {
-      __proto__: null,
-      get: isTrusted,
-      enumerable: true,
-      configurable: false
-    });
     this[kTarget] = null;
     this[kIsBeingDispatched] = false;
   }
@@ -131,7 +134,7 @@ class Event {
       return name;
 
     const opts = ObjectAssign({}, options, {
-      depth: NumberIsInteger(options.depth) ? options.depth - 1 : options.depth
+      depth: NumberIsInteger(options.depth) ? options.depth - 1 : options.depth,
     });
 
     return `${name} ${inspect({
@@ -236,7 +239,7 @@ class Event {
   get returnValue() {
     if (!isEvent(this))
       throw new ERR_INVALID_THIS('Event');
-    return !this.defaultPrevented;
+    return !this.#cancelable || !this.#defaultPrevented;
   }
 
   /**
@@ -323,6 +326,11 @@ ObjectDefineProperties(
     eventPhase: kEnumerableProperty,
     cancelBubble: kEnumerableProperty,
     stopPropagation: kEnumerableProperty,
+    // Don't conform to the spec with isTrusted. The spec defines it as
+    // LegacyUnforgeable but defining it in the constructor has a big
+    // performance impact and the property doesn't seem to be useful outside of
+    // browsers.
+    isTrusted: isTrustedDescriptor,
   });
 
 function isCustomEvent(value) {
@@ -385,7 +393,7 @@ let weakListenersState = null;
 let objectToWeakListenerMap = null;
 function weakListeners() {
   weakListenersState ??= new SafeFinalizationRegistry(
-    (listener) => listener.remove()
+    ({ eventTarget, listener, eventType }) => eventTarget.deref()?.[kRemoveWeakListenerHelper](eventType, listener),
   );
   objectToWeakListenerMap ??= new SafeWeakMap();
   return { registry: weakListenersState, map: objectToWeakListenerMap };
@@ -397,6 +405,7 @@ const kFlagPassive = 1 << 2;
 const kFlagNodeStyle = 1 << 3;
 const kFlagWeak = 1 << 4;
 const kFlagRemoved = 1 << 5;
+const kFlagResistStopPropagation = 1 << 6;
 
 // The listeners for an EventTarget are maintained as a linked list.
 // Unfortunately, the way EventTarget is defined, listeners are accounted
@@ -406,8 +415,8 @@ const kFlagRemoved = 1 << 5;
 // the linked list makes dispatching faster, even if adding/removing is
 // slower.
 class Listener {
-  constructor(previous, listener, once, capture, passive,
-              isNodeStyleListener, weak) {
+  constructor(eventTarget, eventType, previous, listener, once, capture, passive,
+              isNodeStyleListener, weak, resistStopPropagation) {
     this.next = undefined;
     if (previous !== undefined)
       previous.next = this;
@@ -425,13 +434,21 @@ class Listener {
       flags |= kFlagNodeStyle;
     if (weak)
       flags |= kFlagWeak;
+    if (resistStopPropagation)
+      flags |= kFlagResistStopPropagation;
     this.flags = flags;
 
     this.removed = false;
 
     if (this.weak) {
       this.callback = new SafeWeakRef(listener);
-      weakListeners().registry.register(listener, this, this);
+      weakListeners().registry.register(listener, {
+        __proto__: null,
+        // Weak ref so the listener won't hold the eventTarget alive
+        eventTarget: new SafeWeakRef(eventTarget),
+        listener: this,
+        eventType,
+      }, this);
       // Make the retainer retain the listener in a WeakMap
       weakListeners().map.set(weak, listener);
       this.listener = this.callback;
@@ -462,6 +479,9 @@ class Listener {
   get weak() {
     return Boolean(this.flags & kFlagWeak);
   }
+  get resistStopPropagation() {
+    return Boolean(this.flags & kFlagResistStopPropagation);
+  }
   get removed() {
     return Boolean(this.flags & kFlagRemoved);
   }
@@ -558,6 +578,7 @@ class EventTarget {
       signal,
       isNodeStyleListener,
       weak,
+      resistStopPropagation,
     } = validateEventListenerOptions(options);
 
     if (!validateEventListener(listener)) {
@@ -572,7 +593,7 @@ class EventTarget {
       process.emitWarning(w);
       return;
     }
-    type = String(type);
+    type = webidl.converters.DOMString(type);
 
     if (signal) {
       if (signal.aborted) {
@@ -582,16 +603,16 @@ class EventTarget {
       // not prevent the event target from GC.
       signal.addEventListener('abort', () => {
         this.removeEventListener(type, listener, options);
-      }, { once: true, [kWeakHandler]: this });
+      }, { __proto__: null, once: true, [kWeakHandler]: this, [kResistStopPropagation]: true });
     }
 
     let root = this[kEvents].get(type);
 
     if (root === undefined) {
-      root = { size: 1, next: undefined };
+      root = { size: 1, next: undefined, resistStopPropagation: Boolean(resistStopPropagation) };
       // This is the first handler in our linked list.
-      new Listener(root, listener, once, capture, passive,
-                   isNodeStyleListener, weak);
+      new Listener(this, type, root, listener, once, capture, passive,
+                   isNodeStyleListener, weak, resistStopPropagation);
       this[kNewListener](
         root.size,
         type,
@@ -617,9 +638,10 @@ class EventTarget {
       return;
     }
 
-    new Listener(previous, listener, once, capture, passive,
-                 isNodeStyleListener, weak);
+    new Listener(this, type, previous, listener, once, capture, passive,
+                 isNodeStyleListener, weak, resistStopPropagation);
     root.size++;
+    root.resistStopPropagation ||= Boolean(resistStopPropagation);
     this[kNewListener](root.size, type, listener, once, capture, passive, weak);
   }
 
@@ -638,7 +660,7 @@ class EventTarget {
     if (!validateEventListener(listener))
       return;
 
-    type = String(type);
+    type = webidl.converters.DOMString(type);
     const capture = options?.capture === true;
 
     const root = this[kEvents].get(type);
@@ -659,6 +681,28 @@ class EventTarget {
     }
   }
 
+  [kRemoveWeakListenerHelper](type, listener) {
+    const root = this[kEvents].get(type);
+    if (root === undefined || root.next === undefined)
+      return;
+
+    const capture = listener.capture === true;
+
+    let handler = root.next;
+    while (handler !== undefined) {
+      if (handler === listener) {
+        handler.remove();
+        root.size--;
+        if (root.size === 0)
+          this[kEvents].delete(type);
+        // Undefined is passed as the listener as the listener was GCed
+        this[kRemoveListener](root.size, type, undefined, capture);
+        break;
+      }
+      handler = handler.next;
+    }
+  }
+
   /**
    * @param {Event} event
    */
@@ -703,14 +747,21 @@ class EventTarget {
     let handler = root.next;
     let next;
 
-    while (handler !== undefined &&
-           (handler.passive || event?.[kStop] !== true)) {
+    const iterationCondition = () => {
+      if (handler === undefined) {
+        return false;
+      }
+      return root.resistStopPropagation || handler.passive || event?.[kStop] !== true;
+    };
+    while (iterationCondition()) {
       // Cache the next item in case this iteration removes the current one
       next = handler.next;
 
-      if (handler.removed) {
+      if (handler.removed || (event?.[kStop] === true && !handler.resistStopPropagation)) {
         // Deal with the case an event is removed while event handlers are
         // Being processed (removeEventListener called from a listener)
+        // And the case of event.stopImmediatePropagation() being called
+        // For events not flagged as resistStopPropagation
         handler = next;
         continue;
       }
@@ -761,7 +812,7 @@ class EventTarget {
       return name;
 
     const opts = ObjectAssign({}, options, {
-      depth: NumberIsInteger(options.depth) ? options.depth - 1 : options.depth
+      depth: NumberIsInteger(options.depth) ? options.depth - 1 : options.depth,
     });
 
     return `${name} ${inspect({}, opts)}`;
@@ -778,7 +829,7 @@ ObjectDefineProperties(EventTarget.prototype, {
     enumerable: false,
     configurable: true,
     value: 'EventTarget',
-  }
+  },
 });
 
 function initNodeEventTarget(self) {
@@ -822,7 +873,7 @@ class NodeEventTarget extends EventTarget {
   }
 
   /**
-   * @param {string} [type]
+   * @param {string} type
    * @returns {number}
    */
   listenerCount(type) {
@@ -914,7 +965,7 @@ class NodeEventTarget extends EventTarget {
   }
 
   /**
-   * @param {string} type
+   * @param {string} [type]
    * @returns {NodeEventTarget}
    */
   removeAllListeners(type) {
@@ -978,7 +1029,8 @@ function validateEventListenerOptions(options) {
     passive: Boolean(options.passive),
     signal: options.signal,
     weak: options[kWeakHandler],
-    isNodeStyleListener: Boolean(options[kIsNodeStyleListener])
+    resistStopPropagation: options[kResistStopPropagation] ?? false,
+    isNodeStyleListener: Boolean(options[kIsNodeStyleListener]),
   };
 }
 
@@ -1054,7 +1106,7 @@ function defineEventHandler(emitter, name) {
       this[kHandlers].set(name, wrappedHandler);
     },
     configurable: true,
-    enumerable: true
+    enumerable: true,
   });
 }
 
@@ -1093,5 +1145,6 @@ module.exports = {
   kRemoveListener,
   kEvents,
   kWeakHandler,
+  kResistStopPropagation,
   isEventTarget,
 };
