@@ -2,15 +2,17 @@
 
 const {
   ObjectSetPrototypeOf,
+  ReflectApply,
 } = primordials;
+const { kEmptyObject } = require('internal/util');
 
 const { Writable } = require('stream');
 const { closeSync, writeSync } = require('fs');
 
 function SyncWriteStream(fd, options) {
-  Writable.call(this, { autoDestroy: true });
+  ReflectApply(Writable, this, [{ autoDestroy: true }]);
 
-  options = options || {};
+  options = options || kEmptyObject;
 
   this.fd = fd;
   this.readable = false;
@@ -21,9 +23,13 @@ ObjectSetPrototypeOf(SyncWriteStream.prototype, Writable.prototype);
 ObjectSetPrototypeOf(SyncWriteStream, Writable);
 
 SyncWriteStream.prototype._write = function(chunk, encoding, cb) {
-  writeSync(this.fd, chunk, 0, chunk.length);
+  try {
+    writeSync(this.fd, chunk, 0, chunk.length);
+  } catch (e) {
+    cb(e);
+    return;
+  }
   cb();
-  return true;
 };
 
 SyncWriteStream.prototype._destroy = function(err, cb) {
