@@ -161,9 +161,10 @@ function assertEncoding(encoding) {
 }
 
 class Dirent {
-  constructor(name, type, path) {
+  constructor(name, type, path, filepath = path && join(path, name)) {
     this.name = name;
-    this.path = path;
+    this.parentPath = path;
+    this.path = filepath;
     this[kType] = type;
   }
 
@@ -197,8 +198,8 @@ class Dirent {
 }
 
 class DirentFromStats extends Dirent {
-  constructor(name, stats, path) {
-    super(name, null, path);
+  constructor(name, stats, path, filepath) {
+    super(name, null, path, filepath);
     this[kStats] = stats;
   }
 }
@@ -233,7 +234,7 @@ function join(path, name) {
   }
 
   if (typeof path === 'string' && typeof name === 'string') {
-    return pathModule.basename(path) === name ? path : pathModule.join(path, name);
+    return pathModule.join(path, name);
   }
 
   if (isUint8Array(path) && isUint8Array(name)) {
@@ -268,7 +269,7 @@ function getDirents(path, { 0: names, 1: types }, callback) {
             callback(err);
             return;
           }
-          names[idx] = new DirentFromStats(name, stats, path);
+          names[idx] = new DirentFromStats(name, stats, path, filepath);
           if (--toFinish === 0) {
             callback(null, names);
           }
@@ -304,7 +305,7 @@ function getDirent(path, name, type, callback) {
           callback(err);
           return;
         }
-        callback(null, new DirentFromStats(name, stats, filepath));
+        callback(null, new DirentFromStats(name, stats, path, filepath));
       });
     } else {
       callback(null, new Dirent(name, type, path));
@@ -312,9 +313,13 @@ function getDirent(path, name, type, callback) {
   } else if (type === UV_DIRENT_UNKNOWN) {
     const filepath = join(path, name);
     const stats = lazyLoadFs().lstatSync(filepath);
-    return new DirentFromStats(name, stats, path);
-  } else {
+    // callback === true: Quirk to not introduce a breaking change.
+    return new DirentFromStats(name, stats, path, callback === true ? filepath : path);
+  } else if (callback === true) {
+    // callback === true: Quirk to not introduce a breaking change.
     return new Dirent(name, type, path);
+  } else {
+    return new Dirent(name, type, path, path);
   }
 }
 
