@@ -1,6 +1,6 @@
 'use strict';
 const {
-  Symbol
+  Symbol,
 } = primordials;
 const {
   kUpdateTimer,
@@ -37,5 +37,5 @@ class HeapSnapshotStream extends Readable {
 }
 
 module.exports = {
-  HeapSnapshotStream
+  HeapSnapshotStream,
 };
