@@ -2,12 +2,20 @@
 
 const {
   ArrayIsArray,
+  ArrayPrototypeIncludes,
+  ArrayPrototypeMap,
+  ArrayPrototypePush,
   Error,
   MathMax,
   Number,
   ObjectCreate,
+  ObjectDefineProperty,
   ObjectKeys,
-  Set,
+  SafeSet,
+  String,
+  StringFromCharCode,
+  StringPrototypeIncludes,
+  StringPrototypeToLowerCase,
   Symbol,
 } = primordials;
 
@@ -18,17 +26,23 @@ const {
     ERR_HTTP2_INVALID_CONNECTION_HEADERS,
     ERR_HTTP2_INVALID_PSEUDOHEADER,
     ERR_HTTP2_INVALID_SETTING_VALUE,
-    ERR_INVALID_ARG_TYPE
+    ERR_INVALID_ARG_TYPE,
+    ERR_INVALID_HTTP_TOKEN,
   },
-  addCodeToName,
-  hideStackFrames
+  captureLargerStackTrace,
+  getMessage,
+  hideStackFrames,
+  kIsNodeError,
 } = require('internal/errors');
 
+const kSensitiveHeaders = Symbol('nodejs.http2.sensitiveHeaders');
 const kSocket = Symbol('socket');
 const kProxySocket = Symbol('proxySocket');
 const kRequest = Symbol('request');
 
 const {
+  NGHTTP2_NV_FLAG_NONE,
+  NGHTTP2_NV_FLAG_NO_INDEX,
   NGHTTP2_SESSION_CLIENT,
   NGHTTP2_SESSION_SERVER,
 
@@ -56,6 +70,7 @@ const {
   HTTP2_HEADER_ETAG,
   HTTP2_HEADER_EXPIRES,
   HTTP2_HEADER_FROM,
+  HTTP2_HEADER_HOST,
   HTTP2_HEADER_IF_MATCH,
   HTTP2_HEADER_IF_NONE_MATCH,
   HTTP2_HEADER_IF_MODIFIED_SINCE,
@@ -79,30 +94,29 @@ const {
   HTTP2_HEADER_HTTP2_SETTINGS,
   HTTP2_HEADER_TE,
   HTTP2_HEADER_TRANSFER_ENCODING,
-  HTTP2_HEADER_HOST,
   HTTP2_HEADER_KEEP_ALIVE,
   HTTP2_HEADER_PROXY_CONNECTION,
 
   HTTP2_METHOD_DELETE,
   HTTP2_METHOD_GET,
-  HTTP2_METHOD_HEAD
+  HTTP2_METHOD_HEAD,
 } = binding.constants;
 
 // This set is defined strictly by the HTTP/2 specification. Only
 // :-prefixed headers defined by that specification may be added to
 // this set.
-const kValidPseudoHeaders = new Set([
+const kValidPseudoHeaders = new SafeSet([
   HTTP2_HEADER_STATUS,
   HTTP2_HEADER_METHOD,
   HTTP2_HEADER_AUTHORITY,
   HTTP2_HEADER_SCHEME,
   HTTP2_HEADER_PATH,
-  HTTP2_HEADER_PROTOCOL
+  HTTP2_HEADER_PROTOCOL,
 ]);
 
 // This set contains headers that are permitted to have only a single
 // value. Multiple instances must not be specified.
-const kSingleValueHeaders = new Set([
+const kSingleValueHeaders = new SafeSet([
   HTTP2_HEADER_STATUS,
   HTTP2_HEADER_METHOD,
   HTTP2_HEADER_AUTHORITY,
@@ -126,6 +140,7 @@ const kSingleValueHeaders = new Set([
   HTTP2_HEADER_ETAG,
   HTTP2_HEADER_EXPIRES,
   HTTP2_HEADER_FROM,
+  HTTP2_HEADER_HOST,
   HTTP2_HEADER_IF_MATCH,
   HTTP2_HEADER_IF_MODIFIED_SINCE,
   HTTP2_HEADER_IF_NONE_MATCH,
@@ -141,17 +156,17 @@ const kSingleValueHeaders = new Set([
   HTTP2_HEADER_TK,
   HTTP2_HEADER_UPGRADE_INSECURE_REQUESTS,
   HTTP2_HEADER_USER_AGENT,
-  HTTP2_HEADER_X_CONTENT_TYPE_OPTIONS
+  HTTP2_HEADER_X_CONTENT_TYPE_OPTIONS,
 ]);
 
 // The HTTP methods in this set are specifically defined as assigning no
 // meaning to the request payload. By default, unless the user explicitly
 // overrides the endStream option on the request method, the endStream
 // option will be defaulted to true when these methods are used.
-const kNoPayloadMethods = new Set([
+const kNoPayloadMethods = new SafeSet([
   HTTP2_METHOD_DELETE,
   HTTP2_METHOD_GET,
-  HTTP2_METHOD_HEAD
+  HTTP2_METHOD_HEAD,
 ]);
 
 // The following ArrayBuffer instances are used to share memory more efficiently
@@ -330,7 +345,7 @@ function getSettings(session, remote) {
     maxHeaderListSize: settingsBuffer[IDX_SETTINGS_MAX_HEADER_LIST_SIZE],
     maxHeaderSize: settingsBuffer[IDX_SETTINGS_MAX_HEADER_LIST_SIZE],
     enableConnectProtocol:
-      !!settingsBuffer[IDX_SETTINGS_ENABLE_CONNECT_PROTOCOL]
+      !!settingsBuffer[IDX_SETTINGS_ENABLE_CONNECT_PROTOCOL],
   };
 }
 
@@ -362,7 +377,7 @@ function updateSettingsBuffer(settings) {
     if (settings.maxHeaderSize !== undefined &&
       (settings.maxHeaderSize !== settings.maxHeaderListSize)) {
       process.emitWarning(
-        'settings.maxHeaderSize overwrite settings.maxHeaderListSize'
+        'settings.maxHeaderSize overwrite settings.maxHeaderListSize',
       );
       settingsBuffer[IDX_SETTINGS_MAX_HEADER_LIST_SIZE] =
         settings.maxHeaderSize;
@@ -404,7 +419,7 @@ function getSessionState(session) {
     deflateDynamicTableSize:
       sessionState[IDX_SESSION_STATE_HD_DEFLATE_DYNAMIC_TABLE_SIZE],
     inflateDynamicTableSize:
-      sessionState[IDX_SESSION_STATE_HD_INFLATE_DYNAMIC_TABLE_SIZE]
+      sessionState[IDX_SESSION_STATE_HD_INFLATE_DYNAMIC_TABLE_SIZE],
   };
 }
 
@@ -416,7 +431,7 @@ function getStreamState(stream) {
     sumDependencyWeight: streamState[IDX_STREAM_STATE_SUM_DEPENDENCY_WEIGHT],
     localClose: streamState[IDX_STREAM_STATE_LOCAL_CLOSE],
     remoteClose: streamState[IDX_STREAM_STATE_REMOTE_CLOSE],
-    localWindowSize: streamState[IDX_STREAM_STATE_LOCAL_WINDOW_SIZE]
+    localWindowSize: streamState[IDX_STREAM_STATE_LOCAL_WINDOW_SIZE],
   };
 }
 
@@ -424,7 +439,6 @@ function isIllegalConnectionSpecificHeader(name, value) {
   switch (name) {
     case HTTP2_HEADER_CONNECTION:
     case HTTP2_HEADER_UPGRADE:
-    case HTTP2_HEADER_HOST:
     case HTTP2_HEADER_HTTP2_SETTINGS:
     case HTTP2_HEADER_KEEP_ALIVE:
     case HTTP2_HEADER_PROXY_CONNECTION:
@@ -453,24 +467,31 @@ const assertValidPseudoHeaderTrailer = hideStackFrames((key) => {
   throw new ERR_HTTP2_INVALID_PSEUDOHEADER(key);
 });
 
+const emptyArray = [];
+const kNeverIndexFlag = StringFromCharCode(NGHTTP2_NV_FLAG_NO_INDEX);
+const kNoHeaderFlags = StringFromCharCode(NGHTTP2_NV_FLAG_NONE);
 function mapToHeaders(map,
                       assertValuePseudoHeader = assertValidPseudoHeader) {
-  let ret = '';
+  let headers = '';
+  let pseudoHeaders = '';
   let count = 0;
   const keys = ObjectKeys(map);
-  const singles = new Set();
+  const singles = new SafeSet();
   let i, j;
   let isArray;
   let key;
   let value;
   let isSingleValueHeader;
   let err;
+  const neverIndex =
+    ArrayPrototypeMap(map[kSensitiveHeaders] || emptyArray,
+                      StringPrototypeToLowerCase);
   for (i = 0; i < keys.length; ++i) {
     key = keys[i];
     value = map[key];
     if (value === undefined || key === '')
       continue;
-    key = key.toLowerCase();
+    key = StringPrototypeToLowerCase(key);
     isSingleValueHeader = kSingleValueHeaders.has(key);
     isArray = ArrayIsArray(value);
     if (isArray) {
@@ -493,38 +514,53 @@ function mapToHeaders(map,
         throw new ERR_HTTP2_HEADER_SINGLE_VALUE(key);
       singles.add(key);
     }
+    const flags = ArrayPrototypeIncludes(neverIndex, key) ?
+      kNeverIndexFlag :
+      kNoHeaderFlags;
     if (key[0] === ':') {
       err = assertValuePseudoHeader(key);
       if (err !== undefined)
         throw err;
-      ret = `${key}\0${value}\0${ret}`;
+      pseudoHeaders += `${key}\0${value}\0${flags}`;
       count++;
       continue;
     }
+    if (StringPrototypeIncludes(key, ' ')) {
+      throw new ERR_INVALID_HTTP_TOKEN('Header name', key);
+    }
     if (isIllegalConnectionSpecificHeader(key, value)) {
       throw new ERR_HTTP2_INVALID_CONNECTION_HEADERS(key);
     }
     if (isArray) {
       for (j = 0; j < value.length; ++j) {
         const val = String(value[j]);
-        ret += `${key}\0${val}\0`;
+        headers += `${key}\0${val}\0${flags}`;
       }
       count += value.length;
       continue;
     }
-    ret += `${key}\0${value}\0`;
+    headers += `${key}\0${value}\0${flags}`;
     count++;
   }
 
-  return [ret, count];
+  return [pseudoHeaders + headers, count];
 }
 
 class NghttpError extends Error {
-  constructor(ret) {
-    super(binding.nghttp2ErrorString(ret));
-    this.code = 'ERR_HTTP2_ERROR';
-    this.errno = ret;
-    addCodeToName(this, super.name, 'ERR_HTTP2_ERROR');
+  constructor(integerCode, customErrorCode) {
+    super(customErrorCode ?
+      getMessage(customErrorCode, [], null) :
+      binding.nghttp2ErrorString(integerCode));
+    this.code = customErrorCode || 'ERR_HTTP2_ERROR';
+    this.errno = integerCode;
+    captureLargerStackTrace(this);
+    ObjectDefineProperty(this, kIsNodeError, {
+      __proto__: null,
+      value: true,
+      enumerable: false,
+      writable: false,
+      configurable: true,
+    });
   }
 
   toString() {
@@ -548,12 +584,12 @@ const assertWithinRange = hideStackFrames(
       throw new ERR_HTTP2_INVALID_SETTING_VALUE.RangeError(
         name, value, min, max);
     }
-  }
+  },
 );
 
-function toHeaderObject(headers) {
+function toHeaderObject(headers, sensitiveHeaders) {
   const obj = ObjectCreate(null);
-  for (var n = 0; n < headers.length; n += 2) {
+  for (let n = 0; n < headers.length; n += 2) {
     const name = headers[n];
     let value = headers[n + 1];
     if (name === HTTP2_HEADER_STATUS)
@@ -579,7 +615,7 @@ function toHeaderObject(headers) {
           // fields with the same name.  Since it cannot be combined into a
           // single field-value, recipients ought to handle "Set-Cookie" as a
           // special case while processing header fields."
-          existing.push(value);
+          ArrayPrototypePush(existing, value);
           break;
         default:
           // https://tools.ietf.org/html/rfc7230#section-3.2.2
@@ -592,6 +628,7 @@ function toHeaderObject(headers) {
       }
     }
   }
+  obj[kSensitiveHeaders] = sensitiveHeaders;
   return obj;
 }
 
@@ -610,16 +647,30 @@ function sessionName(type) {
   }
 }
 
+function getAuthority(headers) {
+  // For non-CONNECT requests, HTTP/2 allows either :authority
+  // or Host to be used equivalently. The first is preferred
+  // when making HTTP/2 requests, and the latter is preferred
+  // when converting from an HTTP/1 message.
+  if (headers[HTTP2_HEADER_AUTHORITY] !== undefined)
+    return headers[HTTP2_HEADER_AUTHORITY];
+  if (headers[HTTP2_HEADER_HOST] !== undefined)
+    return headers[HTTP2_HEADER_HOST];
+}
+
 module.exports = {
   assertIsObject,
+  assertValidPseudoHeader,
   assertValidPseudoHeaderResponse,
   assertValidPseudoHeaderTrailer,
   assertWithinRange,
+  getAuthority,
   getDefaultSettings,
   getSessionState,
   getSettings,
   getStreamState,
   isPayloadMeaningless,
+  kSensitiveHeaders,
   kSocket,
   kProxySocket,
   kRequest,
@@ -628,5 +679,5 @@ module.exports = {
   sessionName,
   toHeaderObject,
   updateOptionsBuffer,
-  updateSettingsBuffer
+  updateSettingsBuffer,
 };
