@@ -1,8 +1,9 @@
 'use strict';
 
 const {
-  JSONStringify,
-  PromiseResolve,
+  Symbol,
+  RegExpPrototypeExec,
+  globalThis,
 } = primordials;
 
 const path = require('path');
@@ -10,8 +11,9 @@ const path = require('path');
 const {
   codes: {
     ERR_INVALID_ARG_TYPE,
-    ERR_UNCAUGHT_EXCEPTION_CAPTURE_ALREADY_SET
-  }
+    ERR_UNCAUGHT_EXCEPTION_CAPTURE_ALREADY_SET,
+    ERR_EVAL_ESM_CANNOT_PRINT,
+  },
 } = require('internal/errors');
 
 const {
@@ -20,9 +22,12 @@ const {
   clearAsyncIdStack,
   hasAsyncIdStack,
   afterHooksExist,
-  emitAfter
+  emitAfter,
+  popAsyncContext,
 } = require('internal/async_hooks');
-
+const {
+  makeContextifyScript, runScriptInThisContext,
+} = require('internal/vm');
 // shouldAbortOnUncaughtToggle is a typed array for faster
 // communication with JS.
 const { shouldAbortOnUncaughtToggle } = internalBinding('util');
@@ -39,74 +44,86 @@ function tryGetCwd() {
 }
 
 function evalModule(source, print) {
-  const { log, error } = require('internal/console/global');
-  const { decorateErrorStack } = require('internal/util');
-  const asyncESM = require('internal/process/esm_loader');
-  PromiseResolve(asyncESM.ESMLoader).then(async (loader) => {
-    const { result } = await loader.eval(source);
-    if (print) {
-      log(result);
-    }
-  })
-  .catch((e) => {
-    decorateErrorStack(e);
-    error(e);
-    process.exit(1);
-  });
+  if (print) {
+    throw new ERR_EVAL_ESM_CANNOT_PRINT();
+  }
+  const { loadESM } = require('internal/process/esm_loader');
+  const { handleMainPromise } = require('internal/modules/run_main');
+  RegExpPrototypeExec(/^/, ''); // Necessary to reset RegExp statics before user code runs.
+  return handleMainPromise(loadESM((loader) => loader.eval(source)));
 }
 
-function evalScript(name, body, breakFirstLine, print) {
+function evalScript(name, body, breakFirstLine, print, shouldLoadESM = false) {
   const CJSModule = require('internal/modules/cjs/loader').Module;
-  const { kVmBreakFirstLineSymbol } = require('internal/util');
-  const { pathToFileURL } = require('url');
+  const { pathToFileURL } = require('internal/url');
 
   const cwd = tryGetCwd();
-  const origModule = global.module;  // Set e.g. when called from the REPL.
+  const origModule = globalThis.module;  // Set e.g. when called from the REPL.
 
   const module = new CJSModule(name);
   module.filename = path.join(cwd, name);
   module.paths = CJSModule._nodeModulePaths(cwd);
 
-  global.kVmBreakFirstLineSymbol = kVmBreakFirstLineSymbol;
-  global.asyncESM = require('internal/process/esm_loader');
-
+  const { handleMainPromise } = require('internal/modules/run_main');
+  const asyncESM = require('internal/process/esm_loader');
   const baseUrl = pathToFileURL(module.filename).href;
+  const { loadESM } = asyncESM;
 
-  const script = `
-    global.__filename = ${JSONStringify(name)};
-    global.exports = exports;
-    global.module = module;
-    global.__dirname = __dirname;
-    global.require = require;
-    const { kVmBreakFirstLineSymbol, asyncESM } = global;
-    delete global.kVmBreakFirstLineSymbol;
-    delete global.asyncESM;
-    return require("vm").runInThisContext(
-      ${JSONStringify(body)}, {
-        filename: ${JSONStringify(name)},
-        displayErrors: true,
-        [kVmBreakFirstLineSymbol]: ${!!breakFirstLine},
-        async importModuleDynamically (specifier) {
-          const loader = await asyncESM.ESMLoader;
-          return loader.import(specifier, ${JSONStringify(baseUrl)});
-        }
-      });\n`;
-  const result = module._compile(script, `${name}-wrapper`);
-  if (print) {
-    const { log } = require('internal/console/global');
-    log(result);
+  const runScript = () => {
+    // Create wrapper for cache entry
+    const script = `
+      globalThis.module = module;
+      globalThis.exports = exports;
+      globalThis.__dirname = __dirname;
+      globalThis.require = require;
+      return (main) => main();
+    `;
+    globalThis.__filename = name;
+    RegExpPrototypeExec(/^/, ''); // Necessary to reset RegExp statics before user code runs.
+    const result = module._compile(script, `${name}-wrapper`)(() => {
+      const hostDefinedOptionId = Symbol(name);
+      async function importModuleDynamically(specifier, _, importAttributes) {
+        const loader = asyncESM.esmLoader;
+        return loader.import(specifier, baseUrl, importAttributes);
+      }
+      const script = makeContextifyScript(
+        body,                    // code
+        name,                    // filename,
+        0,                       // lineOffset
+        0,                       // columnOffset,
+        undefined,               // cachedData
+        false,                   // produceCachedData
+        undefined,               // parsingContext
+        hostDefinedOptionId,     // hostDefinedOptionId
+        importModuleDynamically, // importModuleDynamically
+      );
+      return runScriptInThisContext(script, true, !!breakFirstLine);
+    });
+    if (print) {
+      const { log } = require('internal/console/global');
+      log(result);
+    }
+
+    if (origModule !== undefined)
+      globalThis.module = origModule;
+  };
+
+  if (shouldLoadESM) {
+    return handleMainPromise(loadESM(runScript));
   }
-
-  if (origModule !== undefined)
-    global.module = origModule;
+  return runScript();
 }
 
-const exceptionHandlerState = { captureFn: null };
+const exceptionHandlerState = {
+  captureFn: null,
+  reportFlag: false,
+};
 
 function setUncaughtExceptionCaptureCallback(fn) {
   if (fn === null) {
     exceptionHandlerState.captureFn = fn;
     shouldAbortOnUncaughtToggle[0] = 1;
+    process.report.reportOnUncaughtException = exceptionHandlerState.reportFlag;
     return;
   }
   if (typeof fn !== 'function') {
@@ -117,6 +134,9 @@ function setUncaughtExceptionCaptureCallback(fn) {
   }
   exceptionHandlerState.captureFn = fn;
   shouldAbortOnUncaughtToggle[0] = 0;
+  exceptionHandlerState.reportFlag =
+    process.report.reportOnUncaughtException === true;
+  process.report.reportOnUncaughtException = false;
 }
 
 function hasUncaughtExceptionCaptureCallback() {
@@ -142,22 +162,6 @@ function createOnGlobalUncaughtException() {
     // call that threw and was never cleared. So clear it now.
     clearDefaultTriggerAsyncId();
 
-    // If diagnostic reporting is enabled, call into its handler to see
-    // whether it is interested in handling the situation.
-    // Ignore if the error is scoped inside a domain.
-    // use == in the checks as we want to allow for null and undefined
-    if (er == null || er.domain == null) {
-      try {
-        const report = internalBinding('report');
-        if (report != null && report.shouldReportOnUncaughtException()) {
-          report.writeReport(er ? er.message : 'Exception',
-                             'Exception',
-                             null,
-                             er ? er : {});
-        }
-      } catch {}  // Ignore the exception. Diagnostic reporting is unavailable.
-    }
-
     const type = fromPromise ? 'unhandledRejection' : 'uncaughtException';
     process.emit('uncaughtExceptionMonitor', er, type);
     if (exceptionHandlerState.captureFn !== null) {
@@ -184,7 +188,11 @@ function createOnGlobalUncaughtException() {
     // Emit the after() hooks now that the exception has been handled.
     if (afterHooksExist()) {
       do {
-        emitAfter(executionAsyncId());
+        const asyncId = executionAsyncId();
+        if (asyncId === 0)
+          popAsyncContext(0);
+        else
+          emitAfter(asyncId);
       } while (hasAsyncIdStack());
     }
     // And completely empty the id stack, including anything that may be
@@ -215,5 +223,5 @@ module.exports = {
   evalScript,
   onGlobalUncaughtException: createOnGlobalUncaughtException(),
   setUncaughtExceptionCaptureCallback,
-  hasUncaughtExceptionCaptureCallback
+  hasUncaughtExceptionCaptureCallback,
 };
