@@ -3,15 +3,35 @@
 const {
   ArrayIsArray,
   Error,
+  ErrorPrototypeToString,
+  ErrorCaptureStackTrace,
+  String,
 } = primordials;
 
-const { ERR_INVALID_ARG_TYPE } = require('internal/errors').codes;
+const assert = require('internal/assert');
+const {
+  codes: {
+    ERR_INVALID_ARG_TYPE,
+  },
+  isErrorStackTraceLimitWritable,
+} = require('internal/errors');
+const { validateString } = require('internal/validators');
 
 // Lazily loaded
 let fs;
 let fd;
 let warningFile;
 let options;
+let traceWarningHelperShown = false;
+
+function resetForSerialization() {
+  if (fd !== undefined) {
+    process.removeListener('exit', closeFdOnExit);
+  }
+  fd = undefined;
+  warningFile = undefined;
+  traceWarningHelperShown = false;
+}
 
 function lazyOption() {
   // This will load `warningFile` only once. If the flag is not set,
@@ -40,6 +60,14 @@ function writeOut(message) {
   error(message);
 }
 
+function closeFdOnExit() {
+  try {
+    fs.closeSync(fd);
+  } catch {
+    // Continue regardless of error.
+  }
+}
+
 function writeToFile(message) {
   if (fd === undefined) {
     fs = require('fs');
@@ -48,11 +76,7 @@ function writeToFile(message) {
     } catch {
       return writeOut(message);
     }
-    process.on('exit', () => {
-      try {
-        fs.closeSync(fd);
-      } catch {}
-    });
+    process.on('exit', closeFdOnExit);
   }
   fs.appendFile(fd, `${message}\n`, (err) => {
     if (err) {
@@ -62,7 +86,7 @@ function writeToFile(message) {
 }
 
 function doEmitWarning(warning) {
-  return () => process.emit('warning', warning);
+  process.emit('warning', warning);
 }
 
 function onWarning(warning) {
@@ -77,14 +101,21 @@ function onWarning(warning) {
   if (trace && warning.stack) {
     msg += `${warning.stack}`;
   } else {
-    const toString =
+    msg +=
       typeof warning.toString === 'function' ?
-        warning.toString : Error.prototype.toString;
-    msg += `${toString.apply(warning)}`;
+        `${warning.toString()}` :
+        ErrorPrototypeToString(warning);
   }
   if (typeof warning.detail === 'string') {
     msg += `\n${warning.detail}`;
   }
+  if (!trace && !traceWarningHelperShown) {
+    const flag = isDeprecation ? '--trace-deprecation' : '--trace-warnings';
+    const argv0 = require('path').basename(process.argv0 || 'node', '.exe');
+    msg += `\n(Use \`${argv0} ${flag} ...\` to show where the warning ` +
+           'was created)';
+    traceWarningHelperShown = true;
+  }
   const warningFile = lazyOption();
   if (warningFile) {
     return writeToFile(msg);
@@ -95,7 +126,7 @@ function onWarning(warning) {
 // process.emitWarning(error)
 // process.emitWarning(str[, type[, code]][, ctor])
 // process.emitWarning(str[, options])
-function emitWarning(warning, type, code, ctor, now) {
+function emitWarning(warning, type, code, ctor) {
   let detail;
   if (type !== null && typeof type === 'object' && !ArrayIsArray(type)) {
     ctor = type.ctor;
@@ -108,42 +139,56 @@ function emitWarning(warning, type, code, ctor, now) {
     code = undefined;
     type = 'Warning';
   }
-  if (type !== undefined && typeof type !== 'string') {
-    throw new ERR_INVALID_ARG_TYPE('type', 'string', type);
-  }
+  if (type !== undefined)
+    validateString(type, 'type');
   if (typeof code === 'function') {
     ctor = code;
     code = undefined;
-  } else if (code !== undefined && typeof code !== 'string') {
-    throw new ERR_INVALID_ARG_TYPE('code', 'string', code);
+  } else if (code !== undefined) {
+    validateString(code, 'code');
   }
   if (typeof warning === 'string') {
-    // Improve error creation performance by skipping the error frames.
-    // They are added in the `captureStackTrace()` function below.
-    const tmpStackLimit = Error.stackTraceLimit;
-    Error.stackTraceLimit = 0;
-    // eslint-disable-next-line no-restricted-syntax
-    warning = new Error(warning);
-    Error.stackTraceLimit = tmpStackLimit;
-    warning.name = String(type || 'Warning');
-    if (code !== undefined) warning.code = code;
-    if (detail !== undefined) warning.detail = detail;
-    // eslint-disable-next-line no-restricted-syntax
-    Error.captureStackTrace(warning, ctor || process.emitWarning);
+    warning = createWarningObject(warning, type, code, ctor, detail);
   } else if (!(warning instanceof Error)) {
     throw new ERR_INVALID_ARG_TYPE('warning', ['Error', 'string'], warning);
   }
   if (warning.name === 'DeprecationWarning') {
     if (process.noDeprecation)
       return;
-    if (process.throwDeprecation)
-      throw warning;
+    if (process.throwDeprecation) {
+      // Delay throwing the error to guarantee that all former warnings were
+      // properly logged.
+      return process.nextTick(() => {
+        throw warning;
+      });
+    }
   }
-  if (now) process.emit('warning', warning);
-  else process.nextTick(doEmitWarning(warning));
+  process.nextTick(doEmitWarning, warning);
+}
+
+function emitWarningSync(warning, type, code, ctor) {
+  process.emit('warning', createWarningObject(warning, type, code, ctor));
+}
+
+function createWarningObject(warning, type, code, ctor, detail) {
+  assert(typeof warning === 'string');
+  // Improve error creation performance by skipping the error frames.
+  // They are added in the `captureStackTrace()` function below.
+  const tmpStackLimit = Error.stackTraceLimit;
+  if (isErrorStackTraceLimitWritable()) Error.stackTraceLimit = 0;
+  // eslint-disable-next-line no-restricted-syntax
+  warning = new Error(warning);
+  if (isErrorStackTraceLimitWritable()) Error.stackTraceLimit = tmpStackLimit;
+  warning.name = String(type || 'Warning');
+  if (code !== undefined) warning.code = code;
+  if (detail !== undefined) warning.detail = detail;
+  ErrorCaptureStackTrace(warning, ctor || process.emitWarning);
+  return warning;
 }
 
 module.exports = {
+  emitWarning,
+  emitWarningSync,
   onWarning,
-  emitWarning
+  resetForSerialization,
 };
