@@ -33,7 +33,7 @@ const visitorsWithoutAncestors = {
       state.prepend(node, `${node.id.name}=`);
       ArrayPrototypePush(
         state.hoistedDeclarationStatements,
-        `let ${node.id.name}; `
+        `let ${node.id.name}; `,
       );
     }
 
@@ -49,7 +49,7 @@ const visitorsWithoutAncestors = {
     state.prepend(node, `this.${node.id.name} = ${node.id.name}; `);
     ArrayPrototypePush(
       state.hoistedDeclarationStatements,
-      `var ${node.id.name}; `
+      `var ${node.id.name}; `,
     );
   },
   FunctionExpression: noop,
@@ -67,7 +67,7 @@ const visitorsWithoutAncestors = {
     const variableKind = node.kind;
     const isIterableForDeclaration = ArrayPrototypeIncludes(
       ['ForOfStatement', 'ForInStatement'],
-      state.ancestors[state.ancestors.length - 2].type
+      state.ancestors[state.ancestors.length - 2].type,
     );
 
     if (variableKind === 'var' || isTopLevelDeclaration(state)) {
@@ -76,7 +76,7 @@ const visitorsWithoutAncestors = {
         node.start + variableKind.length + (isIterableForDeclaration ? 1 : 0),
         variableKind === 'var' && isIterableForDeclaration ?
           '' :
-          'void' + (node.declarations.length === 1 ? '' : ' (')
+          'void' + (node.declarations.length === 1 ? '' : ' ('),
       );
 
       if (!isIterableForDeclaration) {
@@ -99,7 +99,7 @@ const visitorsWithoutAncestors = {
           case 'Identifier':
             ArrayPrototypePush(
               variableIdentifiersToHoist[variableKind === 'var' ? 0 : 1][1],
-              node.name
+              node.name,
             );
             break;
           case 'ObjectPattern':
@@ -125,15 +125,15 @@ const visitorsWithoutAncestors = {
           if (identifiers.length > 0) {
             ArrayPrototypePush(
               state.hoistedDeclarationStatements,
-              `${kind} ${ArrayPrototypeJoin(identifiers, ', ')}; `
+              `${kind} ${ArrayPrototypeJoin(identifiers, ', ')}; `,
             );
           }
-        }
+        },
       );
     }
 
     walk.base.VariableDeclaration(node, state, c);
-  }
+  },
 };
 
 const visitors = {};
@@ -209,7 +209,7 @@ function processTopLevelAwait(src) {
       wrappedArray[node.end - 1] += str;
     },
     containsAwait: false,
-    containsReturn: false
+    containsReturn: false,
   };
 
   walk.recursive(body, state, visitors);
@@ -258,5 +258,5 @@ function processTopLevelAwait(src) {
 }
 
 module.exports = {
-  processTopLevelAwait
+  processTopLevelAwait,
 };
