@@ -115,7 +115,7 @@ function getOriginalSymbolName(sourceMap, trace, curIndex) {
   // First check for a symbol name associated with the enclosing function:
   const enclosingEntry = sourceMap.findEntry(
     trace[curIndex].getEnclosingLineNumber() - 1,
-    trace[curIndex].getEnclosingColumnNumber() - 1
+    trace[curIndex].getEnclosingColumnNumber() - 1,
   );
   if (enclosingEntry.name) return enclosingEntry.name;
   // Fallback to using the symbol name attached to the next stack frame:
@@ -124,7 +124,7 @@ function getOriginalSymbolName(sourceMap, trace, curIndex) {
   if (nextCallSite && currentFileName === nextCallSite.getFileName()) {
     const { name } = sourceMap.findEntry(
       nextCallSite.getLineNumber() - 1,
-      nextCallSite.getColumnNumber() - 1
+      nextCallSite.getColumnNumber() - 1,
     );
     return name;
   }
@@ -137,14 +137,14 @@ function getErrorSource(
   sourceMap,
   originalSourcePath,
   originalLine,
-  originalColumn
+  originalColumn,
 ) {
   const originalSourcePathNoScheme =
     StringPrototypeStartsWith(originalSourcePath, 'file://') ?
       fileURLToPath(originalSourcePath) : originalSourcePath;
   const source = getOriginalSource(
     sourceMap.payload,
-    originalSourcePath
+    originalSourcePath,
   );
   if (typeof source !== 'string') {
     return;
