@@ -1,7 +1,7 @@
 'use strict';
 const {
-  ArrayPrototypeMap,
   ArrayPrototypePush,
+  ArrayPrototypePushApply,
   ArrayPrototypeReduce,
   ArrayPrototypeShift,
   ArrayPrototypeSlice,
@@ -13,16 +13,20 @@ const {
   ObjectSeal,
   PromisePrototypeThen,
   PromiseResolve,
+  SafePromisePrototypeFinally,
   ReflectApply,
   RegExpPrototypeExec,
   SafeMap,
   SafeSet,
   SafePromiseAll,
   SafePromiseRace,
+  SymbolDispose,
+  ObjectDefineProperty,
   Symbol,
 } = primordials;
+const { getCallerLocation } = internalBinding('util');
+const { addAbortListener } = require('events');
 const { AsyncResource } = require('async_hooks');
-const { once } = require('events');
 const { AbortController } = require('internal/abort_controller');
 const {
   codes: {
@@ -31,13 +35,13 @@ const {
   },
   AbortError,
 } = require('internal/errors');
-const { getOptionValue } = require('internal/options');
-const { MockTracker } = require('internal/test_runner/mock');
-const { TapStream } = require('internal/test_runner/tap_stream');
+const { MockTracker } = require('internal/test_runner/mock/mock');
+const { TestsStream } = require('internal/test_runner/tests_stream');
 const {
-  convertStringToRegExp,
   createDeferredCallback,
+  countCompletedTest,
   isTestFailureError,
+  parseCommandLine,
 } = require('internal/test_runner/utils');
 const {
   createDeferredPromise,
@@ -51,46 +55,66 @@ const {
   validateOneOf,
   validateUint32,
 } = require('internal/validators');
-const { setTimeout } = require('timers/promises');
+const { setTimeout } = require('timers');
 const { TIMEOUT_MAX } = require('internal/timers');
-const { cpus } = require('os');
+const { availableParallelism } = require('os');
 const { bigint: hrtime } = process.hrtime;
 const kCallbackAndPromisePresent = 'callbackAndPromisePresent';
 const kCancelledByParent = 'cancelledByParent';
+const kAborted = 'testAborted';
 const kParentAlreadyFinished = 'parentAlreadyFinished';
 const kSubtestsFailed = 'subtestsFailed';
 const kTestCodeFailure = 'testCodeFailure';
 const kTestTimeoutFailure = 'testTimeoutFailure';
 const kHookFailure = 'hookFailed';
 const kDefaultTimeout = null;
-const kDefaultIndent = '    '; // 4 spaces
 const noop = FunctionPrototype;
-const isTestRunner = getOptionValue('--test');
-const testOnlyFlag = !isTestRunner && getOptionValue('--test-only');
-const testNamePatternFlag = isTestRunner ? null :
-  getOptionValue('--test-name-pattern');
-const testNamePatterns = testNamePatternFlag?.length > 0 ?
-  ArrayPrototypeMap(
-    testNamePatternFlag,
-    (re) => convertStringToRegExp(re, '--test-name-pattern')
-  ) : null;
 const kShouldAbort = Symbol('kShouldAbort');
+const kFilename = process.argv?.[1];
 const kHookNames = ObjectSeal(['before', 'after', 'beforeEach', 'afterEach']);
 const kUnwrapErrors = new SafeSet()
   .add(kTestCodeFailure).add(kHookFailure)
   .add('uncaughtException').add('unhandledRejection');
-
+const { testNamePatterns, testOnlyFlag } = parseCommandLine();
+let kResistStopPropagation;
 
 function stopTest(timeout, signal) {
+  const deferred = createDeferredPromise();
+  const abortListener = addAbortListener(signal, deferred.resolve);
+  let timer;
+  let disposeFunction;
+
   if (timeout === kDefaultTimeout) {
-    return once(signal, 'abort');
+    disposeFunction = abortListener[SymbolDispose];
+  } else {
+    timer = setTimeout(() => deferred.resolve(), timeout);
+    timer.unref();
+
+    ObjectDefineProperty(deferred, 'promise', {
+      __proto__: null,
+      configurable: true,
+      writable: true,
+      value: PromisePrototypeThen(deferred.promise, () => {
+        throw new ERR_TEST_FAILURE(
+          `test timed out after ${timeout}ms`,
+          kTestTimeoutFailure,
+        );
+      }),
+    });
+
+    disposeFunction = () => {
+      abortListener[SymbolDispose]();
+      timer[SymbolDispose]();
+    };
   }
-  return PromisePrototypeThen(setTimeout(timeout, null, { ref: false, signal }), () => {
-    throw new ERR_TEST_FAILURE(
-      `test timed out after ${timeout}ms`,
-      kTestTimeoutFailure
-    );
+
+  ObjectDefineProperty(deferred.promise, SymbolDispose, {
+    __proto__: null,
+    configurable: true,
+    writable: true,
+    value: disposeFunction,
   });
+  return deferred.promise;
 }
 
 class TestContext {
@@ -130,12 +154,23 @@ class TestContext {
   }
 
   test(name, options, fn) {
-    // eslint-disable-next-line no-use-before-define
-    const subtest = this.#test.createSubtest(Test, name, options, fn);
+    const overrides = {
+      __proto__: null,
+      loc: getCallerLocation(),
+    };
+
+    const subtest = this.#test.createSubtest(
+      // eslint-disable-next-line no-use-before-define
+      Test, name, options, fn, overrides,
+    );
 
     return subtest.start();
   }
 
+  before(fn, options) {
+    this.#test.createHook('before', fn, options);
+  }
+
   after(fn, options) {
     this.#test.createHook('after', fn, options);
   }
@@ -166,15 +201,15 @@ class SuiteContext {
 }
 
 class Test extends AsyncResource {
-  #abortController;
-  #outerSignal;
+  abortController;
+  outerSignal;
   #reportedSubtest;
 
   constructor(options) {
     super('Test');
 
     let { fn, name, parent, skip } = options;
-    const { concurrency, only, timeout, todo, signal } = options;
+    const { concurrency, loc, only, timeout, todo, signal } = options;
 
     if (typeof fn !== 'function') {
       fn = noop;
@@ -190,23 +225,39 @@ class Test extends AsyncResource {
 
     if (parent === null) {
       this.concurrency = 1;
-      this.indent = '';
+      this.nesting = 0;
       this.only = testOnlyFlag;
-      this.reporter = new TapStream();
+      this.reporter = new TestsStream();
       this.runOnlySubtests = this.only;
       this.testNumber = 0;
       this.timeout = kDefaultTimeout;
+      this.root = this;
+      this.hooks = {
+        __proto__: null,
+        before: [],
+        after: [],
+        beforeEach: [],
+        afterEach: [],
+      };
     } else {
-      const indent = parent.parent === null ? parent.indent :
-        parent.indent + kDefaultIndent;
+      const nesting = parent.parent === null ? parent.nesting :
+        parent.nesting + 1;
 
       this.concurrency = parent.concurrency;
-      this.indent = indent;
+      this.nesting = nesting;
       this.only = only ?? !parent.runOnlySubtests;
       this.reporter = parent.reporter;
       this.runOnlySubtests = !this.only;
       this.testNumber = parent.subtests.length + 1;
       this.timeout = parent.timeout;
+      this.root = parent.root;
+      this.hooks = {
+        __proto__: null,
+        before: [],
+        after: [],
+        beforeEach: ArrayPrototypeSlice(parent.hooks.beforeEach),
+        afterEach: ArrayPrototypeSlice(parent.hooks.afterEach),
+      };
     }
 
     switch (typeof concurrency) {
@@ -217,8 +268,8 @@ class Test extends AsyncResource {
 
       case 'boolean':
         if (concurrency) {
-          // TODO(cjihrig): Use uv_available_parallelism() once it lands.
-          this.concurrency = parent === null ? MathMax(cpus().length - 1, 1) : Infinity;
+          this.concurrency = parent === null ?
+            MathMax(availableParallelism() - 1, 1) : Infinity;
         } else {
           this.concurrency = 1;
         }
@@ -234,16 +285,11 @@ class Test extends AsyncResource {
       this.timeout = timeout;
     }
 
-    if (testNamePatterns !== null) {
-      // eslint-disable-next-line no-use-before-define
-      const match = this instanceof TestHook || ArrayPrototypeSome(
-        testNamePatterns,
-        (re) => RegExpPrototypeExec(re, name) !== null
-      );
+    this.name = name;
+    this.parent = parent;
 
-      if (!match) {
-        skip = 'test name does not match pattern';
-      }
+    if (testNamePatterns !== null && !this.matchesTestNamePatterns()) {
+      skip = 'test name does not match pattern';
     }
 
     if (testOnlyFlag && !this.only) {
@@ -254,20 +300,27 @@ class Test extends AsyncResource {
       fn = noop;
     }
 
-    this.#abortController = new AbortController();
-    this.#outerSignal = signal;
-    this.signal = this.#abortController.signal;
+    this.abortController = new AbortController();
+    this.outerSignal = signal;
+    this.signal = this.abortController.signal;
 
     validateAbortSignal(signal, 'options.signal');
-    this.#outerSignal?.addEventListener('abort', this.#abortHandler);
+    if (signal) {
+      kResistStopPropagation ??= require('internal/event_target').kResistStopPropagation;
+    }
+
+    this.outerSignal?.addEventListener(
+      'abort',
+      this.#abortHandler,
+      { __proto__: null, [kResistStopPropagation]: true },
+    );
 
     this.fn = fn;
+    this.harness = null; // Configured on the root test by the test harness.
     this.mock = null;
-    this.name = name;
-    this.parent = parent;
     this.cancelled = false;
-    this.skipped = !!skip;
-    this.isTodo = !!todo;
+    this.skipped = skip !== undefined && skip !== false;
+    this.isTodo = todo !== undefined && todo !== false;
     this.startTime = null;
     this.endTime = null;
     this.passed = false;
@@ -279,14 +332,30 @@ class Test extends AsyncResource {
     this.pendingSubtests = [];
     this.readySubtests = new SafeMap();
     this.subtests = [];
-    this.hooks = {
-      before: [],
-      after: [],
-      beforeEach: [],
-      afterEach: [],
-    };
     this.waitingOn = 0;
     this.finished = false;
+
+    if (!testOnlyFlag && (only || this.runOnlySubtests)) {
+      const warning =
+        "'only' and 'runOnly' require the --test-only command-line option.";
+      this.diagnostic(warning);
+    }
+
+    if (loc === undefined || kFilename === undefined) {
+      this.loc = undefined;
+    } else {
+      this.loc = {
+        __proto__: null,
+        line: loc[0],
+        column: loc[1],
+        file: loc[2],
+      };
+    }
+  }
+
+  matchesTestNamePatterns() {
+    return ArrayPrototypeSome(testNamePatterns, (re) => RegExpPrototypeExec(re, this.name) !== null) ||
+      this.parent?.matchesTestNamePatterns();
   }
 
   hasConcurrency() {
@@ -300,7 +369,9 @@ class Test extends AsyncResource {
   async processPendingSubtests() {
     while (this.pendingSubtests.length > 0 && this.hasConcurrency()) {
       const deferred = ArrayPrototypeShift(this.pendingSubtests);
-      await deferred.test.run();
+      const test = deferred.test;
+      this.reporter.dequeue(test.nesting, test.loc, test.name);
+      await test.run();
       deferred.resolve();
     }
   }
@@ -334,7 +405,7 @@ class Test extends AsyncResource {
       }
 
       if (i === 1 && this.parent !== null) {
-        this.reportSubtest();
+        this.reportStarted();
       }
 
       // Report the subtest's results and remove it from the ready map.
@@ -379,8 +450,8 @@ class Test extends AsyncResource {
       test.fail(
         new ERR_TEST_FAILURE(
           'test could not be started because its parent finished',
-          kParentAlreadyFinished
-        )
+          kParentAlreadyFinished,
+        ),
       );
     }
 
@@ -389,10 +460,12 @@ class Test extends AsyncResource {
   }
 
   #abortHandler = () => {
-    this.cancel(this.#outerSignal?.reason || new AbortError('The test was aborted'));
+    const error = this.outerSignal?.reason || new AbortError('The test was aborted');
+    error.failureType = kAborted;
+    this.#cancel(error);
   };
 
-  cancel(error) {
+  #cancel(error) {
     if (this.endTime !== null) {
       return;
     }
@@ -400,18 +473,21 @@ class Test extends AsyncResource {
     this.fail(error ||
       new ERR_TEST_FAILURE(
         'test did not finish before its parent and was cancelled',
-        kCancelledByParent
-      )
+        kCancelledByParent,
+      ),
     );
     this.startTime = this.startTime || this.endTime; // If a test was canceled before it was started, e.g inside a hook
     this.cancelled = true;
-    this.#abortController.abort();
+    this.abortController.abort();
   }
 
   createHook(name, fn, options) {
     validateOneOf(name, 'hook name', kHookNames);
     // eslint-disable-next-line no-use-before-define
     const hook = new TestHook(fn, options);
+    if (name === 'before' || name === 'after') {
+      hook.run = runOnce(hook.run);
+    }
     ArrayPrototypePush(this.hooks[name], hook);
     return hook;
   }
@@ -453,6 +529,7 @@ class Test extends AsyncResource {
     // If there is enough available concurrency to run the test now, then do
     // it. Otherwise, return a Promise to the caller and mark the test as
     // pending for later execution.
+    this.reporter.enqueue(this.nesting, this.loc, this.name);
     if (!this.parent.hasConcurrency()) {
       const deferred = createDeferredPromise();
 
@@ -461,6 +538,7 @@ class Test extends AsyncResource {
       return deferred.promise;
     }
 
+    this.reporter.dequeue(this.nesting, this.loc, this.name);
     return this.run();
   }
 
@@ -468,15 +546,15 @@ class Test extends AsyncResource {
     if (this.signal.aborted) {
       return true;
     }
-    if (this.#outerSignal?.aborted) {
-      this.cancel(this.#outerSignal.reason || new AbortError('The test was aborted'));
+    if (this.outerSignal?.aborted) {
+      this.#abortHandler();
       return true;
     }
   }
 
   getRunArgs() {
     const ctx = new TestContext(this);
-    return { ctx, args: [ctx] };
+    return { __proto__: null, ctx, args: [ctx] };
   }
 
   async runHook(hook, args) {
@@ -508,17 +586,27 @@ class Test extends AsyncResource {
     }
 
     const { args, ctx } = this.getRunArgs();
+    const after = async () => {
+      if (this.hooks.after.length > 0) {
+        await this.runHook('after', { __proto__: null, args, ctx });
+      }
+    };
     const afterEach = runOnce(async () => {
       if (this.parent?.hooks.afterEach.length > 0) {
-        await this.parent.runHook('afterEach', { args, ctx });
+        await this.parent.runHook('afterEach', { __proto__: null, args, ctx });
       }
     });
 
+    let stopPromise;
+
     try {
-      if (this.parent?.hooks.beforeEach.length > 0) {
-        await this.parent.runHook('beforeEach', { args, ctx });
+      if (this.parent?.hooks.before.length > 0) {
+        await this.parent.runHook('before', this.parent.getRunArgs());
       }
-      const stopPromise = stopTest(this.timeout, this.signal);
+      if (this.parent?.hooks.beforeEach.length > 0) {
+        await this.parent.runHook('beforeEach', { __proto__: null, args, ctx });
+      }
+      stopPromise = stopTest(this.timeout, this.signal);
       const runArgs = ArrayPrototypeSlice(args);
       ArrayPrototypeUnshift(runArgs, this.fn, ctx);
 
@@ -532,7 +620,7 @@ class Test extends AsyncResource {
         if (isPromise(ret)) {
           this.fail(new ERR_TEST_FAILURE(
             'passed a callback but also returned a Promise',
-            kCallbackAndPromisePresent
+            kCallbackAndPromisePresent,
           ));
           await SafePromiseRace([ret, stopPromise]);
         } else {
@@ -549,30 +637,44 @@ class Test extends AsyncResource {
         return;
       }
 
-      await this.runHook('after', { args, ctx });
       await afterEach();
+      await after();
       this.pass();
     } catch (err) {
-      try { await afterEach(); } catch { /* test is already failing, let's the error */ }
+      try { await afterEach(); } catch { /* test is already failing, let's ignore the error */ }
+      try { await after(); } catch { /* Ignore error. */ }
       if (isTestFailureError(err)) {
         if (err.failureType === kTestTimeoutFailure) {
-          this.cancel(err);
+          this.#cancel(err);
         } else {
           this.fail(err);
         }
       } else {
         this.fail(new ERR_TEST_FAILURE(err, kTestCodeFailure));
       }
+    } finally {
+      stopPromise?.[SymbolDispose]();
+
+      // Do not abort hooks and the root test as hooks instance are shared between tests suite so aborting them will
+      // cause them to not run for further tests.
+      if (this.parent !== null) {
+        this.abortController.abort();
+      }
     }
 
-    // Clean up the test. Then, try to report the results and execute any
-    // tests that were pending due to available concurrency.
-    this.postRun();
+    if (this.parent !== null || typeof this.hookType === 'string') {
+      // Clean up the test. Then, try to report the results and execute any
+      // tests that were pending due to available concurrency.
+      //
+      // The root test is skipped here because it is a special case. Its
+      // postRun() method is called when the process is getting ready to exit.
+      // This helps catch any asynchronous activity that occurs after the tests
+      // have finished executing.
+      this.postRun();
+    }
   }
 
   postRun(pendingSubtestsError) {
-    const counters = { __proto__: null, failed: 0, passed: 0, cancelled: 0, skipped: 0, todo: 0, totalFailed: 0 };
-
     // If the test was failed before it even started, then the end time will
     // be earlier than the start time. Correct that here.
     if (this.endTime < this.startTime) {
@@ -583,41 +685,27 @@ class Test extends AsyncResource {
     // The test has run, so recursively cancel any outstanding subtests and
     // mark this test as failed if any subtests failed.
     this.pendingSubtests = [];
+    let failed = 0;
     for (let i = 0; i < this.subtests.length; i++) {
       const subtest = this.subtests[i];
 
       if (!subtest.finished) {
-        subtest.cancel(pendingSubtestsError);
+        subtest.#cancel(pendingSubtestsError);
         subtest.postRun(pendingSubtestsError);
       }
-
-      // Check SKIP and TODO tests first, as those should not be counted as
-      // failures.
-      if (subtest.skipped) {
-        counters.skipped++;
-      } else if (subtest.isTodo) {
-        counters.todo++;
-      } else if (subtest.cancelled) {
-        counters.cancelled++;
-      } else if (!subtest.passed) {
-        counters.failed++;
-      } else {
-        counters.passed++;
-      }
-
-      if (!subtest.passed) {
-        counters.totalFailed++;
+      if (!subtest.passed && !subtest.isTodo) {
+        failed++;
       }
     }
 
-    if ((this.passed || this.parent === null) && counters.totalFailed > 0) {
-      const subtestString = `subtest${counters.totalFailed > 1 ? 's' : ''}`;
-      const msg = `${counters.totalFailed} ${subtestString} failed`;
+    if ((this.passed || this.parent === null) && failed > 0) {
+      const subtestString = `subtest${failed > 1 ? 's' : ''}`;
+      const msg = `${failed} ${subtestString} failed`;
 
       this.fail(new ERR_TEST_FAILURE(msg, kSubtestsFailed));
     }
 
-    this.#outerSignal?.removeEventListener('abort', this.#abortHandler);
+    this.outerSignal?.removeEventListener('abort', this.#abortHandler);
     this.mock?.reset();
 
     if (this.parent !== null) {
@@ -625,22 +713,50 @@ class Test extends AsyncResource {
       this.parent.addReadySubtest(this);
       this.parent.processReadySubtestRange(false);
       this.parent.processPendingSubtests();
-    } else if (!this.reported) {
-      this.reported = true;
-      this.reporter.plan(this.indent, this.subtests.length);
 
-      for (let i = 0; i < this.diagnostics.length; i++) {
-        this.reporter.diagnostic(this.indent, this.diagnostics[i]);
+      if (this.parent === this.root &&
+          this.root.activeSubtests === 0 &&
+          this.root.pendingSubtests.length === 0 &&
+          this.root.readySubtests.size === 0 &&
+          this.root.hooks.after.length > 0) {
+        // This is done so that any global after() hooks are run. At this point
+        // all of the tests have finished running. However, there might be
+        // ref'ed handles keeping the event loop alive. This gives the global
+        // after() hook a chance to clean them up.
+        this.root.run();
+      }
+    } else if (!this.reported) {
+      const {
+        diagnostics,
+        harness,
+        loc,
+        nesting,
+        reporter,
+      } = this;
+
+      this.reported = true;
+      reporter.plan(nesting, loc, harness.counters.topLevel);
+
+      // Call this harness.coverage() before collecting diagnostics, since failure to collect coverage is a diagnostic.
+      const coverage = harness.coverage();
+      for (let i = 0; i < diagnostics.length; i++) {
+        reporter.diagnostic(nesting, loc, diagnostics[i]);
       }
 
-      this.reporter.diagnostic(this.indent, `tests ${this.subtests.length}`);
-      this.reporter.diagnostic(this.indent, `pass ${counters.passed}`);
-      this.reporter.diagnostic(this.indent, `fail ${counters.failed}`);
-      this.reporter.diagnostic(this.indent, `cancelled ${counters.cancelled}`);
-      this.reporter.diagnostic(this.indent, `skipped ${counters.skipped}`);
-      this.reporter.diagnostic(this.indent, `todo ${counters.todo}`);
-      this.reporter.diagnostic(this.indent, `duration_ms ${this.#duration()}`);
-      this.reporter.push(null);
+      reporter.diagnostic(nesting, loc, `tests ${harness.counters.all}`);
+      reporter.diagnostic(nesting, loc, `suites ${harness.counters.suites}`);
+      reporter.diagnostic(nesting, loc, `pass ${harness.counters.passed}`);
+      reporter.diagnostic(nesting, loc, `fail ${harness.counters.failed}`);
+      reporter.diagnostic(nesting, loc, `cancelled ${harness.counters.cancelled}`);
+      reporter.diagnostic(nesting, loc, `skipped ${harness.counters.skipped}`);
+      reporter.diagnostic(nesting, loc, `todo ${harness.counters.todo}`);
+      reporter.diagnostic(nesting, loc, `duration_ms ${this.duration()}`);
+
+      if (coverage) {
+        reporter.coverage(nesting, loc, coverage);
+      }
+
+      reporter.end();
     }
   }
 
@@ -668,19 +784,20 @@ class Test extends AsyncResource {
     this.finished = true;
   }
 
-  #duration() {
+  duration() {
     // Duration is recorded in BigInt nanoseconds. Convert to milliseconds.
     return Number(this.endTime - this.startTime) / 1_000_000;
   }
 
   report() {
+    countCompletedTest(this);
     if (this.subtests.length > 0) {
-      this.reporter.plan(this.subtests[0].indent, this.subtests.length);
+      this.reporter.plan(this.subtests[0].nesting, this.loc, this.subtests.length);
     } else {
-      this.reportSubtest();
+      this.reportStarted();
     }
     let directive;
-    const details = { __proto__: null, duration: this.#duration() };
+    const details = { __proto__: null, duration_ms: this.duration() };
 
     if (this.skipped) {
       directive = this.reporter.getSkip(this.message);
@@ -688,25 +805,29 @@ class Test extends AsyncResource {
       directive = this.reporter.getTodo(this.message);
     }
 
+    if (this.reportedType) {
+      details.type = this.reportedType;
+    }
+
     if (this.passed) {
-      this.reporter.ok(this.indent, this.testNumber, this.name, details, directive);
+      this.reporter.ok(this.nesting, this.loc, this.testNumber, this.name, details, directive);
     } else {
       details.error = this.error;
-      this.reporter.fail(this.indent, this.testNumber, this.name, details, directive);
+      this.reporter.fail(this.nesting, this.loc, this.testNumber, this.name, details, directive);
     }
 
     for (let i = 0; i < this.diagnostics.length; i++) {
-      this.reporter.diagnostic(this.indent, this.diagnostics[i]);
+      this.reporter.diagnostic(this.nesting, this.loc, this.diagnostics[i]);
     }
   }
 
-  reportSubtest() {
+  reportStarted() {
     if (this.#reportedSubtest || this.parent === null) {
       return;
     }
     this.#reportedSubtest = true;
-    this.parent.reportSubtest();
-    this.reporter.subtest(this.indent, this.name);
+    this.parent.reportStarted();
+    this.reporter.start(this.nesting, this.loc, this.name);
   }
 }
 
@@ -716,59 +837,93 @@ class TestHook extends Test {
     if (options === null || typeof options !== 'object') {
       options = kEmptyObject;
     }
-    const { timeout, signal } = options;
-    super({ __proto__: null, fn, timeout, signal });
+    const { loc, timeout, signal } = options;
+    super({ __proto__: null, fn, loc, timeout, signal });
+
+    this.parentTest = options.parent ?? null;
+    this.hookType = options.hookType;
   }
   run(args) {
+    if (this.error && !this.outerSignal?.aborted) {
+      this.passed = false;
+      this.error = null;
+      this.abortController.abort();
+      this.abortController = new AbortController();
+      this.signal = this.abortController.signal;
+    }
+
     this.#args = args;
     return super.run();
   }
   getRunArgs() {
     return this.#args;
   }
-  postRun() {
+  matchesTestNamePatterns() {
+    return true;
   }
-}
+  postRun() {
+    const { error, loc, parentTest: parent } = this;
 
-class ItTest extends Test {
-  constructor(opt) { super(opt); } // eslint-disable-line no-useless-constructor
-  getRunArgs() {
-    return { ctx: { signal: this.signal, name: this.name }, args: [] };
+    // Report failures in the root test's after() hook.
+    if (error && parent !== null &&
+        parent === parent.root && this.hookType === 'after') {
+
+      if (isTestFailureError(error)) {
+        error.failureType = kHookFailure;
+      }
+
+      parent.reporter.fail(0, loc, parent.subtests.length + 1, loc.file, {
+        __proto__: null,
+        duration_ms: this.duration(),
+        error,
+      }, undefined);
+    }
   }
 }
 
 class Suite extends Test {
+  reportedType = 'suite';
   constructor(options) {
     super(options);
 
+    if (testNamePatterns !== null && !options.skip && !options.todo) {
+      this.fn = options.fn || this.fn;
+      this.skipped = false;
+    }
+    this.runOnlySubtests = testOnlyFlag;
+
     try {
       const { ctx, args } = this.getRunArgs();
-      this.buildSuite = PromisePrototypeThen(
-        PromiseResolve(this.runInAsyncScope(this.fn, ctx, args)),
-        undefined,
-        (err) => {
-          this.fail(new ERR_TEST_FAILURE(err, kTestCodeFailure));
-        });
+      const runArgs = [this.fn, ctx];
+      ArrayPrototypePushApply(runArgs, args);
+      this.buildSuite = SafePromisePrototypeFinally(
+        PromisePrototypeThen(
+          PromiseResolve(ReflectApply(this.runInAsyncScope, this, runArgs)),
+          undefined,
+          (err) => {
+            this.fail(new ERR_TEST_FAILURE(err, kTestCodeFailure));
+          }),
+        () => {
+          this.buildPhaseFinished = true;
+        },
+      );
     } catch (err) {
       this.fail(new ERR_TEST_FAILURE(err, kTestCodeFailure));
+
+      this.buildPhaseFinished = true;
     }
     this.fn = () => {};
-    this.buildPhaseFinished = true;
   }
 
   getRunArgs() {
     const ctx = new SuiteContext(this);
-    return { ctx, args: [ctx] };
+    return { __proto__: null, ctx, args: [ctx] };
   }
 
   async run() {
     const hookArgs = this.getRunArgs();
-    const afterEach = runOnce(async () => {
-      if (this.parent?.hooks.afterEach.length > 0) {
-        await this.parent.runHook('afterEach', hookArgs);
-      }
-    });
 
+    let stopPromise;
     try {
       this.parent.activeSubtests++;
       await this.buildSuite;
@@ -780,28 +935,28 @@ class Suite extends Test {
         return;
       }
 
-      if (this.parent?.hooks.beforeEach.length > 0) {
-        await this.parent.runHook('beforeEach', hookArgs);
+      if (this.parent.hooks.before.length > 0) {
+        await this.parent.runHook('before', this.parent.getRunArgs());
       }
 
       await this.runHook('before', hookArgs);
 
-      const stopPromise = stopTest(this.timeout, this.signal);
+      stopPromise = stopTest(this.timeout, this.signal);
       const subtests = this.skipped || this.error ? [] : this.subtests;
       const promise = SafePromiseAll(subtests, (subtests) => subtests.start());
 
       await SafePromiseRace([promise, stopPromise]);
       await this.runHook('after', hookArgs);
-      await afterEach();
 
       this.pass();
     } catch (err) {
-      try { await afterEach(); } catch { /* test is already failing, let's the error */ }
       if (isTestFailureError(err)) {
         this.fail(err);
       } else {
         this.fail(new ERR_TEST_FAILURE(err, kTestCodeFailure));
       }
+    } finally {
+      stopPromise?.[SymbolDispose]();
     }
 
     this.postRun();
@@ -809,11 +964,11 @@ class Suite extends Test {
 }
 
 module.exports = {
-  ItTest,
   kCancelledByParent,
-  kDefaultIndent,
   kSubtestsFailed,
   kTestCodeFailure,
+  kTestTimeoutFailure,
+  kAborted,
   kUnwrapErrors,
   Suite,
   Test,
