@@ -118,6 +118,9 @@ const {
 } = require('internal/util');
 const { inspect } = require('internal/util/inspect');
 const vm = require('vm');
+
+const { runInThisContext, runInContext } = vm.Script.prototype;
+
 const path = require('path');
 const fs = require('fs');
 const { Interface } = require('readline');
@@ -186,7 +189,9 @@ const {
   extensionFormatMap,
   legacyExtensionFormatMap,
 } = require('internal/modules/esm/formats');
-
+const {
+  makeContextifyScript,
+} = require('internal/vm');
 let nextREPLResourceNumber = 1;
 // This prevents v8 code cache from getting confused and using a different
 // cache from a resource of the same name
@@ -430,8 +435,6 @@ function REPLServer(prompt,
   }
 
   function defaultEval(code, context, file, cb) {
-    const asyncESM = require('internal/process/esm_loader');
-
     let result, script, wrappedErr;
     let err = null;
     let wrappedCmd = false;
@@ -449,6 +452,21 @@ function REPLServer(prompt,
       wrappedCmd = true;
     }
 
+    const hostDefinedOptionId = Symbol(`eval:${file}`);
+    let parentURL;
+    try {
+      const { pathToFileURL } = require('internal/url');
+      // Adding `/repl` prevents dynamic imports from loading relative
+      // to the parent of `process.cwd()`.
+      parentURL = pathToFileURL(path.join(process.cwd(), 'repl')).href;
+    } catch {
+      // Continue regardless of error.
+    }
+    async function importModuleDynamically(specifier, _, importAttributes) {
+      const asyncESM = require('internal/process/esm_loader');
+      return asyncESM.esmLoader.import(specifier, parentURL,
+                                       importAttributes);
+    }
     // `experimentalREPLAwait` is set to true by default.
     // Shall be false in case `--no-experimental-repl-await` flag is used.
     if (experimentalREPLAwait && StringPrototypeIncludes(code, 'await')) {
@@ -466,28 +484,21 @@ function REPLServer(prompt,
       } catch (e) {
         let recoverableError = false;
         if (e.name === 'SyntaxError') {
-          let parentURL;
-          try {
-            const { pathToFileURL } = require('url');
-            // Adding `/repl` prevents dynamic imports from loading relative
-            // to the parent of `process.cwd()`.
-            parentURL = pathToFileURL(path.join(process.cwd(), 'repl')).href;
-          } catch {
-            // Continue regardless of error.
-          }
-
           // Remove all "await"s and attempt running the script
           // in order to detect if error is truly non recoverable
           const fallbackCode = SideEffectFreeRegExpPrototypeSymbolReplace(/\bawait\b/g, code, '');
           try {
-            vm.createScript(fallbackCode, {
-              filename: file,
-              displayErrors: true,
-              importModuleDynamically: (specifier, _, importAttributes) => {
-                return asyncESM.esmLoader.import(specifier, parentURL,
-                                                 importAttributes);
-              },
-            });
+            makeContextifyScript(
+              fallbackCode,            // code
+              file,                    // filename,
+              0,                       // lineOffset
+              0,                       // columnOffset,
+              undefined,               // cachedData
+              false,                   // produceCachedData
+              undefined,               // parsingContext
+              hostDefinedOptionId,     // hostDefinedOptionId
+              importModuleDynamically, // importModuleDynamically
+            );
           } catch (fallbackError) {
             if (isRecoverableError(fallbackError, fallbackCode)) {
               recoverableError = true;
@@ -507,15 +518,6 @@ function REPLServer(prompt,
       return cb(null);
 
     if (err === null) {
-      let parentURL;
-      try {
-        const { pathToFileURL } = require('url');
-        // Adding `/repl` prevents dynamic imports from loading relative
-        // to the parent of `process.cwd()`.
-        parentURL = pathToFileURL(path.join(process.cwd(), 'repl')).href;
-      } catch {
-        // Continue regardless of error.
-      }
       while (true) {
         try {
           if (self.replMode === module.exports.REPL_MODE_STRICT &&
@@ -524,14 +526,17 @@ function REPLServer(prompt,
             // value for statements and declarations that don't return a value.
             code = `'use strict'; void 0;\n${code}`;
           }
-          script = vm.createScript(code, {
-            filename: file,
-            displayErrors: true,
-            importModuleDynamically: (specifier, _, importAttributes) => {
-              return asyncESM.esmLoader.import(specifier, parentURL,
-                                               importAttributes);
-            },
-          });
+          script = makeContextifyScript(
+            code,                    // code
+            file,                    // filename,
+            0,                       // lineOffset
+            0,                       // columnOffset,
+            undefined,               // cachedData
+            false,                   // produceCachedData
+            undefined,               // parsingContext
+            hostDefinedOptionId,     // hostDefinedOptionId
+            importModuleDynamically, // importModuleDynamically
+          );
         } catch (e) {
           debug('parse error %j', code, e);
           if (wrappedCmd) {
@@ -591,9 +596,9 @@ function REPLServer(prompt,
           };
 
           if (self.useGlobal) {
-            result = script.runInThisContext(scriptOptions);
+            result = ReflectApply(runInThisContext, script, [scriptOptions]);
           } else {
-            result = script.runInContext(context, scriptOptions);
+            result = ReflectApply(runInContext, script, [context, scriptOptions]);
           }
         } finally {
           if (self.breakEvalOnSigint) {
