@@ -33,14 +33,15 @@ const errors = require('internal/errors');
 const { ERR_INVALID_FD, ERR_TTY_INIT_FAILED } = errors.codes;
 const {
   getColorDepth,
-  hasColors
+  hasColors,
 } = require('internal/tty');
 
 // Lazy loaded for startup performance.
 let readline;
 
 function isatty(fd) {
-  return NumberIsInteger(fd) && fd >= 0 && isTTY(fd);
+  return NumberIsInteger(fd) && fd >= 0 && fd <= 2147483647 &&
+         isTTY(fd);
 }
 
 function ReadStream(fd, options) {
@@ -50,17 +51,16 @@ function ReadStream(fd, options) {
     throw new ERR_INVALID_FD(fd);
 
   const ctx = {};
-  const tty = new TTY(fd, true, ctx);
+  const tty = new TTY(fd, ctx);
   if (ctx.code !== undefined) {
     throw new ERR_TTY_INIT_FAILED(ctx);
   }
 
   net.Socket.call(this, {
-    highWaterMark: 0,
-    readable: true,
-    writable: false,
+    readableHighWaterMark: 0,
     handle: tty,
-    ...options
+    manualStart: true,
+    ...options,
   });
 
   this.isRaw = false;
@@ -72,7 +72,7 @@ ObjectSetPrototypeOf(ReadStream, net.Socket);
 
 ReadStream.prototype.setRawMode = function(flag) {
   flag = !!flag;
-  const err = this._handle.setRawMode(flag);
+  const err = this._handle?.setRawMode(flag);
   if (err) {
     this.emit('error', errors.errnoException(err, 'setRawMode'));
     return this;
@@ -88,15 +88,15 @@ function WriteStream(fd) {
     throw new ERR_INVALID_FD(fd);
 
   const ctx = {};
-  const tty = new TTY(fd, false, ctx);
+  const tty = new TTY(fd, ctx);
   if (ctx.code !== undefined) {
     throw new ERR_TTY_INIT_FAILED(ctx);
   }
 
   net.Socket.call(this, {
+    readableHighWaterMark: 0,
     handle: tty,
-    readable: false,
-    writable: true
+    manualStart: true,
   });
 
   // Prevents interleaved or dropped stdout/stderr output for terminals.
@@ -132,7 +132,7 @@ WriteStream.prototype._refreshSize = function() {
     this.emit('error', errors.errnoException(err, 'getWindowSize'));
     return;
   }
-  const [newCols, newRows] = winSize;
+  const { 0: newCols, 1: newRows } = winSize;
   if (oldCols !== newCols || oldRows !== newRows) {
     this.columns = newCols;
     this.rows = newRows;
