@@ -33,13 +33,13 @@ const {
 } = primordials;
 
 const { Buffer } = require('buffer');
-const { validateString } = require('internal/validators');
+const { validateString, validateUint32 } = require('internal/validators');
 const {
   Serializer,
-  Deserializer
+  Deserializer,
 } = internalBinding('serdes');
 const {
-  namespace: startupSnapshot
+  namespace: startupSnapshot,
 } = require('internal/v8/startup_snapshot');
 
 let profiler = {};
@@ -55,11 +55,12 @@ const { getValidatedPath } = require('internal/fs/utils');
 const { toNamespacedPath } = require('path');
 const {
   createHeapSnapshotStream,
-  triggerHeapSnapshot
+  triggerHeapSnapshot,
 } = internalBinding('heap_utils');
 const { HeapSnapshotStream } = require('internal/heap_utils');
 const promiseHooks = require('internal/promise_hooks');
-
+const { getOptionValue } = require('internal/options');
+const { JSONParse } = primordials;
 /**
  * Generates a snapshot of the current V8 heap
  * and writes it to a JSON file.
@@ -95,6 +96,7 @@ const {
   updateHeapStatisticsBuffer,
   updateHeapSpaceStatisticsBuffer,
   updateHeapCodeStatisticsBuffer,
+  setHeapSnapshotNearHeapLimit: _setHeapSnapshotNearHeapLimit,
 
   // Properties for heap statistics buffer extraction.
   kTotalHeapSizeIndex,
@@ -173,7 +175,7 @@ function getHeapStatistics() {
     number_of_detached_contexts: buffer[kNumberOfDetachedContextsIndex],
     total_global_handles_size: buffer[kTotalGlobalHandlesSizeIndex],
     used_global_handles_size: buffer[kUsedGlobalHandlesSizeIndex],
-    external_memory: buffer[kExternalMemoryIndex]
+    external_memory: buffer[kExternalMemoryIndex],
   };
 }
 
@@ -198,7 +200,7 @@ function getHeapSpaceStatistics() {
       space_size: buffer[kSpaceSizeIndex],
       space_used_size: buffer[kSpaceUsedSizeIndex],
       space_available_size: buffer[kSpaceAvailableSizeIndex],
-      physical_space_size: buffer[kPhysicalSpaceSizeIndex]
+      physical_space_size: buffer[kPhysicalSpaceSizeIndex],
     };
   }
 
@@ -226,6 +228,18 @@ function getHeapCodeStatistics() {
   };
 }
 
+let heapSnapshotNearHeapLimitCallbackAdded = false;
+function setHeapSnapshotNearHeapLimit(limit) {
+  validateUint32(limit, 'limit', 1);
+  if (heapSnapshotNearHeapLimitCallbackAdded ||
+      getOptionValue('--heapsnapshot-near-heap-limit') > 0
+  ) {
+    return;
+  }
+  heapSnapshotNearHeapLimitCallbackAdded = true;
+  _setHeapSnapshotNearHeapLimit(limit);
+}
+
 /* V8 serialization API */
 
 /* JS methods for the base objects */
@@ -370,6 +384,25 @@ function deserialize(buffer) {
   return der.readValue();
 }
 
+class GCProfiler {
+  #profiler = null;
+
+  start() {
+    if (!this.#profiler) {
+      this.#profiler = new binding.GCProfiler();
+      this.#profiler.start();
+    }
+  }
+
+  stop() {
+    if (this.#profiler) {
+      const data = this.#profiler.stop();
+      this.#profiler = null;
+      return JSONParse(data);
+    }
+  }
+}
+
 module.exports = {
   cachedDataVersionTag,
   getHeapSnapshot,
@@ -387,5 +420,7 @@ module.exports = {
   serialize,
   writeHeapSnapshot,
   promiseHooks,
-  startupSnapshot
+  startupSnapshot,
+  setHeapSnapshotNearHeapLimit,
+  GCProfiler,
 };
