@@ -13,6 +13,7 @@
 #endif
 #ifdef __linux__
 #include <linux/capability.h>
+#include <sys/auxv.h>
 #include <sys/syscall.h>
 #endif  // __linux__
 
@@ -31,9 +32,18 @@ using v8::TryCatch;
 using v8::Uint32;
 using v8::Value;
 
-namespace per_process {
-bool linux_at_secure = false;
-}  // namespace per_process
+bool linux_at_secure() {
+  // This could reasonably be a static variable, but this way
+  // we can guarantee that this function is always usable
+  // and returns the correct value,  e.g. even in static
+  // initialization code in other files.
+#ifdef __linux__
+  static const bool value = getauxval(AT_SECURE);
+  return value;
+#else
+  return false;
+#endif
+}
 
 namespace credentials {
 
@@ -42,7 +52,7 @@ namespace credentials {
 bool HasOnly(int capability) {
   DCHECK(cap_valid(capability));
 
-  struct __user_cap_data_struct cap_data[2];
+  struct __user_cap_data_struct cap_data[_LINUX_CAPABILITY_U32S_3];
   struct __user_cap_header_struct cap_header_data = {
     _LINUX_CAPABILITY_VERSION_3,
     getpid()};
@@ -51,12 +61,11 @@ bool HasOnly(int capability) {
   if (syscall(SYS_capget, &cap_header_data, &cap_data) != 0) {
     return false;
   }
-  if (capability < 32) {
-    return cap_data[0].permitted ==
-        static_cast<unsigned int>(CAP_TO_MASK(capability));
-  }
-  return cap_data[1].permitted ==
-      static_cast<unsigned int>(CAP_TO_MASK(capability));
+
+  static_assert(arraysize(cap_data) == 2);
+  return cap_data[CAP_TO_INDEX(capability)].permitted ==
+             static_cast<unsigned int>(CAP_TO_MASK(capability)) &&
+         cap_data[1 - CAP_TO_INDEX(capability)].permitted == 0;
 }
 #endif
 
@@ -70,11 +79,10 @@ bool SafeGetenv(const char* key,
                 v8::Isolate* isolate) {
 #if !defined(__CloudABI__) && !defined(_WIN32)
 #if defined(__linux__)
-  if ((!HasOnly(CAP_NET_BIND_SERVICE) && per_process::linux_at_secure) ||
+  if ((!HasOnly(CAP_NET_BIND_SERVICE) && linux_at_secure()) ||
       getuid() != geteuid() || getgid() != getegid())
 #else
-  if (per_process::linux_at_secure || getuid() != geteuid() ||
-      getgid() != getegid())
+  if (linux_at_secure() || getuid() != geteuid() || getgid() != getegid())
 #endif
     goto fail;
 #endif
@@ -206,7 +214,8 @@ static const char* name_by_gid(gid_t gid) {
 
 static uid_t uid_by_name(Isolate* isolate, Local<Value> value) {
   if (value->IsUint32()) {
-    return static_cast<uid_t>(value.As<Uint32>()->Value());
+    static_assert(std::is_same<uid_t, uint32_t>::value);
+    return value.As<Uint32>()->Value();
   } else {
     Utf8Value name(isolate, value);
     return uid_by_name(*name);
@@ -215,7 +224,8 @@ static uid_t uid_by_name(Isolate* isolate, Local<Value> value) {
 
 static gid_t gid_by_name(Isolate* isolate, Local<Value> value) {
   if (value->IsUint32()) {
-    return static_cast<gid_t>(value.As<Uint32>()->Value());
+    static_assert(std::is_same<gid_t, uint32_t>::value);
+    return value.As<Uint32>()->Value();
   } else {
     Utf8Value name(isolate, value);
     return gid_by_name(*name);
@@ -445,26 +455,26 @@ static void Initialize(Local<Object> target,
                        Local<Value> unused,
                        Local<Context> context,
                        void* priv) {
+  SetMethod(context, target, "safeGetenv", SafeGetenv);
+
+#ifdef NODE_IMPLEMENTS_POSIX_CREDENTIALS
   Environment* env = Environment::GetCurrent(context);
   Isolate* isolate = env->isolate();
 
-  env->SetMethod(target, "safeGetenv", SafeGetenv);
-
-#ifdef NODE_IMPLEMENTS_POSIX_CREDENTIALS
   READONLY_TRUE_PROPERTY(target, "implementsPosixCredentials");
-  env->SetMethodNoSideEffect(target, "getuid", GetUid);
-  env->SetMethodNoSideEffect(target, "geteuid", GetEUid);
-  env->SetMethodNoSideEffect(target, "getgid", GetGid);
-  env->SetMethodNoSideEffect(target, "getegid", GetEGid);
-  env->SetMethodNoSideEffect(target, "getgroups", GetGroups);
+  SetMethodNoSideEffect(context, target, "getuid", GetUid);
+  SetMethodNoSideEffect(context, target, "geteuid", GetEUid);
+  SetMethodNoSideEffect(context, target, "getgid", GetGid);
+  SetMethodNoSideEffect(context, target, "getegid", GetEGid);
+  SetMethodNoSideEffect(context, target, "getgroups", GetGroups);
 
   if (env->owns_process_state()) {
-    env->SetMethod(target, "initgroups", InitGroups);
-    env->SetMethod(target, "setegid", SetEGid);
-    env->SetMethod(target, "seteuid", SetEUid);
-    env->SetMethod(target, "setgid", SetGid);
-    env->SetMethod(target, "setuid", SetUid);
-    env->SetMethod(target, "setgroups", SetGroups);
+    SetMethod(context, target, "initgroups", InitGroups);
+    SetMethod(context, target, "setegid", SetEGid);
+    SetMethod(context, target, "seteuid", SetEUid);
+    SetMethod(context, target, "setgid", SetGid);
+    SetMethod(context, target, "setuid", SetUid);
+    SetMethod(context, target, "setgroups", SetGroups);
   }
 #endif  // NODE_IMPLEMENTS_POSIX_CREDENTIALS
 }
@@ -472,6 +482,6 @@ static void Initialize(Local<Object> target,
 }  // namespace credentials
 }  // namespace node
 
-NODE_MODULE_CONTEXT_AWARE_INTERNAL(credentials, node::credentials::Initialize)
-NODE_MODULE_EXTERNAL_REFERENCE(credentials,
-                               node::credentials::RegisterExternalReferences)
+NODE_BINDING_CONTEXT_AWARE_INTERNAL(credentials, node::credentials::Initialize)
+NODE_BINDING_EXTERNAL_REFERENCE(credentials,
+                                node::credentials::RegisterExternalReferences)
