@@ -1,13 +1,19 @@
 #include "node_buffer.h"
 #include "node_internals.h"
 #include "libplatform/libplatform.h"
+#include "util.h"
 
 #include <string>
 #include "gtest/gtest.h"
 #include "node_test_fixture.h"
+#include <stdio.h>
+#include <cstdio>
 
 using node::AtExit;
 using node::RunAtExit;
+using node::USE;
+using v8::Context;
+using v8::Local;
 
 static bool called_cb_1 = false;
 static bool called_cb_2 = false;
@@ -24,7 +30,7 @@ static std::string cb_1_arg;  // NOLINT(runtime/string)
 class EnvironmentTest : public EnvironmentTestFixture {
  private:
   void TearDown() override {
-    NodeTestFixture::TearDown();
+    EnvironmentTestFixture::TearDown();
     called_cb_1 = false;
     called_cb_2 = false;
     called_cb_ordered_1 = false;
@@ -66,7 +72,34 @@ TEST_F(EnvironmentTest, EnvironmentWithESMLoader) {
       "})()");
 }
 
+class RedirectStdErr {
+ public:
+  explicit RedirectStdErr(const char* filename) : filename_(filename) {
+    fflush(stderr);
+    fgetpos(stderr, &pos_);
+    fd_ = dup(fileno(stderr));
+    USE(freopen(filename_, "w", stderr));
+  }
+
+  ~RedirectStdErr() {
+    fflush(stderr);
+    dup2(fd_, fileno(stderr));
+    close(fd_);
+    remove(filename_);
+    clearerr(stderr);
+    fsetpos(stderr, &pos_);
+  }
+
+ private:
+  int fd_;
+  fpos_t pos_;
+  const char* filename_;
+};
+
 TEST_F(EnvironmentTest, EnvironmentWithNoESMLoader) {
+  // The following line will cause stderr to get redirected to avoid the
+  // error that would otherwise be printed to the console by this test.
+  RedirectStdErr redirect_scope("environment_test.log");
   const v8::HandleScope handle_scope(isolate_);
   Argv argv;
   Env env {handle_scope, argv, node::EnvironmentFlags::kNoRegisterESMLoader};
@@ -114,8 +147,8 @@ TEST_F(EnvironmentTest, PreExecutionPreparation) {
   v8::Local<v8::Script> script = v8::Script::Compile(
       context,
       v8::String::NewFromOneByte(isolate_,
-                                 reinterpret_cast<const uint8_t*>(run_script),
-                                 v8::NewStringType::kNormal).ToLocalChecked())
+                                 reinterpret_cast<const uint8_t*>(run_script))
+                                 .ToLocalChecked())
       .ToLocalChecked();
   v8::Local<v8::Value> result = script->Run(context).ToLocalChecked();
   CHECK(result->IsString());
@@ -140,8 +173,8 @@ TEST_F(EnvironmentTest, LoadEnvironmentWithCallback) {
         context,
         v8::String::NewFromOneByte(
             isolate_,
-            reinterpret_cast<const uint8_t*>("argv0"),
-            v8::NewStringType::kNormal).ToLocalChecked()).ToLocalChecked();
+            reinterpret_cast<const uint8_t*>("argv0"))
+            .ToLocalChecked()).ToLocalChecked();
     CHECK(argv0->IsString());
 
     return info.process_object;
@@ -165,15 +198,15 @@ TEST_F(EnvironmentTest, LoadEnvironmentWithSource) {
       context,
       v8::String::NewFromOneByte(
           isolate_,
-          reinterpret_cast<const uint8_t*>("process"),
-          v8::NewStringType::kNormal).ToLocalChecked())
+          reinterpret_cast<const uint8_t*>("process"))
+          .ToLocalChecked())
           .ToLocalChecked()->IsObject());
   CHECK(main_ret.As<v8::Object>()->Get(
       context,
       v8::String::NewFromOneByte(
           isolate_,
-          reinterpret_cast<const uint8_t*>("require"),
-          v8::NewStringType::kNormal).ToLocalChecked())
+          reinterpret_cast<const uint8_t*>("require"))
+          .ToLocalChecked())
           .ToLocalChecked()->IsFunction());
 }
 
@@ -182,17 +215,7 @@ TEST_F(EnvironmentTest, AtExitWithEnvironment) {
   const Argv argv;
   Env env {handle_scope, argv};
 
-  AtExit(*env, at_exit_callback1);
-  RunAtExit(*env);
-  EXPECT_TRUE(called_cb_1);
-}
-
-TEST_F(EnvironmentTest, AtExitWithoutEnvironment) {
-  const v8::HandleScope handle_scope(isolate_);
-  const Argv argv;
-  Env env {handle_scope, argv};
-
-  AtExit(at_exit_callback1);  // No Environment is passed to AtExit.
+  AtExit(*env, at_exit_callback1, nullptr);
   RunAtExit(*env);
   EXPECT_TRUE(called_cb_1);
 }
@@ -203,8 +226,8 @@ TEST_F(EnvironmentTest, AtExitOrder) {
   Env env {handle_scope, argv};
 
   // Test that callbacks are run in reverse order.
-  AtExit(*env, at_exit_callback_ordered1);
-  AtExit(*env, at_exit_callback_ordered2);
+  AtExit(*env, at_exit_callback_ordered1, nullptr);
+  AtExit(*env, at_exit_callback_ordered2, nullptr);
   RunAtExit(*env);
   EXPECT_TRUE(called_cb_ordered_1);
   EXPECT_TRUE(called_cb_ordered_2);
@@ -239,8 +262,8 @@ TEST_F(EnvironmentTest, MultipleEnvironmentsPerIsolate) {
   Env env1 {handle_scope, argv};
   Env env2 {handle_scope, argv, node::EnvironmentFlags::kNoFlags};
 
-  AtExit(*env1, at_exit_callback1);
-  AtExit(*env2, at_exit_callback2);
+  AtExit(*env1, at_exit_callback1, nullptr);
+  AtExit(*env2, at_exit_callback2, nullptr);
   RunAtExit(*env1);
   EXPECT_TRUE(called_cb_1);
   EXPECT_FALSE(called_cb_2);
@@ -310,8 +333,8 @@ static void at_exit_js(void* arg) {
   v8::Isolate* isolate = static_cast<v8::Isolate*>(arg);
   v8::HandleScope handle_scope(isolate);
   v8::Local<v8::Object> obj = v8::Object::New(isolate);
-  assert(!obj.IsEmpty());  // Assert VM is still alive.
-  assert(obj->IsObject());
+  EXPECT_FALSE(obj.IsEmpty());  // Assert VM is still alive.
+  EXPECT_TRUE(obj->IsObject());
   called_at_exit_js = true;
 }
 
@@ -333,7 +356,7 @@ TEST_F(EnvironmentTest, SetImmediateCleanup) {
     (*env)->SetImmediate([&](node::Environment* env_arg) {
       EXPECT_EQ(env_arg, *env);
       called++;
-    });
+    }, node::CallbackFlags::kRefed);
     (*env)->SetImmediate([&](node::Environment* env_arg) {
       EXPECT_EQ(env_arg, *env);
       called_unref++;
@@ -447,7 +470,7 @@ TEST_F(EnvironmentTest, InspectorMultipleEmbeddedEnvironments) {
         node::ArrayBufferAllocator::Create();
     uv_loop_t loop;
     uv_loop_init(&loop);
-    v8::Isolate* isolate = NewIsolate(aba.get(), &loop, data->platform);
+    v8::Isolate* isolate = NewIsolate(aba, &loop, data->platform);
     CHECK_NOT_NULL(isolate);
 
     {
@@ -509,8 +532,8 @@ TEST_F(EnvironmentTest, InspectorMultipleEmbeddedEnvironments) {
           context,
           v8::String::NewFromOneByte(
               isolate_,
-              reinterpret_cast<const uint8_t*>("messageFromWorker"),
-              v8::NewStringType::kNormal).ToLocalChecked())
+              reinterpret_cast<const uint8_t*>("messageFromWorker"))
+              .ToLocalChecked())
               .ToLocalChecked();
   CHECK_EQ(data.extracted_value, 42);
   CHECK_EQ(from_inspector->IntegerValue(context).FromJust(), 42);
@@ -530,7 +553,9 @@ TEST_F(EnvironmentTest, ExitHandlerTest) {
     callback_calls++;
     node::Stop(*env);
   });
-  node::LoadEnvironment(*env, "process.exit(42)").ToLocalChecked();
+  // When terminating, v8 throws makes the current embedder call bail out
+  // with MaybeLocal<>()
+  EXPECT_TRUE(node::LoadEnvironment(*env, "process.exit(42)").IsEmpty());
   EXPECT_EQ(callback_calls, 1);
 }
 
@@ -553,9 +578,165 @@ TEST_F(EnvironmentTest, SetImmediateMicrotasks) {
       isolate_->EnqueueMicrotask([](void* arg) {
         ++*static_cast<int*>(arg);
       }, &called);
-    });
+    }, node::CallbackFlags::kRefed);
     uv_run(&current_loop, UV_RUN_DEFAULT);
   }
 
   EXPECT_EQ(called, 1);
 }
+
+#ifndef _WIN32  // No SIGINT on Windows.
+TEST_F(NodeZeroIsolateTestFixture, CtrlCWithOnlySafeTerminationTest) {
+  // We need to go through the whole setup dance here because we want to
+  // set only_terminate_in_safe_scope.
+  // Allocate and initialize Isolate.
+  v8::Isolate::CreateParams create_params;
+  create_params.array_buffer_allocator = allocator.get();
+  create_params.only_terminate_in_safe_scope = true;
+  v8::Isolate* isolate = v8::Isolate::Allocate();
+  CHECK_NOT_NULL(isolate);
+  platform->RegisterIsolate(isolate, &current_loop);
+  v8::Isolate::Initialize(isolate, create_params);
+
+  // Try creating Context + IsolateData + Environment.
+  {
+    v8::Isolate::Scope isolate_scope(isolate);
+    v8::HandleScope handle_scope(isolate);
+
+    auto context = node::NewContext(isolate);
+    CHECK(!context.IsEmpty());
+    v8::Context::Scope context_scope(context);
+
+    std::unique_ptr<node::IsolateData, decltype(&node::FreeIsolateData)>
+      isolate_data{node::CreateIsolateData(isolate,
+                                           &current_loop,
+                                           platform.get()),
+                   node::FreeIsolateData};
+    CHECK(isolate_data);
+
+    std::unique_ptr<node::Environment, decltype(&node::FreeEnvironment)>
+      environment{node::CreateEnvironment(isolate_data.get(),
+                                          context,
+                                          {},
+                                          {}),
+                  node::FreeEnvironment};
+    CHECK(environment);
+    EXPECT_EQ(node::GetEnvironmentIsolateData(environment.get()),
+              isolate_data.get());
+    EXPECT_EQ(node::GetArrayBufferAllocator(isolate_data.get()), nullptr);
+
+    v8::Local<v8::Value> main_ret =
+        node::LoadEnvironment(environment.get(),
+            "'use strict';\n"
+            "const { runInThisContext } = require('vm');\n"
+            "try {\n"
+            "  runInThisContext("
+            "    `process.kill(process.pid, 'SIGINT'); while(true){}`, "
+            "    { breakOnSigint: true });\n"
+            "  return 'unreachable';\n"
+            "} catch (err) {\n"
+            "  return err.code;\n"
+            "}").ToLocalChecked();
+    node::Utf8Value main_ret_str(isolate, main_ret);
+    EXPECT_EQ(std::string(*main_ret_str), "ERR_SCRIPT_EXECUTION_INTERRUPTED");
+  }
+
+  // Cleanup.
+  platform->UnregisterIsolate(isolate);
+  isolate->Dispose();
+}
+#endif  // _WIN32
+
+TEST_F(EnvironmentTest, NestedMicrotaskQueue) {
+  const v8::HandleScope handle_scope(isolate_);
+  const Argv argv;
+
+  std::unique_ptr<v8::MicrotaskQueue> queue = v8::MicrotaskQueue::New(
+      isolate_, v8::MicrotasksPolicy::kExplicit);
+  v8::Local<v8::Context> context = v8::Context::New(
+      isolate_, nullptr, {}, {}, {}, queue.get());
+  node::InitializeContext(context);
+  v8::Context::Scope context_scope(context);
+
+  using IntVec = std::vector<int>;
+  IntVec callback_calls;
+  v8::Local<v8::Function> must_call = v8::Function::New(
+      context,
+      [](const v8::FunctionCallbackInfo<v8::Value>& info) {
+        IntVec* callback_calls = static_cast<IntVec*>(
+            info.Data().As<v8::External>()->Value());
+        callback_calls->push_back(info[0].As<v8::Int32>()->Value());
+      },
+      v8::External::New(isolate_, static_cast<void*>(&callback_calls)))
+          .ToLocalChecked();
+  context->Global()->Set(
+      context,
+      v8::String::NewFromUtf8Literal(isolate_, "mustCall"),
+      must_call).Check();
+
+  node::Environment* env =
+      node::CreateEnvironment(isolate_data_, context, {}, {});
+  CHECK_NE(nullptr, env);
+
+  v8::Local<v8::Function> eval_in_env = node::LoadEnvironment(
+      env,
+      "mustCall(1);\n"
+      "Promise.resolve().then(() => mustCall(2));\n"
+      "require('vm').runInNewContext("
+      "    'Promise.resolve().then(() => mustCall(3))',"
+      "    { mustCall },"
+      "    { microtaskMode: 'afterEvaluate' }"
+      ");\n"
+      "require('vm').runInNewContext("
+      "    'Promise.resolve().then(() => mustCall(4))',"
+      "    { mustCall }"
+      ");\n"
+      "setTimeout(() => {"
+      "  Promise.resolve().then(() => mustCall(5));"
+      "}, 10);\n"
+      "mustCall(6);\n"
+      "return eval;\n").ToLocalChecked().As<v8::Function>();
+  EXPECT_EQ(callback_calls, (IntVec { 1, 3, 6, 2, 4 }));
+  v8::Local<v8::Value> queue_microtask_code = v8::String::NewFromUtf8Literal(
+      isolate_, "queueMicrotask(() => mustCall(7));");
+  eval_in_env->Call(context,
+                    v8::Null(isolate_),
+                    1,
+                    &queue_microtask_code).ToLocalChecked();
+  EXPECT_EQ(callback_calls, (IntVec { 1, 3, 6, 2, 4 }));
+  isolate_->PerformMicrotaskCheckpoint();
+  EXPECT_EQ(callback_calls, (IntVec { 1, 3, 6, 2, 4 }));
+  queue->PerformCheckpoint(isolate_);
+  EXPECT_EQ(callback_calls, (IntVec { 1, 3, 6, 2, 4, 7 }));
+
+  int exit_code = SpinEventLoop(env).FromJust();
+  EXPECT_EQ(exit_code, 0);
+  EXPECT_EQ(callback_calls, (IntVec { 1, 3, 6, 2, 4, 7, 5 }));
+
+  node::FreeEnvironment(env);
+}
+
+static bool interrupted = false;
+static void OnInterrupt(void* arg) {
+  interrupted = true;
+}
+TEST_F(EnvironmentTest, RequestInterruptAtExit) {
+  const v8::HandleScope handle_scope(isolate_);
+  const Argv argv;
+
+  Local<Context> context = node::NewContext(isolate_);
+  CHECK(!context.IsEmpty());
+  context->Enter();
+
+  std::vector<std::string> args(*argv, *argv + 1);
+  std::vector<std::string> exec_args(*argv, *argv + 1);
+  node::Environment* environment =
+      node::CreateEnvironment(isolate_data_, context, args, exec_args);
+  CHECK_NE(nullptr, environment);
+
+  node::RequestInterrupt(environment, OnInterrupt, nullptr);
+  node::FreeEnvironment(environment);
+  EXPECT_TRUE(interrupted);
+
+  context->Exit();
+}
