@@ -2,6 +2,7 @@
 
 const fs = require('fs');
 const path = require('path');
+const { pathToFileURL } = require('url');
 const { isMainThread } = require('worker_threads');
 
 function rmSync(pathname) {
@@ -55,7 +56,26 @@ function onexit() {
   }
 }
 
+function resolve(...paths) {
+  return path.resolve(tmpPath, ...paths);
+}
+
+function hasEnoughSpace(size) {
+  const { bavail, bsize } = fs.statfsSync(tmpPath);
+  return bavail >= Math.ceil(size / bsize);
+}
+
+function fileURL(...paths) {
+  // When called without arguments, add explicit trailing slash
+  const fullPath = path.resolve(tmpPath + path.sep, ...paths);
+
+  return pathToFileURL(fullPath);
+}
+
 module.exports = {
+  fileURL,
+  hasEnoughSpace,
   path: tmpPath,
   refresh,
+  resolve,
 };
