@@ -1,5 +1,3 @@
-/* eslint-disable node-core/required-modules,node-core/require-common-first */
-
 'use strict';
 
 const { runInThisContext } = require('vm');
@@ -10,7 +8,8 @@ const resource = new ResourceLoader(workerData.wptPath);
 
 global.self = global;
 global.GLOBAL = {
-  isWindow() { return false; }
+  isWindow() { return false; },
+  isShadowRealm() { return false; },
 };
 global.require = require;
 
@@ -18,7 +17,7 @@ global.require = require;
 // in Node.js, but some tests and harness depend on this to pull
 // resources.
 global.fetch = function fetch(file) {
-  return resource.read(workerData.filename, file, true);
+  return resource.read(workerData.testRelativePath, file, true);
 };
 
 if (workerData.initScript) {
@@ -26,7 +25,7 @@ if (workerData.initScript) {
 }
 
 runInThisContext(workerData.harness.code, {
-  filename: workerData.harness.filename
+  filename: workerData.harness.filename,
 });
 
 // eslint-disable-next-line no-undef
@@ -42,8 +41,17 @@ add_result_callback((result) => {
   });
 });
 
+// Keep the event loop alive
+const timeout = setTimeout(() => {
+  parentPort.postMessage({
+    type: 'completion',
+    status: { status: 2 },
+  });
+}, 2 ** 31 - 1); // Max timeout is 2^31-1, when overflown the timeout is set to 1.
+
 // eslint-disable-next-line no-undef
 add_completion_callback((_, status) => {
+  clearTimeout(timeout);
   parentPort.postMessage({
     type: 'completion',
     status,
