@@ -4,8 +4,8 @@
 require('../common');
 
 const { strictEqual, throws } = require('assert');
-const { ESMLoader } = require('internal/modules/esm/loader');
-const ModuleMap = require('internal/modules/esm/module_map');
+const { createModuleLoader } = require('internal/modules/esm/loader');
+const { LoadCache, ResolveCache } = require('internal/modules/esm/module_map');
 const ModuleJob = require('internal/modules/esm/module_job');
 const createDynamicModule = require(
   'internal/modules/esm/create_dynamic_module');
@@ -16,7 +16,7 @@ const jsonModuleDataUrl = 'data:application/json,""';
 const stubJsModule = createDynamicModule([], ['default'], jsModuleDataUrl);
 const stubJsonModule = createDynamicModule([], ['default'], jsonModuleDataUrl);
 
-const loader = new ESMLoader();
+const loader = createModuleLoader(false);
 const jsModuleJob = new ModuleJob(loader, stubJsModule.module, undefined,
                                   () => new Promise(() => {}));
 const jsonModuleJob = new ModuleJob(loader, stubJsonModule.module,
@@ -24,11 +24,11 @@ const jsonModuleJob = new ModuleJob(loader, stubJsonModule.module,
                                     () => new Promise(() => {}));
 
 
-// ModuleMap.set and ModuleMap.get store and retrieve module jobs for a
-// specified url/type tuple; ModuleMap.has correctly reports whether such jobs
+// LoadCache.set and LoadCache.get store and retrieve module jobs for a
+// specified url/type tuple; LoadCache.has correctly reports whether such jobs
 // are stored in the map.
 {
-  const moduleMap = new ModuleMap();
+  const moduleMap = new LoadCache();
 
   moduleMap.set(jsModuleDataUrl, undefined, jsModuleJob);
   moduleMap.set(jsonModuleDataUrl, 'json', jsonModuleJob);
@@ -50,10 +50,10 @@ const jsonModuleJob = new ModuleJob(loader, stubJsonModule.module,
   strictEqual(moduleMap.has(jsonModuleDataUrl, 'unknown'), false);
 }
 
-// ModuleMap.get, ModuleMap.has and ModuleMap.set should only accept string
+// LoadCache.get, LoadCache.has and LoadCache.set should only accept string
 // values as url argument.
 {
-  const moduleMap = new ModuleMap();
+  const moduleMap = new LoadCache();
 
   const errorObj = {
     code: 'ERR_INVALID_ARG_TYPE',
@@ -68,10 +68,10 @@ const jsonModuleJob = new ModuleJob(loader, stubJsonModule.module,
   });
 }
 
-// ModuleMap.get, ModuleMap.has and ModuleMap.set should only accept string
+// LoadCache.get, LoadCache.has and LoadCache.set should only accept string
 // values (or the kAssertType symbol) as type argument.
 {
-  const moduleMap = new ModuleMap();
+  const moduleMap = new LoadCache();
 
   const errorObj = {
     code: 'ERR_INVALID_ARG_TYPE',
@@ -86,9 +86,9 @@ const jsonModuleJob = new ModuleJob(loader, stubJsonModule.module,
   });
 }
 
-// ModuleMap.set should only accept ModuleJob values as job argument.
+// LoadCache.set should only accept ModuleJob values as job argument.
 {
-  const moduleMap = new ModuleMap();
+  const moduleMap = new LoadCache();
 
   [{}, [], true, 1].forEach((value) => {
     throws(() => moduleMap.set('', undefined, value), {
@@ -98,3 +98,21 @@ const jsonModuleJob = new ModuleJob(loader, stubJsonModule.module,
     });
   });
 }
+
+{
+  const resolveMap = new ResolveCache();
+
+  strictEqual(resolveMap.serializeKey('./file', { __proto__: null }), './file::');
+  strictEqual(resolveMap.serializeKey('./file', { __proto__: null, type: 'json' }), './file::"type""json"');
+  strictEqual(resolveMap.serializeKey('./file::"type""json"', { __proto__: null }), './file::"type""json"::');
+  strictEqual(resolveMap.serializeKey('./file', { __proto__: null, c: 'd', a: 'b' }), './file::"a""b","c""d"');
+  strictEqual(resolveMap.serializeKey('./s', { __proto__: null, c: 'd', a: 'b', b: 'c' }), './s::"a""b","b""c","c""d"');
+
+  resolveMap.set('key1', 'parent1', 1);
+  resolveMap.set('key2', 'parent1', 2);
+  resolveMap.set('key2', 'parent2', 3);
+
+  strictEqual(resolveMap.get('key1', 'parent1'), 1);
+  strictEqual(resolveMap.get('key2', 'parent1'), 2);
+  strictEqual(resolveMap.get('key2', 'parent2'), 3);
+}
