@@ -39,15 +39,15 @@ function doTest(flags, done) {
   // dictates that it'll resolve relative imports in the main file relative to
   // the symlink, and not relative to the symlink target; the file structure set
   // up above requires this to not crash when loading ./submodule_link.js
-  spawn(process.execPath,
-        flags.concat([
-          '--preserve-symlinks',
-          '--preserve-symlinks-main', entry_link_absolute_path
-        ]),
-        { stdio: 'inherit' }).on('exit', (code) => {
-    assert.strictEqual(code, 0);
-    done();
-  });
+  spawn(process.execPath, [
+    '--preserve-symlinks',
+    '--preserve-symlinks-main',
+    entry_link_absolute_path,
+  ], { stdio: 'inherit' })
+    .on('exit', (code) => {
+      assert.strictEqual(code, 0);
+      done();
+    });
 }
 
 // First test the commonjs module loader
