@@ -1,4 +1,4 @@
-// META: global=worker
+// META: global=window,worker
 
 'use strict';
 
@@ -22,20 +22,6 @@ const badChunks = [
   {
     name: 'array',
     value: [65]
-  },
-  {
-    name: 'SharedArrayBuffer',
-    // Use a getter to postpone construction so that all tests don't fail where
-    // SharedArrayBuffer is not yet implemented.
-    get value() {
-      return new SharedArrayBuffer();
-    }
-  },
-  {
-    name: 'shared Uint8Array',
-    get value() {
-      new Uint8Array(new SharedArrayBuffer())
-    }
   }
 ];
 
@@ -46,7 +32,7 @@ for (const chunk of badChunks) {
     const writer = tds.writable.getWriter();
     const writePromise = writer.write(chunk.value);
     const readPromise = reader.read();
-    await promise_rejects(t, new TypeError(), writePromise, 'write should reject');
-    await promise_rejects(t, new TypeError(), readPromise, 'read should reject');
+    await promise_rejects_js(t, TypeError, writePromise, 'write should reject');
+    await promise_rejects_js(t, TypeError, readPromise, 'read should reject');
   }, `chunk of type ${chunk.name} should error the stream`);
 }
