@@ -3,7 +3,7 @@ const common = require('../common');
 
 // Simulate `cat readfile.js | node readfile.js`
 
-if (common.isWindows || common.isAIX)
+if (common.isWindows || common.isAIX || common.isIBMi)
   common.skip(`No /dev/stdin on ${process.platform}.`);
 
 const assert = require('assert');
@@ -29,8 +29,7 @@ const exec = require('child_process').exec;
 const f = JSON.stringify(__filename);
 const node = JSON.stringify(process.execPath);
 const cmd = `cat ${filename} | ${node} ${f} child`;
-exec(cmd, { maxBuffer: 1000000 }, common.mustCall((err, stdout, stderr) => {
-  assert.ifError(err);
+exec(cmd, { maxBuffer: 1000000 }, common.mustSucceed((stdout, stderr) => {
   assert.strictEqual(
     stdout,
     dataExpected,
@@ -43,7 +42,3 @@ exec(cmd, { maxBuffer: 1000000 }, common.mustCall((err, stdout, stderr) => {
   );
   console.log('ok');
 }));
-
-process.on('exit', function() {
-  fs.unlinkSync(filename);
-});
