@@ -8,7 +8,7 @@ const h2 = require('http2');
 
 const server = h2.createServer();
 
-// we use the lower-level API here
+// We use the lower-level API here
 server.on('stream', common.mustNotCall());
 server.listen(0, common.mustCall(() => {
 
@@ -32,12 +32,12 @@ server.listen(0, common.mustCall(() => {
   }));
 
   req.on('frameError', common.mustCall((type, code) => {
-    assert.strictEqual(code, h2.constants.NGHTTP2_ERR_FRAME_SIZE_ERROR);
+    assert.strictEqual(code, h2.constants.NGHTTP2_FRAME_SIZE_ERROR);
   }));
 
   req.on('error', common.expectsError({
     code: 'ERR_HTTP2_STREAM_ERROR',
-    type: Error,
-    message: 'Stream closed with error code NGHTTP2_REFUSED_STREAM'
+    name: 'Error',
+    message: 'Stream closed with error code NGHTTP2_FRAME_SIZE_ERROR'
   }));
 }));
