@@ -20,8 +20,7 @@ const warnpath = path.join(tmpdir.path, 'warnings.txt');
 
 fork(warnmod, { env: { ...process.env, NODE_REDIRECT_WARNINGS: warnpath } })
   .on('exit', common.mustCall(() => {
-    fs.readFile(warnpath, 'utf8', common.mustCall((err, data) => {
-      assert.ifError(err);
-      assert(/\(node:\d+\) Warning: a bad practice warning/.test(data));
+    fs.readFile(warnpath, 'utf8', common.mustSucceed((data) => {
+      assert.match(data, /\(node:\d+\) Warning: a bad practice warning/);
     }));
   }));
