@@ -1,11 +1,14 @@
+// Flags: --expose-internals
 'use strict';
 const common = require('../common');
 
 if (!common.hasCrypto)
   common.skip('missing crypto');
 
+const assert = require('assert');
 // Monkey-patch SecureContext
-const binding = process.binding('crypto');
+const { internalBinding } = require('internal/test/binding');
+const binding = internalBinding('crypto');
 const NativeSecureContext = binding.SecureContext;
 
 binding.SecureContext = function() {
@@ -17,7 +20,7 @@ binding.SecureContext = function() {
 const tls = require('tls');
 
 {
-  common.expectsError(
+  assert.throws(
     () => { tls.createSecureContext({ clientCertEngine: 'Cannonmouth' }); },
     {
       code: 'ERR_CRYPTO_CUSTOM_ENGINE_NOT_SUPPORTED',
