@@ -23,7 +23,7 @@ const server = tls
     },
     common.mustCall(function(c) {
       assert.strictEqual(c.getPeerCertificate().serialNumber,
-                         'ECC9B856270DA9A8');
+                         '147D36C1C2F74206DE9FAB5F2226D78ADB00A426');
       assert.strictEqual(c.authorizationError, null);
       c.end();
     })
@@ -42,7 +42,7 @@ const server = tls
           // that .getCertificate() does not accidentally decrease the
           // reference count of the X509* certificate on the native side.
           assert.strictEqual(client.getCertificate().serialNumber,
-                             'ECC9B856270DA9A8');
+                             '147D36C1C2F74206DE9FAB5F2226D78ADB00A426');
         }
         client.end();
         server.close();
