@@ -8,7 +8,6 @@ if (!common.hasIntl) {
   common.skip('missing Intl');
 }
 
-const URL = require('url').URL;
 const assert = require('assert');
 const fixtures = require('../common/fixtures');
 
@@ -30,7 +29,7 @@ const typeFailures = [
   { input: new RegExp() },
   { input: 'test', base: null },
   { input: 'http://nodejs.org', base: null },
-  { input: () => {} }
+  { input: () => {} },
 ];
 
 // See https://github.com/w3c/web-platform-tests/pull/10955
@@ -56,11 +55,8 @@ for (const test of failureTests) {
     () => new URL(test.input, test.base),
     (error) => {
       assert.throws(() => { throw error; }, expectedError);
-
-      // The input could be processed, so we don't do strict matching here
-      let match;
-      assert(match = (`${error}`).match(/Invalid URL: (.*)$/));
-      assert.strictEqual(error.input, match[1]);
+      assert.strictEqual(`${error}`, 'TypeError [ERR_INVALID_URL]: Invalid URL');
+      assert.strictEqual(error.message, 'Invalid URL');
       return true;
     });
 }
@@ -82,3 +78,10 @@ for (const test of additional_tests) {
   if (test.search) assert.strictEqual(url.search, test.search);
   if (test.hash) assert.strictEqual(url.hash, test.hash);
 }
+
+assert.throws(() => {
+  new URL();
+}, {
+  name: 'TypeError',
+  code: 'ERR_MISSING_ARGS',
+});
