@@ -1,11 +1,13 @@
+// Flags: --expose-internals --no-warnings
 'use strict';
 
-// see also test/sequential/test-async-wrap-getasyncid.js
+// See also test/sequential/test-async-wrap-getasyncid.js
 
 const common = require('../common');
 const assert = require('assert');
-const tty_wrap = process.binding('tty_wrap');
-const { TTYWRAP } = process.binding('async_wrap').Providers;
+const { internalBinding } = require('internal/test/binding');
+const { TTYWRAP } = internalBinding('async_wrap').Providers;
+const tty_wrap = internalBinding('tty_wrap');
 const providers = { TTYWRAP };
 
 // Make sure that the TTYWRAP Provider is tested.
