@@ -27,9 +27,9 @@ const server = net.createServer(common.mustCall((c) => {
         { node_name: 'Node / NodeBIO', edge_name: 'enc_out' },
         { node_name: 'Node / NodeBIO', edge_name: 'enc_in' },
         // `Node / TLSWrap` (C++) -> `TLSWrap` (JS)
-        { node_name: 'TLSWrap', edge_name: 'wrapped' },
+        { node_name: 'TLSWrap', edge_name: 'native_to_javascript' },
         // pending_cleartext_input could be empty
-      ]
+      ],
     },
   ]);
 }));
