@@ -4,10 +4,8 @@
 # found in the LICENSE file.
 
 import argparse
-import collections
 import json
 import os.path
-import re
 import sys
 
 import pdl
@@ -23,14 +21,12 @@ def main(argv):
     parser.add_argument("json_file", help="The .json output file write.")
     args = parser.parse_args(argv)
     file_name = os.path.normpath(args.pdl_file)
-    input_file = open(file_name, "r")
-    pdl_string = input_file.read()
+    with open(file_name, "r") as input_file:
+        pdl_string = input_file.read()
     protocol = pdl.loads(pdl_string, file_name, args.map_binary_to_string)
-    input_file.close()
 
-    output_file = open(os.path.normpath(args.json_file), 'wb')
-    json.dump(protocol, output_file, indent=4, separators=(',', ': '))
-    output_file.close()
+    with open(os.path.normpath(args.json_file), 'w') as output_file:
+        json.dump(protocol, output_file, indent=4, separators=(',', ': '))
 
 
 if __name__ == '__main__':
