@@ -36,6 +36,7 @@ import os
 import re
 import functools
 import codecs
+import utils
 
 def ReadFile(filename):
   if is_verbose:
@@ -47,24 +48,30 @@ def ReadFile(filename):
 
 TEMPLATE = """
 #include "env-inl.h"
-#include "node_native_module.h"
+#include "node_builtins.h"
 #include "node_internals.h"
 
 namespace node {{
 
-namespace native_module {{
+namespace builtins {{
 
 {0}
 
-void NativeModuleLoader::LoadJavaScriptSource() {{
-  {1}
+namespace {{
+const ThreadsafeCopyOnWrite<BuiltinSourceMap> global_source_map {{
+  BuiltinSourceMap{{ {1} }}
+}};
 }}
 
-UnionBytes NativeModuleLoader::GetConfig() {{
+void BuiltinLoader::LoadJavaScriptSource() {{
+  source_ = global_source_map;
+}}
+
+UnionBytes BuiltinLoader::GetConfig() {{
   return UnionBytes(config_raw, {2});  // config.gypi
 }}
 
-}}  // namespace native_module
+}}  // namespace builtins
 
 }}  // namespace node
 """
@@ -81,11 +88,11 @@ static const uint16_t {0}[] = {{
 }};
 """
 
-INITIALIZER = 'source_.emplace("{0}", UnionBytes{{{1}, {2}}});'
+INITIALIZER = '{{"{0}", UnionBytes{{{1}, {2}}} }},'
 
 CONFIG_GYPI_ID = 'config_raw'
 
-SLUGGER_RE =re.compile('[.\-/]')
+SLUGGER_RE = re.compile(r'[.\-/]')
 
 is_verbose = False
 
@@ -123,7 +130,7 @@ def AddModule(filename, definitions, initializers):
   initializers.append(initializer)
 
 def NormalizeFileName(filename):
-  split = filename.split(os.path.sep)
+  split = filename.split('/')
   if split[0] == 'deps':
     split = ['internal'] + split
   else:  # `lib/**/*.js` so drop the 'lib' part
@@ -140,6 +147,8 @@ def JS2C(source_files, target):
 
   for filename in source_files['.js']:
     AddModule(filename, definitions, initializers)
+  for filename in source_files['.mjs']:
+    AddModule(filename, definitions, initializers)
 
   config_def, config_size = handle_config_gypi(source_files['config.gypi'])
   definitions.append(config_def)
@@ -163,9 +172,11 @@ def handle_config_gypi(config_filename):
 def jsonify(config):
   # 1. string comments
   config = re.sub(r'#.*?\n', '', config)
+  # 2. join multiline strings
+  config = re.sub(r"'$\s+'", '', config, flags=re.M)
   # 3. normalize string literals from ' into "
   config = re.sub('\'', '"', config)
-  # 2. turn pseudo-booleans strings into Booleans
+  # 4. turn pseudo-booleans strings into Booleans
   config = re.sub('"true"', 'true', config)
   config = re.sub('"false"', 'false', config)
   return config
@@ -200,16 +211,36 @@ def main():
     fromfile_prefix_chars='@'
   )
   parser.add_argument('--target', help='output file')
+  parser.add_argument(
+      '--directory',
+      default=None,
+      help='input file directory')
+  parser.add_argument(
+      '--root',
+      default=None,
+      help='root directory containing the sources')
   parser.add_argument('--verbose', action='store_true', help='output file')
   parser.add_argument('sources', nargs='*', help='input files')
   options = parser.parse_args()
   global is_verbose
   is_verbose = options.verbose
-  source_files = functools.reduce(SourceFileByExt, options.sources, {})
-  # Should have exactly 2 types: `.js`, and `.gypi`
-  assert len(source_files) == 2
+  sources = options.sources
+
+  if options.root is not None:
+    os.chdir(options.root)
+
+  if options.directory is not None:
+    js_files = utils.SearchFiles(options.directory, 'js')
+    mjs_files = utils.SearchFiles(options.directory, 'mjs')
+    sources = js_files + mjs_files + options.sources
+
+  source_files = functools.reduce(SourceFileByExt, sources, {})
+
+  # Should have exactly 3 types: `.js`, `.mjs` and `.gypi`
+  assert len(source_files) == 3
   # Currently config.gypi is the only `.gypi` file allowed
-  assert source_files['.gypi'] == ['config.gypi']
+  assert len(source_files['.gypi']) == 1
+  assert os.path.basename(source_files['.gypi'][0]) == 'config.gypi'
   source_files['config.gypi'] = source_files.pop('.gypi')[0]
   JS2C(source_files, options.target)
 
